/*************************************************************************************************
*
*
*			ObjectMacZapp		-- a standard Mac OOP application template
*
*
*
*			ZPictWindow.cpp		-- a window that displays PICT files
*
*
*
*
*
*			 1996, Graham Cox
*
*
*
*
*************************************************************************************************/


#include	"ZPictWindow.h"
#include	"MacZoop.h"
#include	"ZFile.h"

/*--------------------------------***  CONSTRUCTOR  ***---------------------------------*/

ZPictWindow::ZPictWindow( ZCommander* aBoss, const short windID )
	: ZScroller( aBoss, windID, TRUE, TRUE )
{
	thePicture = NULL;
	printable = TRUE;
}


/*---------------------------------***  DESTRUCTOR  ***---------------------------------*/

ZPictWindow::~ZPictWindow()
{
	if (thePicture)
		KillPicture(thePicture);
}

/*--------------------------------***  DRAWCONTENT  ***---------------------------------*/
/*	

overrides ZScroller to draw the picture in the window

----------------------------------------------------------------------------------------*/

void	ZPictWindow::DrawContent()
{
	Rect	pFrame;
	
	if (thePicture)
	{
		pFrame = (*thePicture)->picFrame;
		OffsetRect(&pFrame, -pFrame.left, -pFrame.top);
		DrawPicture(thePicture, &pFrame);
	}
}


/*----------------------------------***  OPENFILE  ***----------------------------------*/
/*	

overrides the base ZWindow to open a PICT file on disk to the window

----------------------------------------------------------------------------------------*/

void	ZPictWindow::OpenFile( const OSType fType )
{
	// read the picture into the picture handle
	
	FInfo	fi;
	short	refNum;
	long	pSize;
	Handle	temp = NULL;
	Rect	pFrame;
	
	if (macFile.vRefNum != kNoFile)
	{
		FailOSErr(FSpGetFInfo(&macFile, &fi));	
		
		if (fi.fdType != 'PICT')
			FailOSErr(paramErr);
			
		FailOSErr(FSpOpenDF(&macFile, fsCurPerm, &refNum ));
		
		// skip the first 512 bytes of the file
		
		try
		{
			FailOSErr(SetFPos(refNum, fsFromStart,  512));
			
			// how big is the rest of the file?
			
			FailOSErr(GetEOF(refNum, &pSize));
			pSize -= 512;
			
			if (pSize > 0)
			{
				FailNIL(temp = NewHandle(pSize));
			
				HLock(temp);
				FailOSErr(FSRead(refNum, &pSize, *temp));
				HUnlock(temp);
			}
			
			FSClose(refNum);
			
			if (thePicture)
				KillPicture(thePicture);
				
			thePicture  = (PicHandle)temp;
			
			// set the scroll dimensions to the picture's frame
			
			pFrame = (*thePicture)->picFrame;
			OffsetRect(&pFrame, -pFrame.left, -pFrame.top);
			SetBounds( pFrame );
			SetScrollAmount(8,8);
			
			pFrame.right += kStdScrollbarWidth;
			pFrame.bottom += kStdScrollbarWidth;
			
			SetSize(pFrame.right, pFrame.bottom);
		}
		catch(OSErr err)
		{
			FSClose(refNum);
			
			if (temp)
			{
				HUnlock(temp);
				DisposeHandle(temp);
			}	
			throw err;
		}
	}
	inherited::OpenFile( fType );
}



/*--------------------------------***  SAVEFILE  ***---------------------------------*/
/*	

overrides the base ZWindow to save the image as a PICT file on disk

----------------------------------------------------------------------------------------*/

void	ZPictWindow::SaveFile()
{
	// this saves the window contents as a PICT file in the file <macFile>.

	Handle		header = NULL;
	ZFile*		theFile;
	
	if ((macFile.vRefNum != kNoFile) && (thePicture != NULL))
	{
		FailNIL( theFile = new ZFile( macFile ));
		theFile->SetType( 'PICT' );
		
		if (! theFile->IsReal())
			theFile->Create();
		
		// open the file for safe-saving
			
		theFile->OpenSafe();
		
		try
		{
			// make a header and write it out. This is just 512 bytes of zeroes.
			
			FailNIL( header = NewHandleClear( 512L ));
			
			theFile->Write( header );
			DisposeHandle(header);
			header = NULL;
			
			// now write the picture handle
			
			theFile->Write((Handle) thePicture );
			theFile->Close();	
		}
		catch( OSErr err )
		{	
			theFile->Close();
			ForgetObject( theFile );
			
			if (header)
				DisposeHandle(header);
			
			throw err;
		}
		
		ForgetObject( theFile );
		inherited::SaveFile();
	}
}

/*--------------------------------***  UPDATEMENUS  ***---------------------------------*/
/*
Enable the Copy command if we have a picture	
----------------------------------------------------------------------------------------*/

void	ZPictWindow::UpdateMenus()
{
	if ( thePicture )
		gMenuBar->EnableCommand( kCmdCopy );

	inherited::UpdateMenus();
}


/*--------------------------------***  CANPASTETYPE  ***--------------------------------*/
/*	
we can accept Paste commands of type PICT
----------------------------------------------------------------------------------------*/

Boolean	ZPictWindow::CanPasteType()
{
	return gClipboard->QueryType( 'PICT' );
}


/*-----------------------------------***  DOCOPY  ***-----------------------------------*/
/*
copy the image to the clipboard	
----------------------------------------------------------------------------------------*/

void	ZPictWindow::DoCopy()
{
	if ( thePicture )
		gClipboard->PutData( thePicture );
}

/*----------------------------------***  DOPASTE  ***-----------------------------------*/
/*
copy a PICT from the clipboard to our local picture	
----------------------------------------------------------------------------------------*/

void	ZPictWindow::DoPaste()
{
	Rect		pFrame;
	PicHandle	p = (PicHandle) gClipboard->GetData( 'PICT' );
	
	FailNIL( p );
	
	if ( thePicture )
		KillPicture( thePicture );
		
	thePicture = p;
	
	pFrame = (*thePicture)->picFrame;
	OffsetRect( &pFrame, -pFrame.left, -pFrame.top );
	SetBounds( pFrame );
	SetScrollAmount( 8, 8 );
	
	pFrame.right += kStdScrollbarWidth;
	pFrame.bottom += kStdScrollbarWidth;
	
	SetSize( pFrame.right, pFrame.bottom );
	
	dirty = TRUE;
}
