/*************************************************************************************************
*
*
*			ObjectMacZapp		-- a standard Mac OOP application template
*
*
*
*			ZButton.cpp			-- a versatile icon button class
*
*
*
*
*
*			 1996, Graham Cox
*
*
*
*
*************************************************************************************************/


#include	"ZButton.h"
#include	"ZCommander.h"
#include	"MacZoop.h"

#include	<icons.h>


/*-------------------------------***   CONSTRUCTOR   ***--------------------------------*/


ZButton::ZButton( ZCommander* aBoss, const short bResID )
	: ZComrade()
{
	itsBoss = aBoss;
	bTemplateID = bResID;
	
	SetRect( &itsFrame, 10, 10, 42, 42 );
	
	itsType = pushButton;
	itsStyle = iconButton | is3DButton;
	itsState = buttonOff;
	
	active = TRUE;
	buttonID = 0;
	puDelay = 0;
	hasSubmenu = FALSE;
	subMenuID = 0;
	menuCheckItem = 0;
	autoCheckMenu = FALSE;
	isPopOutMenu = FALSE;
	itsCommand = kNoCmd;
	buttonData = NULL;
	bDataID = 128;
	borderWidth = 2;
	
	buttonOffColour.red  = buttonOffColour.green = buttonOffColour.blue = 0xCCCC;	// lightish grey
	buttonOnColour.red   = 0x6666;
	buttonOnColour.green = 0xFFFF;
	buttonOnColour.blue  = 0xCCCC;
	
	lastClickTime = 0;
	isDoubleClick = FALSE;
}


/*--------------------------------***   DESTRUCTOR   ***--------------------------------*/

ZButton::~ZButton()
{
	if ( itsSubMenu )
		ReleaseResource((Handle) itsSubMenu );
}


/*-------------------------------***   INITZBUTTON   ***--------------------------------*/
/*
initialise button from template resources	
----------------------------------------------------------------------------------------*/

void	ZButton::InitZButton( const short bIndex )
{
	// if <bIndex> is 0, this is a single button defined by a 'Btn ' resource. If >= 1,
	// this is built as part of a group from a 'Btn#' resource.
	
	ButtonTemplatePtr	btP;
	ButtonListHdl		bListH = NULL;
	ButtonTemplateHdl	bTmplH = NULL;
	
	if ( bIndex > 0 )
	{
		FailNILRes( bListH = (ButtonListHdl) GetResource( kButtonTemplateListResType, bTemplateID ));
		
		// set up btP so it points to the relevant entry, after checking that the index is sensible
		
		if ( bIndex >= 1 && bIndex <= (*bListH)->bCount )
		{
			HNoPurge((Handle) bListH );
			HLock((Handle) bListH );
		
			btP = &(*bListH)->b[ bIndex - 1 ];
		}
		else
		{
			ReleaseResource((Handle) bListH );
			FailOSErr( paramErr );
		}
	}
	else
	{
		FailNILRes( bTmplH = (ButtonTemplateHdl) GetResource( kButtonTemplateResType, bTemplateID ));
		
		HNoPurge((Handle) bTmplH );
		HLock((Handle) bTmplH );
		
		btP = *bTmplH;
	}
	
	// set up the button using the settings pointed at by btP
	
	itsFrame 		= btP->btFrame;
	itsType 		= btP->btType;
	itsStyle 		= btP->btStyle;
	itsState		= buttonOff;
	active			= TRUE;
	buttonID		= bIndex;
	puDelay			= kDefaultPopUpDelay;
	hasSubmenu		= ( btP->btFlags & btHasSubMenu ) && ( btP->btAuxResID1 != 0 );
	subMenuID		= btP->btAuxResID1;
	borderWidth		= MAX( 1, btP->btBorderWidth );
	keyEquivalent	= btP->btKeyEquiv;
	menuCheckItem	= 0;
	autoCheckMenu	= btP->btFlags & btAutoCheckSubmenu;
	isPopOutMenu	= btP->btFlags & btSubMenuIsPopOut;
	itsCommand		= btP->btCommand;
	buttonData		= NULL;
	bDataID			= btP->btAuxResID2;
	itsSubMenu		= NULL;

	// clean up:
	
	if ( bTmplH )
	{
		HUnlock((Handle) bTmplH );
		ReleaseResource((Handle) bTmplH );
	}
	
	if ( bListH )
	{
		HUnlock((Handle) bListH );
		ReleaseResource((Handle) bListH );
	}
	
	// look for a button colour resource to set the colours
	
	ButtonColourHdl		cbH;
	
	cbH = ( ButtonColourHdl ) GetResource( kButtonColoursResType, bTemplateID );
	
	if ( cbH )
	{
		buttonOffColour = (*cbH)->btOffColour;
		buttonOnColour = (*cbH)->btOnColour;
		
		ReleaseResource((Handle) cbH );
	}
}


/*-------------------------------***   CLICKBUTTON   ***--------------------------------*/
/*
process mouse click in the button
----------------------------------------------------------------------------------------*/

Boolean	ZButton::ClickButton( const Point aMousePt, const short modifiers )
{
	// process a mouse click in this button, This returns TRUE if the mouse was in this button,
	// else false.
	
	long		mSelect, ckTime;
	Boolean		wasIn;
	
	ckTime = TickCount();
	
	isDoubleClick = ( ckTime <  ( lastClickTime + GetDblTime()));
	lastClickTime = isDoubleClick? 0 : ckTime;
	
	wasIn = ( PtInRect( aMousePt, &itsFrame ) && active );
	
	if ( wasIn )
	{
		// what we do depends on the type of button. If an interlocking button, we
		// select the button on the mouse down, otherwise we track it and respond to the
		// mouse up.
		
		if ( itsType == interlockButton )
		{
			SelectButton( buttonOn );

			mSelect = PopUpSubmenu();
			SendMenuCommand( mSelect );
		}
		else
		{
			if ( TrackButton( aMousePt ))
			{
				switch ( itsType & 0x0F )
				{
					case pushButton:
						// send the command to the boss
						
						SendCommand();
						break;
					case toggleButton:
						SelectButton(( itsState == buttonOn )? buttonOff : buttonOn );
						break;
				}
			}
		}
	}
	
	return wasIn;
}


/*-------------------------------***   DRAWBUTTON   ***---------------------------------*/
/*
draw the button in the current state, style, etc
----------------------------------------------------------------------------------------*/

void	ZButton::DrawButton()
{
	// draws the button in the current state according to the current style, etc. This draws
	// in the current port, so it's important for the caller to make sure the right (owner)
	// window is currently focussed.
	
	Rect			r = itsFrame;
	CIconHandle		ciH;
			
	PenNormal();
	ForeColor( blackColor );
	BackColor( whiteColor );
	
	ClipRect( &itsFrame );
	
	if ( active && ( itsStyle & is3DButton ))
		InsetRect( &r, borderWidth, borderWidth );
	else
		InsetRect( &r, 1, 1 );
		
	FrameRect( &itsFrame );
	
	// draw background of button in colour
	
	if ( itsState == buttonOn )
		RGBForeColor( &buttonOnColour );
	else
		RGBForeColor( &buttonOffColour );
		
	PaintRect( &r );
	ForeColor( blackColor );
	
	if ( active )
		DrawBorder();
	
	switch ( itsStyle & 0x0F )
	{
		case textButton:
		case plainButton:
			
			if ( itsState == buttonOn )
				PaintRect( &r );
			else
				EraseRect( &r );
			break;
		case iconButton:
			ciH = GetCIcon( bDataID );
			
			if ( ciH )
			{
				// for icon button, plot icon with ID in <bDataID>. We use the
				// icon's true rect, centred in the button
				
				r = (*ciH)->iconMask.bounds;
				OffsetRect( &r, -r.left, -r.top );
				OffsetRect( &r,
							(itsFrame.right + itsFrame.left) / 2 - ( r.right / 2 ),
							(itsFrame.top + itsFrame.bottom) / 2 - ( r.bottom / 2 ));
							
				// if button has submenu, move icon to the left slightly to allow more
				// room for the arrow
				
				if ( hasSubmenu )
					OffsetRect(&r, -3, 0 );
				
				// if button on, offset 1 pixel right and down
				
				if ( itsState == buttonOn )
					OffsetRect( &r, 1, 1 );
					
				short	transform = ttDisabled;
				
				if ( active )
					transform = ( itsState == buttonOn )? ttSelected : ttNone;
				
				PlotCIconHandle( &r,
								 atVerticalCenter + atHorizontalCenter,
								 transform,
								 ciH );
				DisposeCIcon( ciH );
			}
			break;
		case pictButton:
			break;	
	}
	
	// if the button has a submenu attached, draw the arrow icon
	
	if ( hasSubmenu )
	{
		ciH = GetCIcon( kMenuArrowIconResID );
		
		if ( ciH )
		{
			r = (*ciH)->iconMask.bounds;
			
			// position arrow in bottom right corner of button, but inside border
			
			OffsetRect( &r, -r.left, -r.top );
			OffsetRect( &r,
						itsFrame.right - r.right - borderWidth - 2,
						itsFrame.bottom - r.bottom - borderWidth - 2 );
			if ( itsState == buttonOn )
				OffsetRect( &r, 1, 1 );
				
			PlotCIconHandle( &r, atNone, ttNone, ciH );
			DisposeCIcon( ciH );
		}
	}
}			


/*------------------------------***   SELECTBUTTON   ***--------------------------------*/
/*
set button's state, if different to current one
----------------------------------------------------------------------------------------*/

void	ZButton::SelectButton( const ButtonState onOffState )
{
	if ( onOffState == buttonOn &&
		 itsType != pushButton )
		SendCommand();

	if ( itsState != onOffState )
	{
		itsState = onOffState;
		
		if ( itsState == buttonOff )
		{
			lastClickTime = 0;
			SendMessage( kButtonTurningOff, NULL );
		}
		else
			SendMessage( kButtonTurningOn, NULL );

		DrawButton();
	}
}



/*---------------------------------***   ACTIVATE   ***---------------------------------*/
/*
set whether button is active or not
----------------------------------------------------------------------------------------*/

void	ZButton::Activate( const Boolean isActive )
{
	if ( isActive != active )
	{
		active = isActive;
		
		DrawButton();
	}
}



/*-------------------------------***   TRACKBUTTON   ***--------------------------------*/
/*
track mouse in button, highlighting as we go
----------------------------------------------------------------------------------------*/

Boolean	ZButton::TrackButton( Point mouse )	
{
	Boolean		isIn = TRUE, wasIn = FALSE;
	
	// assume mouse is initially in
	
	while( WaitMouseUp())
	{
		if ( isIn != wasIn )
		{
			SelectButton(( ButtonState ) isIn );
		
			wasIn = isIn;
			
			if ( isIn )
				PopUpSubmenu();
		}
		
		GetMouse( &mouse );
		isIn = PtInRect( mouse, &itsFrame );
	}
	
	// mouse released, so unhilight button
	
	SelectButton( buttonOff );
	return wasIn;
}


/*-------------------------------***   SENDCOMMAND   ***--------------------------------*/
/*
send button's command number to boss.
----------------------------------------------------------------------------------------*/

void	ZButton::SendCommand()
{
	if ( isDoubleClick )
		SendMessage( kButtonDoubleClicked, &itsCommand );
	
	if ( itsBoss && itsCommand != kNoCmd )
		itsBoss->HandleCommand( itsCommand );
}


/*-----------------------------***   SENDMENUCOMMAND   ***------------------------------*/
/*
send menu command number to boss.
----------------------------------------------------------------------------------------*/

void	ZButton::SendMenuCommand( const long mCmd )
{
	if ( itsBoss && mCmd != 0 )
		itsBoss->HandleCommand( HiWord( mCmd ), LoWord( mCmd ));
}


/*--------------------------------***   DRAWBORDER   ***--------------------------------*/
/*
draws 3D border of button in relevant state, taking into account colours and border width
----------------------------------------------------------------------------------------*/

void	ZButton::DrawBorder()
{
	if ( itsStyle & is3DButton )
	{
		RGBColor	lightShade;
		RGBColor	darkShade;
		
		// get the shades for the border. This is based on the current OFF colour, not
		// the highlight colour.
		
		lightShade = darkShade = buttonOffColour;
		
		lightShade.red 		= MIN( lightShade.red + 0x3333, 0xFFFF );
		lightShade.green 	= MIN( lightShade.green + 0x3333, 0xFFFF );
		lightShade.blue 	= MIN( lightShade.blue + 0x3333, 0xFFFF );
	
		darkShade.red 		= MAX( darkShade.red - 0x6666, 0 );
		darkShade.green 	= MAX( darkShade.green - 0x6666, 0 );
		darkShade.blue 		= MAX( darkShade.blue - 0x6666, 0 );
	
		// where and how we draw these depends on the state and border width:
		
		Rect	r = itsFrame;
		InsetRect( &r, 1, 1 );
		
		short	t = borderWidth;
		
		while( t-- )
		{
			if ( itsState == buttonOn )
				RGBForeColor( &darkShade );
			else
				RGBForeColor( &lightShade );
		
			MoveTo( r.right - 1, r.top );
			LineTo( r.left, r.top );
			LineTo( r.left, r.bottom - 1 );
			
			if ( itsState == buttonOff )
				RGBForeColor( &darkShade );
			else
				RGBForeColor( &lightShade );
				
			LineTo( r.right - 1, r.bottom - 1 );
			LineTo( r.right - 1, r.top + 1 );
			
			InsetRect( &r, 1, 1 );
		}
		
		ForeColor( blackColor );
	}
}



/*-------------------------------***   POPUPSUBMENU   ***-------------------------------*/
/*
if the menu has an attached menu, this pops it up (or out) and returns the selected item.
This honours the sub-menu delay set for the button, and may keep control for a while.
----------------------------------------------------------------------------------------*/

long	ZButton::PopUpSubmenu()
{
	long	mSelect = 0;
	long	curTim;
	
	// have we got a submenu?
	
	if ( hasSubmenu )
	{
		// yes, so as long as the mouse is down we need to wait for the submenu delay
		// to elapse
		
		curTim = TickCount();
		
		while( Button() && (( curTim + puDelay ) >= TickCount()));
		
		// OK, now it's time to pop up the menu if the mouse is still down
		
		if ( StillDown())
		{
			// compute the position of the menu
			
			Point	menuLoc;
			
			menuLoc.h = itsFrame.right - 4;
			menuLoc.v = itsFrame.bottom - 16;
			LocalToGlobal( &menuLoc );
			
			// get and install the menu if necessary
			
			if ( itsSubMenu == NULL )
				itsSubMenu = GetMenu( subMenuID );		
		
			if ( itsSubMenu )
			{
				// if we're auto-checking the menu, we should do that now:
				
				if ( autoCheckMenu )
					CheckItem( itsSubMenu, menuCheckItem, TRUE );

				InsertMenu( itsSubMenu, hierMenu );
				mSelect = PopUpMenuSelect( itsSubMenu, menuLoc.v, menuLoc.h, 1 );
				DeleteMenu( subMenuID );
				
				if ( HiWord( mSelect ) == subMenuID )
				{
					// if auto-checking, update the checked item
					
					if ( autoCheckMenu && LoWord( mSelect ) != menuCheckItem )
					{
						CheckItem( itsSubMenu, menuCheckItem, FALSE );
						menuCheckItem = LoWord( mSelect );
					}
				}
			}
		}
	}
	
	return mSelect;
}
