/*************************************************************************************************
*
*
*			ObjectMacZapp		-- a standard Mac OOP application template
*
*
*
*			ZButton.h			-- a versatile icon button class
*
*
*
*
*
*			 1996, Graham Cox
*
*
*
*
*************************************************************************************************/


#pragma once

#ifndef	__ZBUTTON__
#define __ZBUTTON__


#include	"ZComrade.h"


class	ZCommander;


enum
{
	pushButton = 1,
	toggleButton,
	interlockButton,
	hasPopUpDelay = 16,
	hasSubMenu = 32
};

typedef short ButtonType;


enum
{
	plainButton = 1,
	textButton,
	iconButton,
	pictButton,
	is3DButton = 16
};

typedef short ButtonStyle;


typedef enum
{
	buttonOff,
	buttonOn
}
ButtonState;

// template resource format

#if PRAGMA_ALIGN_SUPPORTED
#pragma options align=mac68k
#endif

typedef struct
{
	Rect				btFrame;			// button frame	
	short				btType;				// button type
	short				btStyle;			// button style
	short				btAuxResID1;		// res ID of auxiliary resource (1- menu )
	short				btAuxResID2;		// res ID of auxiliary resource (2- cicn, STR# or PICT)
	unsigned char		btFlags;			// button feature flags
	unsigned char		btPadding;
	short				btBorderWidth;		// width of button's border
	unsigned char		btKeyEquiv;			// key equivalent (if used)
	unsigned char		btReserved;			// for future use
	long				btCommand;			// the button's command
}
ButtonTemplateRec, *ButtonTemplatePtr, **ButtonTemplateHdl;

// button flags (undefined flags are reserved for future use):

enum
{
	btHasSubMenu = 0x80,
	btSubMenuIsPopOut = 0x40,
	btAutoCheckSubmenu = 0x20
};

// button template list is straight array of this struct with header:

typedef struct
{
	short				bRows;				// number of rows
	short				bCols;				// number of columns
	Rect				bCommonFrame;		// common frame size of buttons (overrides individual frames)
	long				bFlags;				// feature flags
	long				bRefCon;			// application use
	long				bReserved;			// reserved for future use
	short				bCount;				// count of button records
	ButtonTemplateRec	b[1];
}
ButtonListRec, *ButtonListPtr, **ButtonListHdl;

// colour resource template

typedef struct
{
	RGBColor			btOffColour;
	RGBColor			btOnColour;
}
ButtonColourRec, *ButtonColourPtr, **ButtonColourHdl;


#define		kButtonTemplateResType		'Btn '
#define		kButtonTemplateListResType	'Btn#'
#define		kButtonColoursResType		'Btco'

#define		kMenuArrowIconResID			127
#define		kDefaultPopUpDelay			15

#if PRAGMA_ALIGN_SUPPORTED
#pragma options align=reset
#endif


class	ZButton	: public ZComrade
{
protected:

	ZCommander*		itsBoss;				// owning object
	Rect			itsFrame;				// button's frame in enclosure's coordinates
	ButtonType		itsType;				// button type
	ButtonStyle		itsStyle;				// button style
	ButtonState		itsState;				// button's state (on/off/other)
	Boolean			active;					// TRUE if active
	short			bTemplateID;			// resource ID of template resource
	short			buttonID;				// ID number of this button
	short			puDelay;				// delay for pop-up sub-menu
	short			subMenuID;				// menu ID of submenu
	short			borderWidth;			// width of bevel border, if any
	unsigned char	keyEquivalent;			// keyboard equivalent (may not be used)
	Boolean			hasSubmenu;				// TRUE if there is an attached submenu
	MenuHandle		itsSubMenu;				// handle to submenu
	short			menuCheckItem;			// currently checked item
	Boolean			autoCheckMenu;			// TRUE if we auto-check the menu
	Boolean			isPopOutMenu;			// TRUE if submenu is additional buttons
	long			itsCommand;				// command number this button sends
	Handle			buttonData;				// handle to data- varies with style
	short			bDataID;				// ID of data, depends on style
	RGBColor		buttonOffColour;		// colour of button background (OFF state)
	RGBColor		buttonOnColour;			// colour of button bvackground (ON state )
	long			lastClickTime;			// last click time for detecting double-clicks
	Boolean			isDoubleClick;			// the click was a double

public:

	ZButton( ZCommander* aBoss, const short bResID );
	~ZButton();

	virtual void	InitZButton( const short bIndex );
	
	inline	void	SetCommandNumber( const long aCmd ) { itsCommand = aCmd; };
	inline  long	GetCommandNumber() { return itsCommand; };
	inline	void	SetSubMenuDelay( const short aDelay ) { puDelay = aDelay; };
	inline	void	SetButtonRect( const Rect& aRect ) { itsFrame = aRect; };
	inline	void	GetButtonRect( Rect* aRect ) { *aRect = itsFrame; };
	
	inline	ButtonType	GetButtonType() { return itsType; };
	
	virtual Boolean	ClickButton( const Point aMousePt, const short modifiers );
	virtual void	DrawButton();
	virtual void	SelectButton( const ButtonState onOffState );
	virtual void	Activate( const Boolean isActive );

protected:

	virtual Boolean	TrackButton( Point mouse );	
	virtual void	SendCommand();
	virtual void	SendMenuCommand( const long mCmd );
	virtual void	DrawBorder();
	virtual long	PopUpSubmenu();
};


// comrade messages:

enum
{
	kButtonTurningOn 		= 'bt++',
	kButtonTurningOff 		= 'bt--',
	kButtonSubMenuSelection	= 'btms',
	kButtonDoubleClicked	= 'btkk'
};


/*

ZButtons can be used singly or in groups, and can be used to implement numerous types of
interface elements, such as toolbars, etc. ZButtons work by storing a command number and calling
their boss's HandleCommand method with that command when clicked. They may also send other
messages when clicked, such as informing other buttons in a group that they should turn off
if its part of an interlocked set.

Buttons are generally built from resource templates since their settings are moderately complex.
Single buttons are built from 'Btn ' resources, or from a list of buttons in a set- 'Btn#'
resources. If you want to "Hand build" a button, you'll need to oeverride InitZButton and set
up the fields as necessary.

Buttons can have any rectangular shape, and contain text, icons or pict's, optionally with a 3D
effect border.

Buttons can have an attached submenu. Info about this is supplied in the template- if present
the existence of a submenu causes an arrow to be shown in the bottom right corner of the button.
Submenus can appear after a delay if desired- useful for interlocking buttons. Menus can be
standard text menus, or custom menus including so-called "pop-out" menus which act like extra
buttons in a set.

*/


#endif