/*************************************************************************************************
*
*
*			ObjectMacZapp		-- a standard Mac OOP application template
*
*
*
*			ZGradient.cpp		-- an object for implementing gradients for paint engine
*
*
*
*
*
*			 1996, Graham Cox
*
*
*
*
*************************************************************************************************/


#include	"ZGradient.h"
#include	"ZGWorld.h"
#include	"ZErrors.h"
#include	"ZDefines.h"

#include	<fp.h>


#include	"ZObjectArray.cpp"


ZGradientList*	gGradientList = new ZGradientList();


/*------------------------------***   CONSTRUCTOR   ***---------------------------------*/

ZGradient::ZGradient()
	: ZComrade()
{
	InitGradient();
}

/*-------------------------------***   DESTRUCTOR   ***---------------------------------*/

ZGradient::~ZGradient()
{
}


/*------------------------***   GRADIENTFILLREGIONINPORT   ***--------------------------*/
/*
main external call- this fills the passed region with the set gradient in the current
grafPort.
----------------------------------------------------------------------------------------*/

void	ZGradient::GradientFillRegionInPort( RgnHandle aRgn, Point centrePt )
{
	Rect	src, dest;
	
	FailNILParam( aRgn );
	destRgn = aRgn;
	
	dest = (*aRgn)->rgnBBox;
	BuildGradientWorld( &dest, centrePt );
	
	grTemp->GetSize( &src );
	
	ForeColor( blackColor );
	BackColor( whiteColor );
	
	grTemp->CopyOut( &qd.thePort->portBits, &src, &dest, ditherCopy, aRgn );
	
	ForgetObject( grTemp );
	
	destRgn = NULL;
}


/*--------------------------***   GRADIENTFILLRECTINPORT   ***--------------------------*/
/*
alternative call uses a rect parameter
----------------------------------------------------------------------------------------*/

void	ZGradient::GradientFillRectInPort( Rect* aFrame, Point centrePt )
{
	RgnHandle	temp;
	
	FailNIL( temp = NewRgn());
	
	RectRgn( temp, aFrame );
	GradientFillRegionInPort( temp, centrePt );
	DisposeRgn( temp );
}


/*-------------------------***   GETGRADIENTTHUMBNAIL   ***-----------------------------*/
/*
returns a picture of an example gradient. This can be used in user interfaces that
display a set of preset gradients.
----------------------------------------------------------------------------------------*/

PicHandle		ZGradient::GetGradientThumbnail()
{
	Rect		r = { 0, 0, 32, 32 };
	Point		exPt = { 8, 8 };
	PicHandle	p = NULL;
	
	BuildGradientWorld( &r, exPt );
	
	p = grTemp->MakePicture();	
	ForgetObject( grTemp );
	
	return p;
}


/*---------------------------***   SETGRADIENTCOLOUR   ***------------------------------*/
/*
sets the indexed gradient colour to <aColour>
----------------------------------------------------------------------------------------*/

void	ZGradient::SetGradientColour( const short index, RGBColor* aColour )
{
	grColours[ index ] = *aColour;
}


/*---------------------------***   SETGRADIENTOFFSET   ***------------------------------*/
/*
sets the indexed gradient offset to <offset>
----------------------------------------------------------------------------------------*/

void	ZGradient::SetGradientOffset( const short index, const short offset )
{
	grOffsets[ index ] = offset;
}

/*------------------------------***   INITGRADIENT   ***--------------------------------*/
/*
initialises the gradient to black/white top to bottom with evenly spaced offsets
----------------------------------------------------------------------------------------*/

void	ZGradient::InitGradient()
{
	grType = gradTopToBottom;
	grCount = 1;
	grTemp = NULL;
	grAngle = 45;
	
	RGBColor	white = {-1,-1,-1};
	RGBColor	black = { 0, 0, 0 };
	short		i;
	
	for( i = 0; i < kMaxGradientColours; i++)
	{
		grColours[i] = white;
		grOffsets[i] = i * 20;
	}
		
	grColours[0] = black;
}
	
/*---------------------------***   BUILDGRADIENTWORLD   ***-----------------------------*/
/*
makes an offscreen 32-bit GWorld containing the set gradient. This is then usually
CopyBitsed to the current port to draw the gradient.
----------------------------------------------------------------------------------------*/

void	ZGradient::BuildGradientWorld( Rect* aFrame, Point centrePt )
{
	Rect		r = *aFrame;
	Rect		sq;
	CGrafPtr	savePort;
	GDHandle	saveDev;
	short		i, p, aStart, gc, scanDir;
	Point		ls, le;
	long		range;
	CGradInfo	ci;
	RGBColor	c1, c2;
	RgnHandle	shape = NULL;
	
	OffsetRect( &r, -r.left, -r.top );
	FailNIL( grTemp = new ZGWorld( r, 32 )); // always built with 32-bits
	
	MapPt( &centrePt, aFrame, &r );
	
	grTemp->SetPortToGW( &savePort, &saveDev );
	grTemp->Lock();
	
	// now we can do the "tricky" stuff. For each colour in the gradient, we need to calculate
	// the range over which it is applied, compute the diffs and increments, and then draw in
	// that colour according to the style chosen.
	
	gc = grCount;
	
	switch ( grType )
	{
		case gradTopToBottom:
		case gradLeftToRight:
			MoveTo( r.left, r.top );
			break;
		case gradBottomToTop:
			MoveTo( r.left, r.bottom - 1 );
			break;
		case gradRightToLeft:
			MoveTo( r.right - 1, r.top );
			break;
		case gradShapeburst:
			shape = NewRgn();
			CopyRgn( destRgn, shape );
			OffsetRgn( shape, -(*shape)->rgnBBox.left, -(*shape)->rgnBBox.top );
			break;
		case gradSunburst:
			sq = r;
			CalcEnclosingSquare( &sq, centrePt );
			break;
		case gradRadial:
			sq = r;
			CalcEnclosingSquare( &sq, centrePt );
			aStart = grAngle;
			gc = grCount + 1;
			break;
		case gradDirectional:
			scanDir = CalcDirectionalScan( &r, &ls, &le );
			break;
	}
		
	for( i = 0; i < gc; i++ )
	{
		c1 = grColours[i];
		c2 = grColours[i + 1];
		
		if (( grType == gradRadial ) && ( i == grCount ))
			c2 = grColours[0];
		
		range = CalcGradRange( &r, i, centrePt );
		CalcColours( range, &c1, &c2, &ci );
		
		// now loop through the range and draw
		
		for( p = 0; p <= range; p++ )
		{
			RGBForeColor( &c1 );
			
			switch ( grType )
			{
				case gradTopToBottom:
					// linear gradients drawn with lines
					Line( r.right - r.left, 0 );
					Move( -( r.right - r.left ), 1 );
					break;
				case gradLeftToRight:
					Line( 0, r.bottom - r.top );
					Move( 1, -( r.bottom - r.top ));
					break;
				case gradBottomToTop:
					Line( r.right - r.left, 0 );
					Move( -( r.right - r.left ), -1 );
					break;
				case gradRightToLeft:
					Line( 0, r.bottom - r.top );
					Move( -1, -( r.bottom - r.top ));
					break;
				case gradSunburst:
					// sunburst is circular oval repeatedly inset
					FrameOval( &sq );
					InsetRect( &sq, 1, 1 );
					break;
				case gradShapeburst:
					// shapeburst is region repeatedly inset
					FrameRgn( shape );
					InsetRgn( shape, 1, 1 );
					break;
				case gradRadial:
					// radials are drawn as arcs in 1 increments
					PaintArc( &sq, aStart % 360, 1 );
					aStart++;
					break;
				case gradDirectional:
					MoveTo( ls.h, ls.v );
					LineTo( le.h, le.v );
					switch ( scanDir )
					{
						case scanLeftToRight:
							ls.h++;
							le.h++;
							break;
						case scanTopToBottom:
							ls.v++;
							le.v++;
							break;
						case scanBottomToTop:
							ls.v--;
							le.v--;
							break;
						case scanRightToLeft:
							ls.h--;
							le.h--;
							break;
					}
					break;
					
			}
			IncrementColour( &c1, &ci );
		}		
	}
	
	if ( shape )
		DisposeRgn( shape );
	
	grTemp->Unlock();
	SetGWorld( savePort, saveDev );
}


/*-------------------------------***   CALCCOLOURS   ***--------------------------------*/
/*
computes the differences and increments between two colours c1 and c2, based on the
number of iterations passed in <gradRange>. This is an atomic function that is used to
calculate most gradients.
----------------------------------------------------------------------------------------*/

void	ZGradient::CalcColours( long gradRange, RGBColor* c1, RGBColor* c2, CGradInfo* grInfo )	
{
	grInfo->redDiff   = ((long) c2->red 	& 0xFF00 ) - ((long) c1->red 	& 0xFF00 );	
	grInfo->greenDiff = ((long) c2->green 	& 0xFF00 ) - ((long) c1->green 	& 0xFF00 );	
	grInfo->blueDiff  = ((long) c2->blue 	& 0xFF00 ) - ((long) c1->blue 	& 0xFF00 );	
	
	// increments are set to the difference divided by the range
	
	grInfo->redInc 	  = grInfo->redDiff 	/ gradRange;
	grInfo->greenInc  = grInfo->greenDiff 	/ gradRange;
	grInfo->blueInc   = grInfo->blueDiff 	/ gradRange;
}


/*-----------------------------***   INCREMENTCOLOUR   ***------------------------------*/
/*
modify colour c according to the increments set in the gradInfo record.
----------------------------------------------------------------------------------------*/

void	ZGradient::IncrementColour( RGBColor* c, CGradInfo* grInfo )
{
	c->red 	 += grInfo->redInc;
	c->green += grInfo->greenInc;
	c->blue  += grInfo->blueInc;
}


/*------------------------------***   CALCGRADRANGE   ***-------------------------------*/
/*
compute the range over which the indexed gradient will be applied. This is a number of
pixels between each colour for the range given.
----------------------------------------------------------------------------------------*/

long	ZGradient::CalcGradRange( Rect* r, short index, Point centre )
{
	long	range = 1;
	Rect	sq;
	
	switch ( grType )
	{
		case gradTopToBottom:
		case gradBottomToTop:
			range = ( r->bottom - r->top ) / grCount;
			break;
		case gradLeftToRight:
		case gradRightToLeft:
			range = (r->right - r->left ) / grCount;
			break;
		case gradDirectional:
			Point	a, b;
			short sd = CalcDirectionalScan( r, &a, &b );
			if ( sd == scanTopToBottom ||
				 sd == scanBottomToTop )
				range = ( r->bottom - r->top ) / grCount;
			else
				range = ( r->right - r->left ) / grCount;	
			break;
		case gradRadial:
			range = 360 / ( grCount + 1 );
			break;
		case gradShapeburst:
			// for this style, the range depends on how many insets are needed to the shape
			// before it becomes empty. Thus we "preflight" this using a temporary region.
			RgnHandle	t = NewRgn();
			long		p = 1;
			CopyRgn( destRgn, t );
			
			while( ! EmptyRgn( t ))
			{
				InsetRgn( t, 1, 1 );
				p++;
			}
			DisposeRgn( t );
			range = p / grCount;
			break;
		case gradSunburst:
			sq = *r;
			CalcEnclosingSquare( &sq, centre );
			range = ( sq.right - sq.left ) / ( grCount * 2 );
			break;
	}
	
	return range;
}	


/*---------------------------***   CALCENCLOSINGSQUARE   ***----------------------------*/
/*
for the sunburst fill, we need to know what the enclosing square around r is that has
the arbitrary point <centre> is its centre. The fill takes place within this imaginary
square to give the expected result.
----------------------------------------------------------------------------------------*/

void	ZGradient::CalcEnclosingSquare( Rect* r, Point centre )	
{
	short	dx, dy, sr;
	
	dx = MAX( centre.h - r->left, r->right - centre.h );
	dy = MAX( centre.v - r->top, r->bottom - centre.v );
	
	// radius of enclosing circle is the larger dimension of the two * root 2.
	
	dx = MAX( dx, dy );
	sr = sqrt((dx * dx) + (dx * dx));
	
	r->left   = centre.h - sr;
	r->right  = centre.h + sr;
	r->top    = centre.v - sr;
	r->bottom = centre.v + sr;
}


/*---------------------------***   CALCDIRECTIONALSCAN   ***----------------------------*/
/*
the directional scan proceeds across or down depending on the angle- we always scan the
shorter dimension.
----------------------------------------------------------------------------------------*/

short	ZGradient::CalcDirectionalScan( Rect* aFrame, Point* start, Point* end )	
{
	short	scanDir;
	
	if ( grAngle > 315 || grAngle <= 45 )
	{
		scanDir = scanTopToBottom;
		*start = topLeft( *aFrame );
	}
	else
	{
		if ( grAngle > 45 && grAngle <= 135 )
		{
			scanDir = scanRightToLeft;
			start->h = aFrame->right - 1;
			start->v = aFrame->top;	
		}
		else
		{
			if ( grAngle > 135 && grAngle <= 225 )
			{
				scanDir = scanBottomToTop;
				start->h = aFrame->left;
				start->v = aFrame->bottom - 1;
			}
			else
			{
				scanDir = scanLeftToRight;
				*start = topLeft( *aFrame );
			}
		}
	}
	
	// compute the initial points <start> and <end> for the line to give the correct angle
	
	Fixed slope = SlopeFromAngle( grAngle );
	
	return scanDir;
}



/*------------------------------***   GETGRADHANDLE   ***-------------------------------*/
/*
get the gradient info as a handle
----------------------------------------------------------------------------------------*/

GradientRecHdl	ZGradient::GetGradHandle()
{
	// return a handle with the gradient info in it. This is used to save the gradient in
	// an external prefs file
	
	GradientRecHdl grH = NULL;
	
	FailNIL( grH = (GradientRecHdl) NewHandleClear( sizeof( GradientRecord )));
	
	(*grH)->grType = grType;
	(*grH)->grCount = grCount;
	(*grH)->grAngle = grAngle;
	
	for( short i = 0; i < kMaxGradientColours; i++ )
	{
		(*grH)->grColours[i] = grColours[i];
		(*grH)->grOffsets[i] = grOffsets[i];
	}
	
	return grH;
}


/*------------------------------***   SETGRADHANDLE   ***-------------------------------*/
/*
set the gradient info from a handle
----------------------------------------------------------------------------------------*/

void			ZGradient::SetGradHandle( GradientRecHdl aHandle )
{
	// set the gradient from an external handle, possibnly obtained from e.g. a prefs file.
	
	FailNILParam( aHandle );
	
	grType = (*aHandle)->grType;
	grCount = (*aHandle)->grCount;
	grAngle = (*aHandle)->grAngle;
	
	for( short i = 0; i < kMaxGradientColours; i++ )
	{
		grColours[i] = (*aHandle)->grColours[i];
		grOffsets[i] = (*aHandle)->grOffsets[i];
	}
}

