/*************************************************************************************************
*
*
*			ObjectMacZapp		-- a standard Mac OOP application template
*
*
*
*			ZGradient.h			-- an object for implementing gradients for paint engine
*
*
*
*
*
*			 1996, Graham Cox
*
*
*
*
*************************************************************************************************/


#pragma once

#ifndef	__ZGRADIENT__
#define __ZGRADIENT__


#include	"ZComrade.h"
#include	"ZObjectArray.h"

class	ZGWorld;

// gradient types:

enum
{
	gradTopToBottom = 1,
	gradBottomToTop,
	gradLeftToRight,
	gradRightToLeft,
	gradDirectional,
	gradRadial,
	gradShapeburst,
	gradSunburst
};

typedef short GradientType;

#define		kMaxGradientColours		5

// structure used internally to compute gradient increments:

typedef struct
{
	long	redDiff;
	long	greenDiff;
	long	blueDiff;
	short	redInc;
	short	greenInc;
	short	blueInc;
}
CGradInfo;

// constants for scanning directional fills

enum
{
	scanLeftToRight,
	scanTopToBottom,
	scanRightToLeft,
	scanBottomToTop
};

// structure for saving gradient data externally:

#if PRAGMA_ALIGN_SUPPORTED
#pragma options align=mac68k
#endif

typedef struct
{
	GradientType			grType;								// gradient style
	short					grCount;							// how many colours
	RGBColor				grColours[ kMaxGradientColours ];	// the colours
	short					grOffsets[ kMaxGradientColours ];	// positions
	short					grAngle;							// angle for directional ones
}
GradientRecord, *GradientRecPtr, **GradientRecHdl;


#if PRAGMA_ALIGN_SUPPORTED
#pragma options align=reset
#endif

// gradient fill class:


class	ZGradient	: public ZComrade
{
protected:
	GradientType			grType;								// gradient style
	short					grCount;							// how many colours
	RGBColor				grColours[ kMaxGradientColours ];	// the colours
	short					grOffsets[ kMaxGradientColours ];	// positions
	short					grAngle;							// angle for directional ones
	ZGWorld*				grTemp;								// temporary offscreen world
	RgnHandle				destRgn;							// the region we're filling

public:
	
	ZGradient();
	~ZGradient();

// drawing calls:
	
	virtual void			GradientFillRegionInPort( RgnHandle aRgn, Point centrePt );
	virtual void			GradientFillRectInPort( Rect* aFrame, Point centrePt );

// getters and setters:
	
	inline	void			SetGradientType( const GradientType aType ) { grType = aType; };
	inline	GradientType	GetGradientType() { return grType; };
	inline 	void			SetGradCount( const short aCount ) { grCount = aCount; };
	
	virtual	void			SetGradientColour( const short index, RGBColor* aColour );
	virtual void			SetGradientOffset( const short index, const short offset );
	
	virtual PicHandle		GetGradientThumbnail();
	
	inline	void			GetGradientColour( RGBColor* aColour, const short index ) { *aColour = grColours[ index ]; };
	inline	short			GetGradientAngle() { return grAngle; };
	inline	short			GetGradientCount() { return grCount; };

// for storing gradient settings externally:
	
	virtual GradientRecHdl	GetGradHandle();
	virtual void			SetGradHandle( GradientRecHdl aHandle );
	
protected:
	
	virtual void			InitGradient();	
	virtual void			BuildGradientWorld( Rect* aFrame, Point centrePt );
	
	virtual void			CalcColours( long gradRange, RGBColor* c1, RGBColor* c2, CGradInfo* grInfo );	
	virtual void			IncrementColour( RGBColor* c, CGradInfo* grInfo );
	virtual long			CalcGradRange( Rect* r, short index, Point centre );
	virtual void			CalcEnclosingSquare( Rect* r, Point centre );
	virtual short			CalcDirectionalScan( Rect* aFrame, Point* start, Point* end );	
};

// n.b. in this version the gradient offsets are ignored- the gradient is distributed evenly
// across the fill area. Variable offsets is a possible future enhancement.


typedef ZObjectArray<ZGradient>		ZGradientList;





#endif