/*************************************************************************************************
*
*
*			ObjectMacZapp			-- a standard Mac OOP application template
*
*
*
*			ZMenuBarPane.cpp		-- a menubar in a window
*
*
*
*
*
*			 1996, Graham Cox
*
*
*
*
*************************************************************************************************/


#include	"ZMenuBarPane.h"
#include	"RPaneWindow.h"
#include	"MacZoop.h"
#include	<Palettes.h>

static RgnHandle	gCurrMBarRgn = NULL;


/*--------------------------------***  CONSTRUCTOR  ***---------------------------------*/

ZMenuBarPane::ZMenuBarPane( ZCommander* aBoss, const short mBarResID, const Boolean rounded )
	: RPane( aBoss ), ZMenuBar( mBarResID )
{
	mbFont = menuFont = geneva;
	mbSize = menuFontSize = 9;
	mbStyle = bold;
	roundCorners = rounded;
	useMTitleArrows = TRUE;
	menuCheckChar = '';	
	paneActive = FALSE;	
	curTHiliteIndex = 0;
	
	// make sure that if the window we're part of is resized, we stretch
	// to fit across
	
	alignLeft = TRUE;
	alignRight = TRUE;
}

/*---------------------------------***  DESTRUCTOR  ***---------------------------------*/

ZMenuBarPane::~ZMenuBarPane()
{
	// we need to dispose the menus that belong to this, since
	// the system won't do it, and since this pane may come and go we
	// don't want a memory leak.
	
	short		i;
	MenuHandle	mH;
	
	for ( i = 1; i <= mbCount; i++ )
	{
		mH = GetIndexedMenuHandle( i );	
	
		if ( mH )
			DisposeMenuAndSubMenus( mH );
	}
	
	// get rid of our MBAR handle too.
	
	if ( mBarH )
		DisposeHandle((Handle) mBarH );
}


/*--------------------------------***  INITMENUBAR  ***---------------------------------*/
/*	
MUST be called after construction- this sets up the pane and the menus
----------------------------------------------------------------------------------------*/

void		ZMenuBarPane::InitMenuBar()
{
	// by default, the pane's frame is the same width as its boss, aligned to the boss's
	// top edge, and a fixed height- just like a menubar should be!
	
	RPane*	myBoss = dynamic_cast<RPane*>( itsBoss );		// Is boss a pane?
	
	if ( myBoss )
	{
		Rect	bf;
		
		myBoss->GetContentRect( &bf );
	
		SetSize( bf.right - bf.left, MB_STD_MBAR_HEIGHT );
		PlaceAt( bf.left, bf.top );
	}

	// unlike ZMenuBar, we need to keep the MBAR resource handle around permanently so that
	// we can look up the menuID from the menu index number. Otherwise the behaviour of the
	// menu loading is largely similar, with the exception that menus are not inserted into
	// the system menu list until they are popped up.
	
	FailNIL( theMenuCmds = new ZArray( sizeof( MenuCmd )));
	
	mbCount = 0;
	
	// load the MBAR resource
	
	mBarH = (short**) GetResource( 'MBAR', mBarID );
	FailOSErr( ResError());
	
	HNoPurge((Handle) mBarH );
	DetachResource((Handle) mBarH );
	
	// read in menus from the MBAR resource, but do NOT
	// automatically install them into the system list just yet
	
	LoadMenus( FALSE );
}


/*--------------------------------***  SETMBARFONT  ***---------------------------------*/
/*	
sets the font, size and style of the menu title text. Does not affect the menus.
----------------------------------------------------------------------------------------*/

void		ZMenuBarPane::SetMBarFont( const short aFont, const short aSize, const short aStyle )
{
	if ( aFont != mbFont	||
		 aSize != mbSize	||
		 aStyle != mbStyle )
	{
		mbFont = aFont;
		mbSize = aSize;
		mbStyle = aStyle;
		
		Focus();
		Refresh();
	}
}


/*--------------------------------***  SETMENUFONT  ***---------------------------------*/
/*	
sets the font, size and checkmark to use for the menus. The style of menu items is
given by the original menu resource, or can be changed with normal toolbox calls
----------------------------------------------------------------------------------------*/

void		ZMenuBarPane::SetMenuFont( const short aFont, const short aSize, const char aCheckChar )
{
	menuFont = aFont;
	menuFontSize = aSize;
	menuCheckChar = aCheckChar;
}


/*-------------------------------***  SETTITLEHILITE  ***-------------------------------*/
/*	
draws the title of the indexed menu in the given state.
----------------------------------------------------------------------------------------*/

void		ZMenuBarPane::SetTitleHilite( const short menuID, const Boolean state )
{
	// the <menuID> passed here is the index of the menu- that is, the <nth> menu at the top
	// level of the bar. Since we need to know what the true menu ID is, this class keeps the
	// original MBAR handle around so that we can look it up.
	
	short	mID = menuID;
	
	if ( menuID == 0 )
	{
		mID = curTHiliteIndex;
		curTHiliteIndex = 0;
	}
	
	TextFont( mbFont );
	TextSize( mbSize );
	TextFace( mbStyle );

	MenuHandle mH = GetIndexedMenuHandle( mID );
	
	if ( mH )
	{
		// given the handle, we can obtain the title string, which is what we're going
		// to draw in the bar. We also need the title rect so we know where to position it.
		
		Rect			tr;
		Str255			mtStr;
		
		GetTitleRect( mID, &tr );
		
		// copy the title locally
		
		CopyPString((*mH)->menuData, mtStr );
		
		// if title is hilited, we paint it in the hilite colour for menu titles
		// first, then draw the text in the contrasting colour. Otherwise we paint in
		// background colour and draw the text in the text colour
		
		SetUpMBarColours();
		
		if ( state )
		{
			if ( ! paneActive)
				PenPat( &qd.gray );
			
			PaintRect( &tr );
			TextMode( srcBic );
			PenNormal();
		}
		else
		{
			EraseRect( &tr );
			
			#ifdef MB_FULL_COLOUR_MBAR
			
			// if we are using the full 3D effect bar, we need to put back the shadow line
			// we just obliterated
			
			RGBColor	saveFore, f, b;
			
			GetForeColor( &saveFore );
			GetBackColor( &b );
			f = saveFore;
			GetGray( GetMainDevice(), &b, &f );
	
			RGBForeColor( &f );
			MoveTo( tr.left, itsFrame.bottom - 2 );
			LineTo( tr.right, itsFrame.bottom - 2 );
			
			RGBForeColor( &saveFore );
			
			#endif
		
			if ( paneActive )
				TextMode( srcOr );
			else
				TextMode( grayishTextOr );
		}
		
		MoveTo( tr.left + MB_MENU_TITLE_EXTRA, tr.bottom - 4 );
		DrawString( mtStr );
		
		// if we want arrows as well, plot the little buggers! Note that
		// if this is not a colour mac, this feature is unavailable.
		
		if ( useMTitleArrows && gMacInfo.supportsColour )
		{
			CIconHandle		arrow;
			
			arrow = GetCIcon( MB_ARROW_ICON_RES_ID );
			
			if ( arrow )
			{
				Rect	ar = (*arrow)->iconMask.bounds;
				
				OffsetRect( &ar,
							tr.right - MB_ARROW_WIDTH_EXTRA - ar.left - 5,
							tr.bottom - 10 - ar.top );
				
				PlotCIconInCurrentTextMode( &ar, arrow );
				DisposeCIcon( arrow );
			}	
		}
		
		ForeColor( blackColor );
		BackColor( whiteColor );
	}
}


/*---------------------------***  GETINDEXEDMENUHANDLE  ***-----------------------------*/
/*	
returns the mac menu handle associated with the indexed menu in this bar
----------------------------------------------------------------------------------------*/

MenuHandle	ZMenuBarPane::GetIndexedMenuHandle( const short index )
{
	FailNILParam( mBarH );
	
	// to obtain the menuHandle, given that the menu is not in the system list, we
	// need to look for our menu ID in the list of commands we have. That record
	// contains the menu handle
	
	MenuCmd		mCmd;
	short		id;
	long		mSelect;
	
	// menuID is obtained by look-up in the mBarH list
	
	mCmd.macMenu = NULL;
	
	if ( index > 0 )
	{
		id = (*mBarH)[ index ];
		
		// iterate through commands until we match the ID
		
		mSelect = ((long) id << 16) + 1;
		
		FindMCmd( mSelect, &mCmd );
		
		// if found, return the handle, otherwise NULL
	}	
	
	return mCmd.macMenu;
}


/*--------------------------------***  DRAWCONTENT  ***---------------------------------*/
/*	
draws the menubar when called upon to do so
----------------------------------------------------------------------------------------*/

void		ZMenuBarPane::DrawContent()
{
	Rect		rr;
	
	PenNormal();
	
	// if we're in colour and there's a special colour for the bar, paint it first
	
	SetUpMBarColours();
	
	rr = itsFrame;
	rr.bottom--;
	EraseRect( &rr );
	
	// if we have round corners, form the corner region and paint them
	
	if ( roundCorners )
	{
		#ifndef MB_FULL_COLOUR_MBAR
		
		RgnHandle	temp, rrRgn;
		
		rr = itsFrame;
		rr.bottom += rr.bottom - rr.top;
		
		OpenRgn();
		FrameRoundRect( &rr, MB_ROUND_RECT_RADIUS, MB_ROUND_RECT_RADIUS );
		CloseRgn( temp = NewRgn());
		
		RectRgn( rrRgn = NewRgn(), &itsFrame );
		DiffRgn( rrRgn, temp, rrRgn );
		
		PaintRgn( rrRgn );
		
		DisposeRgn( rrRgn );
		DisposeRgn( temp );
		
		#else
		
		// for 3D effect Mbar, we actually plot colour icons for the corners that have a
		// 3D effect built in- this is easier than trying to calculate a decent effect on the
		// fly.
		
		CIconHandle		ciH;
		
		ciH = GetCIcon( MB_LCORNER_ICON_RES_ID );
		
		if ( ciH )
		{
			rr = (*ciH)->iconMask.bounds;
			OffsetRect( &rr, -rr.left, -rr.top );
			
			PlotCIcon( &rr, ciH );
			DisposeCIcon( ciH );
		}
		
		ciH = GetCIcon( MB_RCORNER_ICON_RES_ID );
		
		if ( ciH )
		{
			rr = (*ciH)->iconMask.bounds;
			OffsetRect( &rr, itsFrame.right - rr.left - ( rr.right - rr.left ), -rr.top );
			
			PlotCIcon( &rr, ciH );
			DisposeCIcon( ciH );
		}
		
		
		#endif
	}
	
	// draw the menu titles in the bar in the correct font, etc.
	
	if ( mbCount > 0 )
	{
		// for each menu entry, obtain the title and draw it.
		
		short		i;

		for ( i = 1; i <= mbCount; i++ )
			SetTitleHilite( i, ( i == curTHiliteIndex ));
	}
	
	TextFace( 0 );
	TextMode( srcOr );
	
	// draw a single line below the bar
	
	ForeColor( blackColor );
	BackColor( whiteColor );
	
	MoveTo( 0, itsFrame.bottom - 1 );
	LineTo( itsFrame.right, itsFrame.bottom - 1 );
	
	// if we want the 3D effect, draw the shadow line at the bottom, just above the black line
	
	#ifdef MB_FULL_COLOUR_MBAR
	
	// we need a colour half way between fore and back to do the shadow:
	
	RGBColor	f, b;
	
	SetUpMBarColours();
	
	GetForeColor( &f );
	GetBackColor( &b );
	GetGray( GetMainDevice(), &b, &f );
	
	RGBForeColor( &f );
	MoveTo( 0, itsFrame.bottom - 2 );
	LineTo( itsFrame.right - 1, itsFrame.bottom - 2 );
	Line( 0, -6 );
	
	// likewise we put a light coloured highlight across the top of the bar:
	
	f.red = f.green = f.blue = 0xEEEE;
	RGBForeColor( &f );
	MoveTo( 7 , itsFrame.top );
	LineTo( itsFrame.right - 7 , itsFrame.top );
	
	ForeColor( blackColor );
	BackColor( whiteColor );
	
	#endif
}


/*----------------------------------***  ACTIVATE  ***----------------------------------*/
/*	
activate the menubar if this option is compiled in
----------------------------------------------------------------------------------------*/

void		ZMenuBarPane::Activate()
{
 	RPane::Activate();
	paneActive = TRUE;
	
	#ifdef MB_AUTO_DIMMING
	Focus();
	Refresh();
	#endif
}
 
 
/*---------------------------------***  DEACTIVATE  ***---------------------------------*/
/*	
deactivate the menubar if this option is compiled in
----------------------------------------------------------------------------------------*/

void		ZMenuBarPane::Deactivate()
{
	RPane::Deactivate();
	
	#ifdef MB_AUTO_DIMMING
	paneActive = FALSE;
	Refresh();
	#endif
}


/*------------------------------------***  CLICK  ***-----------------------------------*/
/*	
handle mouse-down in the menubar, tracking menu titles, pulling down menus, etc. When an
item is chosen from a menu, the command is dispatched back to this object, which by default
passes it to its boss.
----------------------------------------------------------------------------------------*/

void		ZMenuBarPane::Click( const Point mouse, const short modifiers )
{
	long			mSelect;
	ZCommander*		hSave;
	
	DimMenus();
	UpdateMenus();
	
	TextFont( mbFont );
	TextSize( mbSize );
	TextFace( mbStyle );
	TextMode( srcOr );
	
	mSelect = TrackMenuBar( mouse );
	
	// as a quick solution to sending the command to ourselves, we simply
	// set gCurHandler to this and restore it afterwards.
	
	hSave = gCurHandler;
	gCurHandler = this;
	
	DispatchCommand( mSelect );
	
	gCurHandler = hSave;
	
	TextFace( 0 );
}


/*----------------------------------***  DIMMENUS  ***----------------------------------*/
/*	
dim all menus prior to tracking the bar
----------------------------------------------------------------------------------------*/

void		ZMenuBarPane::DimMenus()
{
	// dim our menus. Instead of walking the system list as our base class does, since the menus
	// are not there, we simply index through them ourselves
	
	short		i;
	MenuHandle	mH;
	
	for ( i = 1; i <= mbCount; i++ )
	{
		mH = GetIndexedMenuHandle( i );
		
		if ( mH )
			PredimMenu( mH );
	}
}


/*--------------------------------***  TRACKMENUBAR  ***--------------------------------*/
/*	
track the mouse in the menubar, returning the selected menu, if any.
----------------------------------------------------------------------------------------*/

long		ZMenuBarPane::TrackMenuBar( const Point mouse )
{
	long 	mSelect = 0;
	short	i, lasti = 0;
	Rect	tr = {0,0,0,0};
	Point	lastMouse = {-1, -1};
	Point	curMouse;
	
	while( WaitMouseUp())
	{
		// where is the mouse?
		
		GetMouse( &curMouse );
		
		// has it moved at all?
		
		if ( DeltaPoint( curMouse, lastMouse ))
		{
			// has it moved outside the title rect?
			
			if ( ! PtInRect( curMouse, &tr ))
			{	
				// yes, so find out which title it's gone to
				
				i = GetPtInIndex( curMouse, &tr );
				
				// is that a new index?
				
				if ( i != lasti )
				{
					// we've moved into a new title, so hilite it and track its menu.
					
					if ( lasti )
						SetTitleHilite( lasti, FALSE );
					
					if ( i )
					{
						SetTitleHilite( i, TRUE );
						mSelect = TrackMenu( curMouse, i, &tr );
					}
					else
						SetRect( &tr, 0, 0, 0, 0 );
							
					lasti = i;
				}
			}
			lastMouse = curMouse;
		}
	}
	
	if ( mSelect == 0 )
	{
		curTHiliteIndex = 0;
		SetTitleHilite( lasti, FALSE );
	}
	else
		curTHiliteIndex = lasti;
	
	return mSelect;
}


/*---------------------------------***  TRACKMENU  ***----------------------------------*/
/*	
pull down and track the menu indexed. This installs regions and hook procs and sets up
the font for the menu as necessary
----------------------------------------------------------------------------------------*/

long		ZMenuBarPane::TrackMenu( const Point mouse, const short mIndex, Rect* titleRect )
{
	long			mSelect = 0;
	MenuHandle		mH;
	Point			popPt, fOffset = { 0, 0 };
	MenuHookUPP		mhUPP,oldMHook;	
	Rect			tr;
	
	// insert, pop-up and track the menu indexed.
	
	mH = GetIndexedMenuHandle( mIndex );
	
	if ( mH )
	{
		RgnHandle	temp;
		
		// insert the menu into the system list temporarily
		
		PreInsertSubMenus( mH );
		
		// to make sure that the bar works properly, we need to set up a MenuHook proc
		// that will abort the menu tracking when the mouse moves into another menu's title. We
		// also set up a global region representing the menubar LESS our own title, so we can
		// find where the mouse is in the hook proc
		
		tr = *titleRect;
		tr.bottom += 2;
		
		gCurrMBarRgn = NewRgn();
		RectRgn( gCurrMBarRgn, &itsFrame );
		RectRgn( temp = NewRgn(), &tr );
		DiffRgn( gCurrMBarRgn, temp, gCurrMBarRgn );
		DisposeRgn( temp );
		
		LocalToGlobal( &fOffset );
		OffsetRgn( gCurrMBarRgn, fOffset.h, fOffset.v );
		
		// install hook proc
		
		oldMHook = LMGetMenuHook();
		mhUPP = NewMenuHookProc( MHookProc );
		LMSetMenuHook( mhUPP );
		
		// the menu itself is going to be drawn in a font we dictate, not the system
		// font. To make this work, we need to correctly manipulate the low memory
		// globals that control this
		
		SetUpSystemFont( menuFont, menuFontSize );
		ResetSystemFont();
		
		// calculate the point to pop it up
		
		popPt.h = titleRect->left + 1;
		popPt.v = titleRect->bottom + 1;
		LocalToGlobal( &popPt );
		
		// do all the tricky toolbox stuff:
		
		mSelect = PopUpMenuSelect( mH, popPt.v, popPt.h, 1 );
		
		// reset the system font
		
		SetUpSystemFont( 0, 0 );
		ResetSystemFont();
		
		// remove menu from system list again
		
		PostDeleteSubMenus( mH );
		
		// clean up:
		
		LMSetMenuHook( oldMHook );
		DisposeRoutineDescriptor( mhUPP );
		
		DisposeRgn( gCurrMBarRgn );
		gCurrMBarRgn = NULL;
	}
	
	return mSelect;
}


/*--------------------------------***  GETPTININDEX  ***--------------------------------*/
/*	
returns the index and title rect of the menu title under the mouse, if any
----------------------------------------------------------------------------------------*/

short		ZMenuBarPane::GetPtInIndex( const Point where, Rect* titleRect )
{
	// return the index of the menu containing the mouse point. This also returns the title
	// rect so we avoid having to calculate this twice.
	
	short	i, j = 0;
	Rect	tr;
	
	for ( i = 1; i <= mbCount; i++ )
	{
		GetTitleRect( i, &tr );
	
		if ( PtInRect( where, &tr ))
		{
			*titleRect = tr;
			j = i;
			break;
		}
	}
	
	return j;	
}


/*--------------------------------***  GETTITLERECT  ***--------------------------------*/
/*	
calculates the title rect of the indexed menu. This is done dynamically to titles
can be changed more or less on the fly if you really want to have an ugly interface!
----------------------------------------------------------------------------------------*/

void		ZMenuBarPane::GetTitleRect( const short index, Rect* theRect )
{
	// return the rect of the menu title corresponding to to the indexed menu. Menus are drawn
	// from left to right in index order.
	
	Rect		r;
	short		rightEdge = MB_FIRST_TITLE_OFFSET;
	short		i;
	MenuHandle	mH;
	
	r = itsFrame;
	r.top = r.bottom - MB_STD_MBAR_HEIGHT + 1;
	r.bottom--;

	// we need to iterate through the menus until we reach the index of interest
	// so we can keep a running position of the edges of the titles.
	
	for ( i = 1; i <= index; i++ )
	{
		// get the menu handle for the one we're looking at
	
		r.left = rightEdge;
		mH = GetIndexedMenuHandle( i );
		
		if ( mH )
		{
			// title rect width is the width of the string plus some extra, plus
			// space for the arrows if switched on
			
			rightEdge += ( MB_MENU_TITLE_EXTRA * 2 ) + StringWidth((*mH)->menuData);
			if ( useMTitleArrows )
				rightEdge += MB_ARROW_WIDTH_EXTRA;
			r.right = rightEdge;
		}
		else
		{
			rightEdge += 16;
			r.right = rightEdge;
		}
	}
	
	*theRect = r;
}


/*------------------------------***  SETUPMBARCOLOURS  ***------------------------------*/
/*	
set the fore and back colours of the port to those indicated in the system mctb resource
----------------------------------------------------------------------------------------*/

void		ZMenuBarPane::SetUpMBarColours()
{
	ForeColor( blackColor );
	BackColor( whiteColor );
	
	if ( gMacInfo.supportsColour )
	{
		MCEntryPtr		mceP;
		RGBColor		fColour, bColour;

		mceP = GetMCEntry( 0, 0 );
		
		if ( mceP )
		{
			fColour = mceP->mctRGB1;
			bColour = mceP->mctRGB4;
			
			RGBForeColor( &fColour );
			RGBBackColor( &bColour );
		}
	}
}



void	ZMenuBarPane::PreInsertSubMenus( MenuHandle parentMenu )
{
	// scans through the menu looking for submenus. For any it finds, it inserts them
	// into the system list.
	
	InsertMenu( parentMenu, hierMenu );
	
	short	m, i, cmd, subID;
	
	m = CountMenuItems( parentMenu );
	
	for ( i = 1; i <= m; i++ )
	{
		GetItemCmd( parentMenu, i, &cmd );
	
		if ( cmd == hMenuCmd )
		{
			GetItemMark( parentMenu, i, &subID );
		
			MenuHandle	subMenu = FindMenuID( subID );
			
			if ( subMenu )
				PreInsertSubMenus( subMenu );
		}
	}
}



void	ZMenuBarPane::PostDeleteSubMenus( MenuHandle parentMenu )
{
	// undoes the above by deleting the submenus attached to the menu from the system list
	
	DeleteMenu((*parentMenu)->menuID );
	
	short	m, i, cmd, subID;
	
	m = CountMenuItems( parentMenu );
	
	for ( i = 1; i <= m; i++ )
	{
		GetItemCmd( parentMenu, i, &cmd );
	
		if ( cmd == hMenuCmd )
		{
			GetItemMark( parentMenu, i, &subID );
		
			MenuHandle	subMenu = FindMenuID( subID );
			
			if ( subMenu )
				PostDeleteSubMenus( subMenu );
		}
	}

}


/*----------------------------***  DISPOSEMENUANDSUBMENUS  ***--------------------------*/
/*	
disposes of the menu passed and any submenus it refers to. Called only by the destructor.
----------------------------------------------------------------------------------------*/

void	ZMenuBarPane::DisposeMenuAndSubMenus( MenuHandle parentMenu )
{
	short	i, m, cmd, subID;
	
	m = CountMenuItems( parentMenu );
	
	for( i = 1; i <= m; i++ )
	{
		GetItemCmd( parentMenu, i, &cmd );
		
		if ( cmd == hMenuCmd )
		{
			GetItemMark( parentMenu, i, &subID );
	
			MenuHandle	subMenu = FindMenuID( subID );
			
			if ( subMenu )
				DisposeMenuAndSubMenus( subMenu );
		}
	}
	
	DisposeMenu( parentMenu );
}



#pragma mark -

/*-------------------------------***  SETUPSYSTEMFONT  ***------------------------------*/
/*	
set the system font to the font and size passed. USE WITH EXTREME CARE!
----------------------------------------------------------------------------------------*/

void SetUpSystemFont( short font, short size )
{
	if ( LMGetSysFontFam() != font || LMGetSysFontSize() != size )
	{
		LMSetSysFontFam( font );
		LMSetSysFontSize( size );
		LMSetLastSPExtra( -1L );		// Signal to flush font cache
	}
}


/*-------------------------------***  RESETSYSTEMFONT  ***------------------------------*/
/*	
reset the system font to the default value
----------------------------------------------------------------------------------------*/

void ResetSystemFont()
{
	GrafPtr	savePort;
	GrafPtr wMgrPort;

	GetWMgrPort( &wMgrPort );
	GetPort( &savePort );
	SetPort( wMgrPort );
	TextFont( 0 );
	TextSize( 0 );
	SetPort( savePort );
}


/*-----------------------***  PLOTCICONINCURRENTTEXTMODE  ***---------------------------*/
/*	
plot the B+W part of a colour icon using the current text mode as the transfer mode. This
allows the arrow icon to be drawn as if it were a text character.
----------------------------------------------------------------------------------------*/

void PlotCIconInCurrentTextMode( Rect* aRect, CIconHandle iconH )
{
	// plots the icon in aRect in the current port's transfer mode.
	
	GrafPtr	thePort;
	short	mode;
	Rect	srcRect;
	
	GetPort( &thePort );
	mode = thePort->txMode;
	srcRect = (*iconH)->iconPMap.bounds;
	
	// if the text mode is grayishTextOr, we need to modify things a bit so
	// that the icon is drawn in the correct colour.
	
	if ( mode == grayishTextOr )
	{
		RGBColor	saveFore, back, newFore;
		
		GetForeColor( &saveFore );
		GetBackColor( &back );
		newFore = saveFore;
		
		GetGray( GetMainDevice(), &back, &newFore );
	
		RGBForeColor( &newFore );
		mode = srcOr;
	}
	
	HLock((Handle) iconH );
	
	(*iconH)->iconBMap.baseAddr = (Ptr) &(*iconH)->iconMaskData;
	CopyBits(&(*iconH)->iconBMap, &thePort->portBits, &srcRect, aRect, mode, NULL );
	HUnlock((Handle) iconH );
	
	
}


#pragma mark -


/*---------------------------------***  MHOOKPROC  ***----------------------------------*/
/*	
in order to correctly switch to a new menu when inside the toolbox modal PopUpMenuSelect
call, we install this proc which checks that the mouse is not in another part of the
menubar, and if it is, aborts the menu tracking by posting a mouse-up event. This is the
documented technique for achieving this behaviour.
----------------------------------------------------------------------------------------*/

pascal void	MHookProc()
{
	// this is the MenuHook function we install to make sure that when the mouse moves into
	// another part of the menubar, PopUpMenuSelect is abandoned.

	Point	where;
	
	GetMouse( &where );
	if ( PtInRgn( where, gCurrMBarRgn ))
	{
		// menu is to be abandoned, so we post a mouse-up event to make this
		// happen. Hopefully, this method will not break in the future (finger's crossed!)
		
		(void) PostEvent( mouseUp, 0 );
	}
}





