/*************************************************************************************************
*
*
*			ObjectMacZapp		-- a standard Mac OOP application template
*
*
*
*			ZMenuBarPane.h		-- a menubar in a window
*
*
*
*
*
*			 1996, Graham Cox
*
*
*
*
*************************************************************************************************/


#pragma once

#ifndef __ZMENUBARPANE__
#define	__ZMENUBARPANE__

#include	"ZMenuBar.h"
#include	"RPane.h"

// class definition:


class	ZMenuBarPane	: public RPane, public ZMenuBar
{
protected:
	short		mbFont;					// font to use for menu titles
	short		mbSize;					// font size of titles
	short		mbStyle;				// font style of titles
	short		menuFont;				// font to use for menus
	short		menuFontSize;			// font size to use for menus
	short		curTHiliteIndex;		// current hilited title index
	Boolean		roundCorners;			// TRUE if bar is drawn with rounded corners
	Boolean		useMTitleArrows;		// TRUE if titles have hint arrows
	Boolean		paneActive;				// TRUE if the bar is activated
	
	
public:

	ZMenuBarPane( ZCommander* aBoss, const short mBarResID, const Boolean rounded = TRUE );
	~ZMenuBarPane();

	virtual void		InitMenuBar();
	virtual void		SetMBarFont( const short aFont, const short aSize, const short aStyle);
	virtual void		SetMenuFont( const short aFont, const short aSize, const char aCheckChar = '');
	virtual void		SetTitleHilite( const short menuID, const Boolean state );
	virtual void		DimMenus();

	virtual void		DrawContent();
	virtual void		Click( const Point mouse, const short modifiers );
	
	virtual void		Activate();
	virtual void		Deactivate();

protected:
	
	virtual long		TrackMenuBar( const Point mouse );
	virtual long		TrackMenu( const Point mouse, const short mIndex, Rect* titleRect );
	virtual void		SetUpMBarColours();

private:
	
	short				GetPtInIndex( const Point where, Rect* titleRect );
	void				GetTitleRect( const short index, Rect* theRect );
	MenuHandle			GetIndexedMenuHandle( const short index );
	void				PreInsertSubMenus( MenuHandle parentMenu );
	void				PostDeleteSubMenus( MenuHandle parentMenu );
	void				DisposeMenuAndSubMenus( MenuHandle parentMenu );
};

// menu hook function:

static pascal void	MHookProc( void );

// global functions for manipulating low memory font variables:

void SetUpSystemFont( short font, short size );
void ResetSystemFont();
void PlotCIconInCurrentTextMode( Rect* aRect, CIconHandle iconH );

// some dimensions of the mbar layout are hard-coded using these defines:

#define		MB_ROUND_RECT_RADIUS	16
#define		MB_MENU_TITLE_EXTRA		5
#define		MB_FIRST_TITLE_OFFSET	8
#define		MB_STD_MBAR_HEIGHT		16
#define		MB_ARROW_WIDTH_EXTRA	10

// some features of this class require icon resources- this defines their ID numbers

#define		MB_ARROW_ICON_RES_ID	200
#define		MB_LCORNER_ICON_RES_ID	201
#define		MB_RCORNER_ICON_RES_ID	202

// if you want a full-colour, copland style menubar, define the follwoing. Note that even
// if this is not defined, the menubar will pick up the colours set for the system's menus.
// This define gives the bar an extra subtle 3D appearance.

#define	MB_FULL_COLOUR_MBAR

// if you want the menubar to dim on deactivation, define this:

#define		MB_AUTO_DIMMING

/*

This class allows you to have a menubar in any window. It is based on a combination of ZMenuBar
and RPane, so requires RPane and related classes which are available from Uli Kusterer.

This interface element is extremely useful if used appropriately, and is one item the Mac lacks
as part of its standard UI toolbox. Also one of the few Windows features that is actually nice
to have. Unlike the Windows version though, this looks good without taking up excessive space
in the window!

This works just like the main menu bar in every respect, and dispatches commands to itself when
menu items are selected. If you don't override it, these commands will end up at the window class.


*/


#endif