/*************************************************************************************************
*
*
*			ObjectMacZapp			-- a standard Mac OOP application template
*
*
*
*			ZPaintEngine.cpp		-- an object for implementing paint functions in a window
*
*
*
*
*
*			 1996, Graham Cox
*
*
*
*
*************************************************************************************************/


#include	"ZPaintEngine.h"
#include	"ZUndoIPTask.h"
#include	"ZUndoPaintSelectionTask.h"
#include	"ZGradient.h"
#include	"ZGWorldWindow.h"
#include	"MacZoop.h"
#include	"ZMouserWindow.h"

#include	<TextEdit.h>
#include	<fp.h>

// static func prototypes:

static short 			abs( short a );
static pascal Boolean 	WandSearchProc( RGBColor* rgb, long* position );
static pascal Boolean	MakeMaskSearchProc( RGBColor* rgb, long* position );
static pascal void 		PEDrawHookProc( unsigned short textOffset, unsigned short drawLen, void *textBufferPtr, TEPtr pTE, TEHandle hTE );

// gradient globals (extern):

extern short			gCurGradIndex;
extern	ZGradientList*	gGradientList;
extern	ZGradient*		gCurGradient;
extern	ZPaintEngine*	gPaintEngine;

// draw hook globals

DrawHookUPP		gPEHook = NewDrawHookProc( PEDrawHookProc );

/*------------------------------***   CONSTRUCTOR   ***---------------------------------*/

ZPaintEngine::ZPaintEngine( ZGWorld* anArtBuffer )
	: ZComrade()
{
	FailNILParam( anArtBuffer );
	
	artBuffer = anArtBuffer;
	ownsArtBuffer = FALSE;
	
	tempBuffer = NULL;
	windowPort = NULL;
	curPTask = NULL;
	floatBuffer = NULL;
	FailNIL( theSelection = NewRgn());
	antsActive = FALSE;
	hAlign = vAlign = 1;
	
	FailNIL( paintMask = NewRgn());
	paintMaskActive = FALSE;
	paintMaskVisible = FALSE;
	
	artBuffer->GetSize( &artFrame );
	SetDefaults();
}

/*------------------------------***   CONSTRUCTOR   ***---------------------------------*/

ZPaintEngine::ZPaintEngine( const Rect& artworkSize )
	: ZComrade()
{
	FailNIL( artBuffer = new ZGWorld( artworkSize, 8 ));
	ownsArtBuffer = TRUE;
	
	tempBuffer = NULL;
	windowPort = NULL;
	floatBuffer = NULL;
	FailNIL( theSelection = NewRgn());
	
	FailNIL( paintMask = NewRgn());
	paintMaskActive = FALSE;
	paintMaskVisible = FALSE;

	antsActive = FALSE;
	hAlign = vAlign = 1;
	artFrame = artworkSize;
	SetDefaults();
}

/*------------------------------***   CONSTRUCTOR   ***---------------------------------*/

ZPaintEngine::ZPaintEngine()
	: ZComrade()
{
	artBuffer = NULL;
	ownsArtBuffer = FALSE;
	tempBuffer = NULL;
	windowPort = NULL;
	curPTask = NULL;
	floatBuffer = NULL;
	FailNIL( theSelection = NewRgn());
	
	FailNIL( paintMask = NewRgn());
	paintMaskActive = FALSE;
	paintMaskVisible = FALSE;

	antsActive = FALSE;
	hAlign = vAlign = 1;
	
	SetRect( &artFrame, 0, 0, 0, 0 );
	SetDefaults();
}


/*-------------------------------***   DESTRUCTOR   ***---------------------------------*/

ZPaintEngine::~ZPaintEngine()
{
	KillFloatingSelection( FALSE );

	if ( ownsArtBuffer && artBuffer )
		ForgetObject( artBuffer );
		
	DisposeRgn( theSelection );
	DisposeRgn( paintMask );
	
	if ( inTextMode )
		EndText();
		
	if ( curGradient )
		ForgetObject( curGradient );
}


/*------------------------------***   SETARTBUFFER   ***--------------------------------*/
/*
set a new art buffer. Caller's responsibility to manage or delete old one.	
----------------------------------------------------------------------------------------*/

void		ZPaintEngine::SetArtBuffer( ZGWorld* aBuf, Boolean ownsBuf )
{
	artBuffer = aBuf;
	ownsArtBuffer = ownsBuf;
	
	if ( artBuffer )
		artBuffer->GetSize( &artFrame );
}

/*------------------------------***   SETDEFAULTS   ***---------------------------------*/
/*	
set all paint parameters to their default values. Also resets the selection tools.
----------------------------------------------------------------------------------------*/

void		ZPaintEngine::SetDefaults()
{
	currentTool = selectionTool;
	antiAlias = TRUE;
	drawFromCentre = FALSE;
	tweenEnable = TRUE;
	frameWidth = frameHeight = 1;
	paintColour.red = paintColour.green = paintColour.blue = 0;
	lineColour = paintColour;
	eraseColour.red = eraseColour.green = eraseColour.blue = 0xFFFF;
	fillPattern = qd.black;
	framePattern = qd.black;
	cFillPattern = NULL;
	cFramePattern = NULL;
	currentFill = plainColourFill;
	frameBkTransparent = FALSE;
	fillBkTransparent = FALSE;
	paintBkTransparent = FALSE;
	txFillTransparent = FALSE;
	currentPaint = colourPaint;
	currentBrush = ovalBrush;
	brushWidth = 8;
	brushHeight = 8;
	currentAirMode = stdAirMode;
	airRadius = 16;
	airPressure = 50;
	textShadowWidth = textShadowHeight = 1;
	textFont = geneva;
	textSize = 12;
	textStyle = 0;
	textFill = plainColourFill;
	textAlign = teFlushLeft;
	textBorderWidth = 1;
	regularPolySides = 6;
	roundRectHCorner = 16;
	roundRectVCorner = 16;
	curGradient = NULL;
	
	// no initial selection
	
	SetEmptyRgn( theSelection );
	currentSelType = rectSelection;
	currentSelMod = replaceSelection;
	
	// 45 striped pattern
	
	StuffHex( &ants, "\pF0E1C3870F1E3C78" );
	
	fbHOffset = fbVOffset = 0;
	matchTolerance = 4096;
	
	// not in text editing mode
	
	currTE = NULL;
	inTextMode = FALSE;
	SetRect( &textRect, 0, 0, 0, 0 );
	
	// initialise the brush bitmap
	
	Rect	r = { 0, 0, 10, 10 };
	MakeBitMap( &r, &brushBits );
	
	// clone anchor point not yet set
	
	cloneAnchor.h = cloneAnchor.v = -1;
	cloneHOffset = 0;
	cloneVOffset = 0;
	hadFirstCloneClick = FALSE;
}


/*----------------------------*** SETSTATEFROMHANDLE ***--------------------------------*/
/*	
set all of the parameters from the state handle. This could have come from e.g. the prefs
file for the application that is using this object.
----------------------------------------------------------------------------------------*/

void		ZPaintEngine::SetStateFromHandle( PEStateHdl p )
{
	FailNILParam( p );
	
	antiAlias 			= (*p)->aaLines;
	antiAliasText 		= (*p)->aaText;
	tweenEnable			= (*p)->tween;
	frameBkTransparent	= (*p)->frTsp;
	fillBkTransparent	= (*p)->fillTsp;
	paintBkTransparent	= (*p)->paintTsp;
	matchTolerance		= (*p)->slipTolerance;
	frameWidth			= (*p)->frWidth;
	frameHeight			= (*p)->frHeight;
	//(*p)->cPatFillID		= 1;
	//(*p)->cPatFrameID		= 1;
	brushWidth			= (*p)->brWidth;
	brushHeight			= (*p)->brHeight;
	airRadius			= (*p)->airRadius;
	airPressure			= (*p)->airPressure;
	textShadowWidth		= (*p)->txShadowWidth;
	textShadowHeight	= (*p)->txShadowHeight;
	textFont			= (*p)->txFont;
	textSize			= (*p)->txSize;
	textStyle			= (*p)->txStyle;
	textBorderWidth		= (*p)->txBorder;
	regularPolySides	= (*p)->rPolySides;
	roundRectHCorner	= (*p)->rRectHCorner;
	roundRectVCorner	= (*p)->rRectVCorner;
	paintColour			= (*p)->paint;
	lineColour			= (*p)->lines;
	eraseColour			= (*p)->erase;
	currentSelType		= (*p)->curSelType;
	currentTool			= (*p)->curTool;
	currentFill			= (*p)->curFill;
	currentPaint		= (*p)->curPaint;
	currentBrush		= (*p)->curBrush;
	currentAirMode		= (*p)->curAirMode;
	textFill			= (*p)->txFill;
	fillPattern			= (*p)->pFill;
	framePattern		= (*p)->pFrame;
	
	// set up the gradient:
	
	gCurGradIndex = (*p)->curGradID;
	
	// install this gradient as initial one, if exists
	
	long	n = gGradientList->CountItems();
	
	if ( n )
	{
		gCurGradIndex = MAX( MIN( n, gCurGradIndex ), 1 );
		
		curGradient = gCurGradient = gGradientList->GetObject( gCurGradIndex );
	}
}


/*-------------------------------*** GETSTATEHANDLE ***---------------------------------*/
/*	
return the current state of the paint engine in a handle for saving externally, e.g. in
the prefs file.
----------------------------------------------------------------------------------------*/

PEStateHdl	ZPaintEngine::GetStateHandle()
{
	PEStateHdl	p;
	
	FailNIL( p = ( PEStateHdl ) NewHandleClear( sizeof( PEStateRec )));
	
	(*p)->sVersNum			= 0x0100;
	(*p)->aaLines			= antiAlias;
	(*p)->aaText			= antiAliasText;
	(*p)->tween				= tweenEnable;
	(*p)->frTsp				= frameBkTransparent;
	(*p)->fillTsp			= fillBkTransparent;
	(*p)->paintTsp			= paintBkTransparent;
	(*p)->slipTolerance		= matchTolerance;
	(*p)->frWidth			= frameWidth;
	(*p)->frHeight			= frameHeight;
	(*p)->cPatFillID		= 1;
	(*p)->cPatFrameID		= 1;
	(*p)->brWidth			= brushWidth;
	(*p)->brHeight			= brushHeight;
	(*p)->airRadius			= airRadius;
	(*p)->airPressure		= airPressure;
	(*p)->txShadowWidth		= textShadowWidth;
	(*p)->txShadowHeight	= textShadowHeight;
	(*p)->txFont			= textFont;
	(*p)->txSize			= textSize;
	(*p)->txStyle			= textStyle;
	(*p)->txBorder			= textBorderWidth;
	(*p)->rPolySides		= regularPolySides;
	(*p)->rRectHCorner		= roundRectHCorner;
	(*p)->rRectVCorner		= roundRectVCorner;
	(*p)->curGradID			= gCurGradIndex;
	(*p)->paint				= paintColour;
	(*p)->lines				= lineColour;
	(*p)->erase				= eraseColour;
	(*p)->curSelType		= currentSelType;
	(*p)->curTool			= currentTool;
	(*p)->curFill			= currentFill;
	(*p)->curPaint			= currentPaint;
	(*p)->curBrush			= currentBrush;
	(*p)->curAirMode		= currentAirMode;
	(*p)->txFill			= textFill;
	(*p)->pFill				= fillPattern;
	(*p)->pFrame			= framePattern;

	return p;
}


/*-------------------------------*** SETCURRENTTOOL ***---------------------------------*/
/*	
set the paint tool to use for subsequent drawing
----------------------------------------------------------------------------------------*/

void		ZPaintEngine::SetCurrentTool( const PaintTool aTool )
{
	currentTool = aTool;
	
	if ( inTextMode && ( currentTool != textTool ))
		EndText();
		
	// reset the clone anchor
	
	cloneAnchor.h = cloneAnchor.v = -1;
	cloneHOffset = cloneVOffset = 0;
	hadFirstCloneClick = FALSE;
	
	ResetAlertStage();
}


/*------------------------------*** SETSELECTIONTYPE ***--------------------------------*/
/*	
set the type of selection the selection tool implements
----------------------------------------------------------------------------------------*/

void		ZPaintEngine::SetSelectionType( const SelectionType aType )
{
	currentSelType = aType;
}


/*-------------------------------***  SETARTBOUNDS ***----------------------------------*/
/*	
set the art rect to blit into in the destination port. This is a rect that represents the
artwork size at the displayed scale- allowing for zoom and scroll, etc. This also computes
the hAlign and vAlign members that affect the way point calculations are done- this ensures
that tools work with great accuracy at zoomed sizes.
----------------------------------------------------------------------------------------*/

void		ZPaintEngine::SetArtBounds( const Rect& aRect, const short scaling )
{
	artBounds = aRect;
	
	if ( scaling > 100 )
		hAlign = vAlign = ( scaling / 100 );
	else
		hAlign = vAlign = 1;
}


/*------------------------------***  SETPAINTBRUSH  ***---------------------------------*/
/*	
set the type of paint brush- affects the shape and drawing methods
----------------------------------------------------------------------------------------*/

void		ZPaintEngine::SetPaintBrush( const BrushType aType )
{
	currentBrush = aType;	
}


/*--------------------------------*** SETPAINTTYPE ***----------------------------------*/
/*	
set the type of paint to paint with- colour, patterned, etc.
----------------------------------------------------------------------------------------*/

void		ZPaintEngine::SetPaintType( const PaintType ptType )
{
	currentPaint = ptType;
}


/*-----------------------------*** SETPAINTBRUSHSIZE ***--------------------------------*/
/*	
set the size of the brush to paint with.
----------------------------------------------------------------------------------------*/

void		ZPaintEngine::SetPaintBrushSize( const short aWidth, const short aHeight )
{
	brushWidth = aWidth;
	brushHeight = aHeight;
}


/*------------------------------*** SETPAINTCOLOUR  ***---------------------------------*/
/*	
set the colour of the paint
----------------------------------------------------------------------------------------*/

void		ZPaintEngine::SetPaintColour( const RGBColor& aColour )
{
	paintColour = aColour;
}


/*------------------------------***  SETFRAMECOLOUR ***---------------------------------*/
/*	
set the colour for frames and lines
----------------------------------------------------------------------------------------*/

void		ZPaintEngine::SetFrameColour( const RGBColor& aColour )
{
	lineColour = aColour;
}


/*-----------------------------***  SETERASECOLOUR  ***---------------------------------*/
/*	
set the colour of the background when erased, etc
----------------------------------------------------------------------------------------*/

void		ZPaintEngine::SetEraseColour( const RGBColor& aColour )
{
	eraseColour = aColour;
}


/*-----------------------------***  SETPAINTPATTERN  ***--------------------------------*/
/*	
set the pattern to use for painting
----------------------------------------------------------------------------------------*/

void		ZPaintEngine::SetPaintPattern( const Pattern& aPat )
{
	SetPaintCPattern( NULL );
	fillPattern = aPat;
}

/*-----------------------------***  SETPAINTPATTERN  ***--------------------------------*/

void		ZPaintEngine::SetPaintPattern( const short patListID, const short patID )
{
	SetPaintCPattern( NULL );
	GetIndPattern( &fillPattern, patListID, patID );
}


/*------------------------------*** SETFRAMEPATTERN ***---------------------------------*/
/*	

----------------------------------------------------------------------------------------*/

void		ZPaintEngine::SetFramePattern( const Pattern& aPat )
{
	SetFrameCPattern( NULL );
	framePattern = aPat;
}

/*------------------------------*** SETFRAMEPATTERN ***---------------------------------*/

void		ZPaintEngine::SetFramePattern( const short patListID, const short patID )
{
	SetFrameCPattern( NULL );
	GetIndPattern( &framePattern, patListID, patID );
}


/*-----------------------------***  SETPAINTCPATTERN ***--------------------------------*/
/*	
set the colour pattern to use for painting
----------------------------------------------------------------------------------------*/

void		ZPaintEngine::SetPaintCPattern( PixPatHandle pp )
{
	if ( cFillPattern )
		DisposePixPat( cFillPattern );
	
	cFillPattern = pp;
}


/*-----------------------------***  SETPAINTCPATTERN ***--------------------------------*/
/*	
set the colour pattern to use for frames and lines
----------------------------------------------------------------------------------------*/

void		ZPaintEngine::SetFrameCPattern( PixPatHandle pp )
{
	if ( cFramePattern )
		DisposePixPat( cFramePattern );
	
	cFramePattern = pp;
}


/*-------------------------------***  SETFILLTYPE  ***----------------------------------*/
/*	
set the type of fill for the bucket tool
----------------------------------------------------------------------------------------*/

void		ZPaintEngine::SetFillType( const FillType aType )
{
	if ( aType != currentFill )
	{
		currentFill = aType;
		
		SendMessage( peMsgFillStyleChanged, &currentFill );
	}
}


/*-----------------------------***  SETGRADIENTFILL  ***--------------------------------*/
/*	
set the current gradient style for gradient fills. This is supplied as an object that
can both store and generate gradients
----------------------------------------------------------------------------------------*/

void		ZPaintEngine::SetGradientFill( ZGradient* aGradient )
{
	curGradient = aGradient;
	
	SendMessage( peMsgFillStyleChanged, &currentFill );
}


/*------------------------------*** SETAIRBRUSHTYPE  ***--------------------------------*/
/*	
set the type of spray for the airbrush tool
----------------------------------------------------------------------------------------*/

void		ZPaintEngine::SetAirbrushType( const AirMode abType )
{
	currentAirMode = abType;
}


/*---------------------------------***  SETAIR   ***------------------------------------*/
/*	
set spray paramaters radius and pressure
----------------------------------------------------------------------------------------*/

void		ZPaintEngine::SetAir( const short aRadius, const short aPressure )
{
	airRadius = aRadius;
	airPressure = aPressure;
}


/*-------------------------------*** SETTEXTPARAMS ***----------------------------------*/
/*	

----------------------------------------------------------------------------------------*/

void		ZPaintEngine::SetTextParams( const short aFont, const short aSize, const short aStyle, const short align )
{
	textFont = aFont;
	textSize = aSize;
	textStyle = aStyle;
	textAlign = align;
	
	// if currently editing text, apply the new settings immediately
	
	if ( inTextMode && currTE )
	{
		CGrafPtr	savePort;
		GDHandle	saveDev;
		
		tempBuffer->Lock();
		tempBuffer->SetPortToGW( &savePort, &saveDev );
		
		(*currTE)->txFont = textFont;
		(*currTE)->txSize = textSize;
		(*currTE)->txFace = textStyle;
		
		TESetAlignment( textAlign , currTE );
		
		FontInfo	fi;
		Rect		r;
	
		TextFont( textFont );
		TextSize( textSize );
		TextFace( textStyle );
		
		GetFontInfo( &fi );
		
		(*currTE)->lineHeight = fi.ascent + fi.descent + fi.leading + textShadowHeight;
		(*currTE)->fontAscent = fi.ascent;
		
		r = textRect;
		InsetRect( &r, 2, 2);
		
		//TECalText( currTE );
		
		tempBuffer->Clear();
		TEUpdate( &r, currTE );
		
		tempBuffer->Unlock();
		SetGWorld( savePort, saveDev );
		InvalRect( &textRect );
	}
	
	if ( currentTool == textTool )
		SendMessage( peMsgTextParamsChanged, NULL );
}


/*-------------------------------*** SETTEXTSHADOW ***----------------------------------*/
/*	

----------------------------------------------------------------------------------------*/

void		ZPaintEngine::SetTextShadow( const short shadowH, const short shadowV )
{
	textShadowWidth = shadowH;
	textShadowHeight = shadowV;
	
	if ( inTextMode )
	{
		SetPort( windowPort );
		InvalRect( &textRect );
	}
}


/*----------------------------*** SETTEXTBORDERFRAME  ***-------------------------------*/
/*

----------------------------------------------------------------------------------------*/

void		ZPaintEngine::SetTextBorderFrame( const short bfWidth )
{
	textBorderWidth = bfWidth;
	
	if ( inTextMode )
	{
		SetPort( windowPort );
		InvalRect( &textRect );
	}
}

/*----------------------------***  SETREGULARPOLYSIDES ***------------------------------*/
/*	

----------------------------------------------------------------------------------------*/

void		ZPaintEngine::SetRegularPolySides( const short sides )
{
	regularPolySides = sides;
}

#pragma mark -
/*---------------------------------*** MOUSECLICK ***-----------------------------------*/
/*	
this is the main call. This applies the current paint tool to the artwork, tracking the
mouse as needed, and updating the display window as needed. At the end of the call, an
undo task is generated and the artwork modified.
----------------------------------------------------------------------------------------*/

void		ZPaintEngine::MouseClick( Point mClickPt, const short modifiers )
{
	curPTask = NULL;
	
	// set up data members for common modifiers such as constrain and draw from centre
	
	constrain = ( modifiers & shiftKey ) == shiftKey;
	drawFromCentre = ( modifiers & controlKey ) == controlKey;
	
	TranslatePointToOffscreen( &mClickPt );
	
	if ( inTextMode )
		TextClick( mClickPt, modifiers );
	else
	{
		try
		{
			if (( currentTool != selectionTool ) &&
				( modifiers & optionKey) == optionKey )
			{
				// the option key means we want to pick up the colour under the mouse for use with
				// the current tool. If the clone tool is in use, we set the anchor point of the
				// tool for subsequent cloning ops.
				
				if ( currentTool == stampTool )
				{
					cloneAnchor = mClickPt;
					cloneHOffset = 0;
					cloneVOffset = 0;
					hadFirstCloneClick = FALSE;
				}
				else
					PickUpPaintColour( mClickPt, modifiers );	
			}
			else
			{
				switch( currentTool )
				{
					case selectionTool:
						DragSelection( mClickPt, modifiers );
						break;
					case paintBrushTool:
					case eraserTool:
					case pencilTool:
						PaintTrail( mClickPt );
						break;
					case fillTool:
						ClickPaint( mClickPt );
						break;
					case stampTool:
					case airbrushTool:
						AirbrushTrail( mClickPt );
						break;
					case rectFrameTool:
					case rectSolidTool:
					case roundRectFrameTool:
					case roundRectSolidTool:
					case ovalFrameTool:
					case ovalSolidTool:
					case arcFrameTool:
					case arcSolidTool:
						DragElasticShape( mClickPt );
						break;
					case lineTool:
						DragLineOrPoly( mClickPt, FALSE );
						break;
					case regularPolyFrameTool:
					case regularPolySolidTool:
						DragLineOrPoly( mClickPt, TRUE );
						break;
					case irregPolyFrameTool:
					case irregPolySolidTool:
						break;
					case textTool:
						TextClick( mClickPt, modifiers );
						break;
					case colourPickTool:
						PickUpPaintColour( mClickPt, modifiers );
						break;
				}
			}
		}
		catch( OSErr err )
		{
			// if an exception was throwm the reason is most probably lack of memory to make
			// the edition undoable. We make sure any existing undo task is released since
			// we've now gone to the next step in any case.
			
			if (( err == userCanceledErr ) && windowPort )
			{
				ZWindow*	zw = GetZWindow( windowPort );
				zw->SetTask( NULL );
			}
			throw err;
		}
	
		// pass the undo task to the window. It's possible that the task is legitimately NULL,
		// so we do not bother to validate it.
		
		if ( ! (( currentTool == selectionTool ) && ( curPTask == NULL )))
		{
			if ( windowPort )
			{	
				ZWindow*	zw = GetZWindow( windowPort );
				
				zw->SetTask( curPTask );
			}
		}
	}
}

#pragma mark -

/*-----------------------------*** SETUPSECONDBUFFER  ***-------------------------------*/
/*	
some painting requires double-buffering techniques. This makes the buffer.
----------------------------------------------------------------------------------------*/

void		ZPaintEngine::SetUpSecondBuffer( Rect* aRect, Boolean autoClear )
{
	FailNIL( tempBuffer = new ZGWorld( artBuffer ));
	
	if ( autoClear )
		tempBuffer->Clear();
		
	CGrafPtr	savePort;
	GDHandle	saveDev;
	
	tempBuffer->SetPortToGW( &savePort, &saveDev );
	PrepareMaskClipForPort();
	
	SetGWorld( savePort, saveDev );
}

/*-----------------------------*** REMOVESECONDBUFFER ***-------------------------------*/
/*	
deletes the temporary buffer
----------------------------------------------------------------------------------------*/

void		ZPaintEngine::RemoveSecondBuffer()
{
	if ( tempBuffer )
		ForgetObject( tempBuffer );
		
	tempBuffer = NULL;
}


/*---------------------------------*** MAKEBITMAP ***-----------------------------------*/
/*	
make a bitmap with the bounds rect passed. This is a true 1-bit offscreen classic bitmap
which is needed for some particular QD operations
----------------------------------------------------------------------------------------*/

void		ZPaintEngine::MakeBitMap( Rect* aRect, BitMap* aBMap )
{
	short	width, height;
	long	bufSize;
	
	aBMap->bounds = *aRect;
	
	width = aRect->right - aRect->left;
	height = aRect->bottom - aRect->top;
	
	aBMap->rowBytes = ((width + 15) / 16) * 2;
	
	bufSize = (long) aBMap->rowBytes * (long) height;
	
	FailNIL( aBMap->baseAddr = NewPtrClear( bufSize ));
}


/*-------------------------------*** DISPOSEBITMAP ***----------------------------------*/
/*	
releases memory used by bitmap made above
----------------------------------------------------------------------------------------*/

void		ZPaintEngine::DisposeBitMap( BitMap* aBMap )
{
	DisposePtr( aBMap->baseAddr );
}


/*------------------------------*** DRAGELASTICSHAPE ***--------------------------------*/
/*	
lower level method to draw "elastic" shapes. This is a whole class of shapes such as rects,
ovals, roundrects and  regular polys, both framed and filled.
----------------------------------------------------------------------------------------*/

void		ZPaintEngine::DragElasticShape( Point mouseStart )
{
	// elastic drawing uses a 2nd buffer if we have the memory to do so- this gives nice
	// smooth drawing. If there is not enough memory, we blit directly, and incur some
	// flicker.
	
	Point		lastMouse, anchor, anchor2;
	Rect		lastRect, shapeRect, uRect, mURect;
	CGrafPtr	savePort;
	GDHandle	saveDev;
	
	// make a second buffer for doing smooth dragging
	
	SetUpSecondBuffer( &artFrame );
	
	// set up initial tracking variables
	
	Pt2Rect( mouseStart, mouseStart, &lastRect );
	lastMouse = anchor = anchor2 = mouseStart;
	
	ReportMouseLoc( anchor, mouseStart, TRUE );
	
	// track as long as the mouse is down
	
	while( WaitMouseUp())
	{
		GetMouse( &mouseStart );
		TranslatePointToOffscreen( &mouseStart );
		
		// if constraining, make sure width and height are equal
		
		if ( constrain )
		{
			long mouseDiff = DeltaPoint( mouseStart, anchor2 );
		
			if ( HiWord( mouseDiff ) != LoWord( mouseDiff ))
				mouseStart.v = anchor2.v + mouseStart.h - anchor2.h;
		}
		
		// if drawing from centre, modify anchor to be negative offset
		
		if ( drawFromCentre )
		{
			anchor.h = anchor2.h - ( mouseStart.h - anchor2.h );
			anchor.v = anchor2.v - ( mouseStart.v - anchor2.v );
		}
		
		// has the mouse actually moved since last time? If not, nothing to do
		
		if ( DeltaPoint( mouseStart, lastMouse ))
		{
			// the rect of the shape is between the anchor and the mouse
			
			Pt2Rect( anchor, mouseStart, &shapeRect );
			
			// compute the union of the last rect and the new one
			
			UnionRect( &shapeRect, &lastRect, &uRect );
			mURect = uRect;
			MapRect( &mURect, &artFrame, &artBounds );
			
			// blit the artwork to the window where this rect is, either from
			// our second buffer or directly from the original. If no second buffer,
			// dragging shapes is fairly flickery, but does work OK.
			
			tempBuffer->SetPortToGW( &savePort, &saveDev );
			tempBuffer->SetPortColours();
			
			// the clip region of the temp buffer may include the mask clip. If so
			// we need to temporarily disable it or the drawing won't be right.
			
			RgnHandle	saveClip;
				
			if ( paintMaskActive )
			{
				GetClip( saveClip = NewRgn());
				ClipRect( &uRect );
			}
			
			tempBuffer->CopyIn( artBuffer, &uRect, &uRect );
			
			if ( paintMaskActive )
			{
				SetClip( saveClip );
				DisposeRgn( saveClip );
			}
			
			if ( currentTool == arcFrameTool ||
				 currentTool == arcSolidTool )
				DrawArcInPort( anchor, mouseStart ); 
			else
				DrawShapeInPort( &shapeRect );
			
			SetGWorld( savePort, saveDev );
			tempBuffer->CopyOut( &windowPort->portBits, &uRect, &mURect, ditherCopy ); 
			
			// keep track of last mouse location and shape
			
			lastRect = shapeRect;
			lastMouse = mouseStart;
			
			ReportMouseLoc( anchor, mouseStart, TRUE );
		}
	}
	
	// mouse has been released, so we need to apply the shape drawn to the artwork. If we have
	// a second buffer, we can merely copy it back in. If not, we have to drawthe shape to the
	// artwork directly.
	
	try
	{
		curPTask = MakeUndoTask();
	}
	catch( OSErr err )
	{
		RemoveSecondBuffer();
		throw err;
	}
	
	artBuffer->CopyIn( tempBuffer, &uRect, &uRect );
	RemoveSecondBuffer();
}


/*-------------------------------*** DRAWARCINPORT ***----------------------------------*/
/*
draw an arc from anchor to location using current paint settings. This takes account of
the positional ordering of the points to draw the arc intuitively.	
----------------------------------------------------------------------------------------*/

void		ZPaintEngine::DrawArcInPort( Point anchor, Point location )
{
	Rect	arcRect;
	short	arcAngle;
	short	startAngle;
	
	if ( location.h >= anchor.h )
	{
		arcAngle = 90;
		arcRect.right = location.h;	
		arcRect.left = anchor.h - ( location.h - anchor.h );
	}
	else
	{
		arcAngle = -90;
		arcRect.left = location.h;
		arcRect.right = anchor.h + ( anchor.h - location.h );
	}	
	
	if ( location.v >= anchor.v )
	{
		startAngle = 0;
		arcRect.top = anchor.v;
		arcRect.bottom = location.v + ( location.v - anchor.v );	
	}
	else
	{
		if ( location.h >= anchor.h )
			startAngle = 90;
		else
			startAngle = -90;
		arcRect.bottom = anchor.v;
		arcRect.top = location.v - ( anchor.v - location.v );
	}
	
	// OK, all set up so draw the arc:
	
	PenSize( frameWidth, frameHeight );
	if ( fillBkTransparent )
		PenMode( patOr );
	else
		PenMode( patCopy );
		
	RGBBackColor( &eraseColour );
	
	switch ( currentTool )
	{
		case arcSolidTool:
			RGBForeColor( &paintColour );
			SetPortPenPattern( kFill );
			PaintArc( &arcRect, startAngle, arcAngle );
		case arcFrameTool:
			RGBForeColor( &lineColour );
			SetPortPenPattern( kFrame );
			if ( frameBkTransparent )
				PenMode( patOr );
			else
				PenMode( patCopy );
			FrameArc( &arcRect, startAngle, arcAngle );
			break;
	}
	
	PenNormal();
	BackColor( whiteColor );
	
}


/*------------------------------*** DRAWSHAPEINPORT ***---------------------------------*/
/*	
draw the currently called for "elastic" shape to fit the given rect in the current port.
----------------------------------------------------------------------------------------*/

void		ZPaintEngine::DrawShapeInPort( Rect* aRect )
{
	Boolean		isFilledShape = FALSE;
	
	PenSize( frameWidth, frameHeight );
	if ( fillBkTransparent )
		PenMode( patOr );
	else
		PenMode( patCopy );
	RGBBackColor( &eraseColour );
	
	switch ( currentTool )
	{
		case rectSolidTool:
			RGBForeColor( &paintColour );
			SetPortPenPattern( kFill );
			PaintRect( aRect );
		case rectFrameTool:
			RGBForeColor( &lineColour );
			SetPortPenPattern( kFrame );
			if ( frameBkTransparent )
				PenMode( patOr );
			else
				PenMode( patCopy );
			FrameRect( aRect );
			break;
		case ovalSolidTool:
			RGBForeColor( &paintColour );
			SetPortPenPattern( kFill );
			PaintOval( aRect );
		case ovalFrameTool:
			RGBForeColor( &lineColour );
			SetPortPenPattern( kFrame );
			if ( frameBkTransparent )
				PenMode( patOr );
			else
				PenMode( patCopy );
			FrameOval( aRect );
			break;
		case roundRectSolidTool:
			RGBForeColor( &paintColour );
			SetPortPenPattern( kFill );
			PaintRoundRect( aRect, roundRectHCorner, roundRectVCorner );
		case roundRectFrameTool:
			RGBForeColor( &lineColour );
			SetPortPenPattern( kFrame );
			if ( frameBkTransparent )
				PenMode( patOr );
			else
				PenMode( patCopy );
			FrameRoundRect( aRect, roundRectHCorner, roundRectVCorner );
			break;
	}
	
	PenNormal();
	BackColor( whiteColor );
}



void		ZPaintEngine::DrawShapeInPort( RgnHandle aRgn )
{
	// as above, but works with arbitrary region shape
	
	Boolean		isFilledShape = FALSE;
	
	PenSize( frameWidth, frameHeight );
	if ( fillBkTransparent )
		PenMode( patOr );
	else
		PenMode( patCopy );
	RGBBackColor( &eraseColour );
	
	switch ( currentTool )
	{
		case regularPolySolidTool:
			RGBForeColor( &paintColour );
			SetPortPenPattern( kFill );
			PaintRgn( aRgn );
		case regularPolyFrameTool:
			RGBForeColor( &lineColour );
			SetPortPenPattern( kFrame );
			if ( frameBkTransparent )
				PenMode( patOr );
			else
				PenMode( patCopy );
			FrameRgn( aRgn );
			break;
	}
}


/*-------------------------------***  DRAGLINEORPOLY ***--------------------------------*/
/*	
draws lines or regular polygons using similar technique to DragShape. Lines and polys
are actually fairly similar, so we can make the one routine do both according to
<currentTool>
----------------------------------------------------------------------------------------*/

void		ZPaintEngine::DragLineOrPoly( Point mouseStart, Boolean isPoly )
{
	Point		lastMouse, anchor;
	Rect		shapeRect, lastShapeRect, uRect, mURect;
	CGrafPtr	savePort;
	GDHandle	saveDev;
	short		constrainV = -1;
	RgnHandle	polyRgn = NULL;

	lastMouse = anchor = mouseStart;
	Pt2Rect( mouseStart, mouseStart, &lastShapeRect );
	
	// make offscreen buffer
	
	SetUpSecondBuffer( &artFrame );
	ReportMouseLoc( mouseStart, mouseStart, TRUE );

	// track mouse
	
	while( WaitMouseUp())
	{
		GetMouse( &mouseStart );
		TranslatePointToOffscreen( &mouseStart );
		
		// if constraining, make sure width and height are equal
		
		if ( constrain )
		{
			long mouseDiff = DeltaPoint( mouseStart, anchor );
			
			if ( constrainV == -1 )
			{
				if ( abs( HiWord( mouseDiff )) > abs( LoWord( mouseDiff )))
					constrainV = 1;	// v
				else
					constrainV = 0;
			}
			
			if ( constrainV )
				mouseStart.h = anchor.h;
			else
				mouseStart.v = anchor.v;
		}
		
		if ( DeltaPoint( mouseStart, lastMouse ))
		{
			// mouse moved, draw the line
			// the rect enclosing the line is between the anchor and the mouse
			
			if ( isPoly )
			{
				polyRgn = MakeRegularPoly( anchor, mouseStart );
				shapeRect = (*polyRgn)->rgnBBox;
			}
			else
			{
				Pt2Rect( anchor, mouseStart, &shapeRect );
				shapeRect.right += frameWidth;
				shapeRect.bottom += frameHeight;
			}
			// compute the union of the last rect and the new one
			
			UnionRect( &shapeRect, &lastShapeRect, &uRect );
			mURect = uRect;
			MapRect( &mURect, &artFrame, &artBounds );
			
			
			tempBuffer->SetPortToGW( &savePort, &saveDev );
			tempBuffer->SetPortColours();
			tempBuffer->CopyIn( artBuffer, &uRect, &uRect );	
			
			if ( isPoly )
				DrawShapeInPort( polyRgn );
			else
				DrawLineInPort( mouseStart, anchor );
			
			SetGWorld( savePort, saveDev );
			ForeColor( blackColor );
			BackColor( whiteColor );
			tempBuffer->CopyOut( &windowPort->portBits, &uRect, &mURect, ditherCopy ); 
			
			lastMouse = mouseStart;
			lastShapeRect = shapeRect;
			
			if ( isPoly && polyRgn )
				DisposeRgn( polyRgn );
				
			ReportMouseLoc( anchor, mouseStart, TRUE );
		}
	}
	
	// make undo task:
	
	try
	{
		curPTask = MakeUndoTask();
	}
	catch( OSErr err )
	{
		RemoveSecondBuffer();
		throw err;
	}
	
	// make change permanent
	
	artBuffer->CopyIn( tempBuffer, &uRect, &uRect );
	RemoveSecondBuffer();
}


/*------------------------------***  DRAWLINEINPORT ***---------------------------------*/
/*
draws a line from a to b in the current port, applying current frame parameters	
----------------------------------------------------------------------------------------*/

void		ZPaintEngine::DrawLineInPort( Point a, Point b )
{
	PenSize( frameWidth, frameHeight );
	
	if ( frameBkTransparent )
		PenMode( patOr );
	else
		PenMode( patCopy );
		
	SetPortPenPattern( kFrame );
	RGBForeColor( &lineColour );
	RGBBackColor( &eraseColour );
	
	MoveTo( a.h, a.v );
	LineTo( b.h, b.v );
	
	PenNormal();
	ForeColor( blackColor );
	BackColor( whiteColor );
}


/*--------------------------------***  PAINTTRAIL ***-----------------------------------*/
/*	
lower level method to draw "trail" paint. This is a whole class of painting including pencil,
airbrush, paintbrush and erase tool.
----------------------------------------------------------------------------------------*/

void		ZPaintEngine::PaintTrail( Point mouseStart )
{
	Point		lastMouse, anchor;
	short		constrainV = -1;
	CGrafPtr	savePort;
	GDHandle	saveDev;
	Rect		uRect, mURect, pbRect, lastPBRect;
	RGBColor	pencilColour;
	
	SetUpSecondBuffer( &artFrame, FALSE );
	
	// were going to draw directly into the second buffer, so set it as the current port
	
	tempBuffer->SetPortToGW( &savePort, &saveDev );
	anchor = lastMouse = mouseStart;
	
	// this operation does not use an offscreen buffer to record the change, but records the
	// sequence of calls into a QuickDraw picture instead. This may be changed later.
	
	switch ( currentTool )
	{
		default:
			RGBForeColor( &paintColour );
			RGBBackColor( &eraseColour );
			SetPortPenPattern( kFill );
			break;
		case eraserTool:
			RGBForeColor( &eraseColour );
			break;
		case pencilTool:
			GetClickedColour( mouseStart, &pencilColour );
			
			if ( EqualColour( &pencilColour, &paintColour ))
				RGBForeColor( &eraseColour );
			else
				RGBForeColor( &paintColour );
			break;
	}
	
	// draw the initial point so that if the mouse is clicked and not dragged, you get at
	// least 1 "blob"
	
	DrawPaint( lastMouse );
	
	Pt2Rect( lastMouse, lastMouse, &uRect );
	
	if ( currentTool == eraserTool )
		InsetRect( &uRect, -kEraserToolHalfWidth, -kEraserToolHalfWidth );
	else
		InsetRect( &uRect, -( brushWidth >> 1), -( brushHeight >> 1 ));
	mURect = uRect;
	MapRect( &mURect, &artFrame, &artBounds );
	
	SetGWorld( savePort, saveDev );
	
	ForeColor( blackColor );
	BackColor( whiteColor );
	tempBuffer->CopyOut( &windowPort->portBits, &uRect, &mURect );
	
	pbRect= lastPBRect = uRect;
	
	ReportMouseLoc( anchor, mouseStart, TRUE );

	// track the mouse and trail the paint
	
	while ( WaitMouseUp())
	{
		GetMouse( &mouseStart );
		TranslatePointToOffscreen( &mouseStart );
		
		if ( DeltaPoint( mouseStart, lastMouse ))
		{
			// mouse moved, so tween the brush if enabled, else just draw the new "blob"
			// at the current mouse location.
			
			// apply contraint, if any. This is a vertical or horizontal constraint
			// depnding on the direction we moved first time
			
			if ( constrain )
			{
				if ( constrainV == -1 )
				{
					// compute which way we are going to constrain in future
					
					long	mDiff = DeltaPoint( mouseStart, lastMouse );
					
					if ( abs( HiWord( mDiff )) > abs( LoWord( mDiff )))
						constrainV = 1;		// vertical
					else
						constrainV = 0;		// horizontal			
				}

				if ( constrainV )
					mouseStart.h = lastMouse.h;
				else
					mouseStart.v = lastMouse.v;
			}
	
			tempBuffer->SetPortToGW( &savePort, &saveDev );
			
			// tween the brush if enabled. Note that eraser or pencil tool set, paint
			// is always tweened regardless of setting
			
			if ( tweenEnable 					||
				( currentTool == eraserTool ) 	||
				( currentTool == pencilTool ))
				TweenPaint( mouseStart, lastMouse );
			else
				DrawPaint( mouseStart );
			
			Pt2Rect( mouseStart, mouseStart, &pbRect );
			
			if ( currentTool == eraserTool )
				InsetRect( &pbRect, -kEraserToolHalfWidth, -kEraserToolHalfWidth );
			else
				InsetRect( &pbRect, -( brushWidth >> 1), -( brushHeight >> 1 ));
			UnionRect( &pbRect, &lastPBRect, &uRect );
			
			mURect = uRect;
			MapRect( &mURect, &artFrame, &artBounds );
			
			SetGWorld( savePort, saveDev );
			
			ForeColor( blackColor );
			BackColor( whiteColor );
			tempBuffer->CopyOut( &windowPort->portBits, &uRect, &mURect, ditherCopy );

			lastMouse = mouseStart;
			lastPBRect = pbRect;
			
			ReportMouseLoc( anchor, mouseStart, TRUE );
		}
	}
	
	// mouse released, clean up and make the change permanent.
	
	PenNormal();
	ForeColor( blackColor );
	
	try
	{
		curPTask = MakeUndoTask();
	}
	catch( OSErr err )
	{
		RemoveSecondBuffer();
		SetGWorld( savePort, saveDev );
		throw err;
	}	
	// draw the recorded strokes into the art buffer
	
	SetGWorld( savePort, saveDev );
	artBuffer->SetPortToGW( &savePort, &saveDev );	
	artBuffer->CopyIn( tempBuffer, &artFrame, &artFrame );
	SetGWorld( savePort, saveDev );
	
	RemoveSecondBuffer();
}


/*-------------------------------*** AIRBRUSHTRAIL  ***---------------------------------*/
/*
similar to paint trail, but continously calls the draw method so that paint "builds up"
over time, as you'd expect from an airbrush.	
----------------------------------------------------------------------------------------*/

void		ZPaintEngine::AirbrushTrail( Point mouseStart )
{
	Rect		shape, mURect, artSource;
	CGrafPtr	savePort;
	GDHandle	saveDev;
	Boolean		usingAirPort = FALSE;
	Point		anchor = mouseStart;
	
	// for non-splatter airbrush, we use yet another offscreen port to provide a mask for
	// the spray effect.
	
	if ( currentAirMode == stdAirMode	||
		 currentAirMode == darkenArt	||
		 currentAirMode == lightenArt 	||
		 currentTool == stampTool )
	{
		// set up clone tool offsets
		
		if ( currentTool == stampTool ) 
		{
			if ( ! hadFirstCloneClick )
			{
				// if the clone anchor not set, abort...
				
				if ( cloneAnchor.h == -1 &&
					 cloneAnchor.v == -1 )
				{
					(void) Alert( 203, NULL );
					return;
				}
				
				cloneHOffset = anchor.h - cloneAnchor.h;
				cloneVOffset = anchor.v - cloneAnchor.v;
				hadFirstCloneClick = TRUE;
			}
			else
			{
				cloneAnchor.h = anchor.h - cloneHOffset;
				cloneAnchor.v = anchor.v - cloneVOffset;
			}
		}

		// 2nd buffer has full copy of art to give
		// a good, smooth airbrush effect. Colour blending at other depths gives disappointing
		// results.
		
		FailNIL( tempBuffer = new ZGWorld( artFrame, artBuffer->GetDepth()));
		tempBuffer->CopyIn( artBuffer, &artFrame, &artFrame );
		
		// set mask
		
		tempBuffer->SetPortToGW( &savePort, &saveDev );
		PrepareMaskClipForPort();
		SetGWorld( savePort, saveDev );
	
		usingAirPort = TRUE;
		SetRect( &shape, 0, 0, airRadius * 2, airRadius * 2 );
		FailNIL( airGW = new ZGWorld( shape, 32 ));
		artSource = shape;
		
		airGW->Clear;
		
		// paint a circle into the port, feathered towards the edges
		
		airGW->SetPortToGW( &savePort, &saveDev );
		airGW->Lock();
		
		RGBColor	feather = {-1,-1,-1}; 	// start at white
		short		i;
		
		i = airRadius;
		
		do
		{
			RGBForeColor( &feather );
			PaintOval( &shape );
			InsetRect( &shape, 1, 1 );
			
			// gradually feather the paint towards black from white. The greater the pressure,
			// the more black the mask becomes.
			
			feather.green = feather.blue = feather.red -=  ( 655L * airPressure ) / (long) airRadius;
		}
		while( --i );
		
		airGW->Unlock();
		SetGWorld( savePort, saveDev );
	}
	else
		SetUpSecondBuffer( &artFrame, FALSE );
		
	ReportMouseLoc( anchor, mouseStart, TRUE );
	
	// track the mouse. For the airbrush, we just keep applying the paint, even if the mouse
	// never moves. This allows the paint to "build up" in a more realistic fashion.
		
	while ( WaitMouseUp())
	{
		GetMouse( &mouseStart );
		TranslatePointToOffscreen( &mouseStart );
		
		tempBuffer->SetPortToGW( &savePort, &saveDev );
		
		// render the airbrush, or paint using the clone tool ( the same method does both ).
		
		if ( currentTool == stampTool )
		{
			// for cloning, we need to offset the <artSource> rect to represent where in the original
			// buffer we are cloning FROM. This rect is based on the cloneAnchor data member and
			// how far the mouse has moved since then.
			
			short dh, dv;
			
			dh = mouseStart.h - anchor.h;
			dv = mouseStart.v - anchor.v;	
			
			Pt2Rect( cloneAnchor, cloneAnchor, &artSource );
			InsetRect( &artSource, -airRadius, -airRadius );
			OffsetRect( &artSource, dh, dv );
		}
		
		DrawAirInPort( mouseStart, &artSource );	
	
		// copy the sprayed area to the window
		
		Pt2Rect( mouseStart, mouseStart, &shape );
		InsetRect( &shape, -airRadius, -airRadius );
		
		mURect = shape;
		MapRect( &mURect, &artFrame, &artBounds );
		
		SetGWorld( savePort, saveDev );
		
		tempBuffer->CopyOut( &windowPort->portBits, &shape, &mURect, ditherCopy );
		
		ReportMouseLoc( anchor, mouseStart, TRUE );
	}
	
	if ( usingAirPort )
		ForgetObject( airGW );
	
	// make change permanent
	
	try
	{
		curPTask = MakeUndoTask();
	}
	catch( OSErr err )
	{
		RemoveSecondBuffer();
		
		throw err;
	}
	
	artBuffer->SetPortToGW( &savePort, &saveDev );
	artBuffer->CopyIn( tempBuffer, &artFrame, &artFrame, srcCopy );	
	SetGWorld( savePort, saveDev );

	RemoveSecondBuffer();
}


/*--------------------------------*** CLICKPAINT  ***-----------------------------------*/
/*	
lower-level method to handle "one-shot" tools. This is a class of tool that requires a single
click such as fill bucket.
----------------------------------------------------------------------------------------*/

void		ZPaintEngine::ClickPaint( Point mouseStart )
{
	RgnHandle		maskRgn;
	
	ReportMouseLoc( mouseStart, mouseStart, TRUE );
	FailNIL( maskRgn = NewRgn());
	
	// do all the really difficult stuff:
	
	try
	{
		SeedFillToRegion( mouseStart, maskRgn );
		
		if ( ! EmptyRgn( maskRgn ))
			FillSelection( mouseStart, maskRgn );
	}
	catch( OSErr err )
	{
		DisposeRgn( maskRgn );
		
		throw err;
	}
	
	DisposeRgn( maskRgn );
}

/*-----------------------------*** SEEDFILLTOREGION  ***--------------------------------*/
/*
used as part of a paint fill and a "magic wand" click. This forms a region based on a
seed fill from the clicked point in the art buffer. This applies the search proc that
allows a tolerance	on the clicked colour. This gives good "natural" results and the
tolerance parameter can be adjusted.
----------------------------------------------------------------------------------------*/


void		ZPaintEngine::SeedFillToRegion( Point clickPt, RgnHandle aRgn )
{
	// SeedCFill REQUIRES the destination bitmap to be a true 2-byte aligned
	// bitmap- unfortunately a GWorld won't work.
	
	BitMap			maskBuffer;
	PixMapHandle	srcP;
	ColorSearchUPP	csUPP;
	CGrafPtr		savePort;
	GDHandle		saveDev;
	
	FailNILParam( aRgn );
	
	SetWatchCursor();
	
	// we're going to make a 1-bit mask buffer to form the seed fill mask
	
	MakeBitMap( &artFrame, &maskBuffer );
	
	// now use the SeedCFill procedure to actually calculate the mask. The filled area is
	// calculated in the artBuffer, and the maskBuffer is the destination. We use the same
	// search proc that the wand and lasso use so that some 'tolerance' is allowed for in
	// matching the seed colour. This gives better natural results than the default.
	
	csUPP = NewColorSearchProc( WandSearchProc );
	
	// undocumented, but true: The source image must be the current port for
	// SeedCFill to work correctly.
	
	artBuffer->SetPortToGW( &savePort, &saveDev );
	
	srcP = artBuffer->GetPixMap( TRUE );	// also lock it
	
	SeedCFill(  (BitMap*) *srcP, &maskBuffer,
				&artFrame, &artFrame,
				clickPt.h, clickPt.v,
				csUPP, matchTolerance );
	
	artBuffer->Unlock();
	DisposeRoutineDescriptor( csUPP );
	
	SetGWorld( savePort, saveDev );
	
	// now we can get the fill mask as a region:
	
	try
	{
		FailOSErr( BitMapToRegion( aRgn, &maskBuffer ));
	}
	catch( OSErr err )
	{
		DisposeBitMap( &maskBuffer );
		throw err;
	}
	DisposeBitMap( &maskBuffer );
}


/*-----------------------------*** FILLREGIONINPORT  ***--------------------------------*/
/*
fills the given region in the current port according to the fill style currently set. This
may be complex- for example if we have a gradient fill, this has to calculate the gradient
too.
----------------------------------------------------------------------------------------*/

void		ZPaintEngine::FillRegionInPort( RgnHandle aRgn, Point centre )
{
	RGBForeColor( &paintColour );
	
	switch ( currentFill )
	{
		case plainColourFill:
			PenPat( &qd.black );
			break;
		case patternFill:
			PenPat( &fillPattern );
			RGBBackColor( &eraseColour );
			break;
		case colourPatternFill:
			if ( cFillPattern )
				PenPixPat( cFillPattern );
			else
				PenPat( &fillPattern );
			break;
		case gradientFill:
			if ( curGradient )
				curGradient->GradientFillRegionInPort( aRgn, centre );
			break;
	}
	
	if ( currentFill != gradientFill )
	{
		if ( fillBkTransparent )
			PenMode( patOr );
		else
			PenMode( patCopy );
			
		PaintRgn( aRgn );
	}
		
	PenNormal();
	ForeColor( blackColor );
	BackColor( whiteColor );
}


/*----------------------------***  GETCLICKEDCOLOUR  ***--------------------------------*/
/*	
returns the RGB value of the colour in the artwork buffer located at point.
----------------------------------------------------------------------------------------*/

void		ZPaintEngine::GetClickedColour( Point mouse, RGBColor* aColour )
{
	CGrafPtr	savePort;
	GDHandle	saveDev;
	
	artBuffer->SetPortToGW( &savePort, &saveDev );
	GetCPixel( mouse.h, mouse.v, aColour );
	
	SetGWorld( savePort, saveDev );
}


/*-----------------------------***  GETPAINTCOLOURS  ***--------------------------------*/
/*	
returns the colour table used by the art buffer, if it has one. If the artBuffer is 16
or 24-bit colour, this returns NULL. The table returned is NOT a copy, so the caller should
take care not to dispose of it.
----------------------------------------------------------------------------------------*/

CTabHandle		ZPaintEngine::GetPaintColours()	
{
	CTabHandle	ct = NULL;
	
	if ( artBuffer )
	{
		if ( artBuffer->GetDepth() <= 8 )
			ct = artBuffer->GetColours();
	}
	
	return ct;
}


/*-------------------------------***  TWEENPAINT   ***----------------------------------*/
/*	
used by trail painting to fill gaps between sampled mouse points as the tool is dragged.
This uses Bresenham's algorithm to form the pixels along the line.
----------------------------------------------------------------------------------------*/

void		ZPaintEngine::TweenPaint( Point start, Point end )
{
   	register short 	d, x, y, ax, ay, sx, sy, dx, dy, x1, x2, y1, y2;
    register Point	sexy;

    x1 = start.h;
    x2 = end.h;
    y1 = start.v;
    y2 = end.v;
    
    
    dx = x2 - x1;  ax = abs( dx ) << 1;  sx = SGN( dx );
    dy = y2 - y1;  ay = abs( dy ) << 1;  sy = SGN( dy );
	x = x1;
	y = y1;
    
    if (ax >= ay)
    {	
    	// x dominant
		d = ay - (ax >> 1);
		while (x != x2)
		{
	    	sexy.h = x;
	    	sexy.v = y;
	    	DrawPaint( sexy );
	    	if (d >= 0)
	    	{
				y += sy;
				d -= ax;
	    	}
	    	x += sx;
	    	d += ay;
		}
    }
    else
    {	
    	//y dominant
		d = ax - (ay >> 1);
		while (y != y2)
		{
	    	sexy.h = x;
	    	sexy.v = y;
	    	DrawPaint( sexy );
	    	if (d >= 0)
	    	{
				x += sx;
				d -= ay;
	    	}
	    	y += sy;
	    	d += ax;
		}
    }
}


/*---------------------------------*** DRAWPAINT ***------------------------------------*/
/*	
draws a single "blob" of paint according to the current requirements in the current port
centred at the location passed. This is called by the tweening method.
----------------------------------------------------------------------------------------*/

void		ZPaintEngine::DrawPaint( Point location )
{
	Rect	pBlobRect;
	
	switch ( currentTool )
	{
		case paintBrushTool:
			Pt2Rect( location, location, &pBlobRect );
			InsetRect( &pBlobRect, -( brushWidth >> 1 ), -( brushHeight >> 1 )); 
			
			if ( paintBkTransparent )
				PenMode( patOr );
			else
				PenMode( patCopy );
			
			switch ( currentBrush )
			{
				case ovalBrush:
					PaintOval( &pBlobRect );
					break;
				case rectBrush:
					PaintRect( &pBlobRect );
					break;
				case slantLeftBrush:
					PenSize( 2, 2 );
					MoveTo( pBlobRect.left, pBlobRect.top );
					LineTo( pBlobRect.right, pBlobRect.bottom );
					break;
				case slantRightBrush:
					PenSize( 2, 2 );
					MoveTo( pBlobRect.right, pBlobRect.top );
					LineTo( pBlobRect.left, pBlobRect.bottom );
					break;
				case bitMapBrush:
					Rect	srcR = brushBits.bounds;
					CopyBits( 	&brushBits, &qd.thePort->portBits,
								&srcR, &pBlobRect,
								paintBkTransparent? srcOr : srcCopy,
								NULL );
				
					break;
				default:
					break;
			}
			break;
		case eraserTool:
			Pt2Rect( location, location, &pBlobRect );
			InsetRect( &pBlobRect, -kEraserToolHalfWidth, -kEraserToolHalfWidth );
			 
			PaintRect( &pBlobRect );
			break;
		case airbrushTool:
			break;
		case pencilTool:
			PenSize( 1, 1 );
			MoveTo( location.h, location.v );
			LineTo( location.h, location.v );
			break;
	}
}


/*-----------------------------*** PICKUPPAINTCOLOUR ***--------------------------------*/
/*
picks up the colour under the mouse location, and sets the appropriate colour data
members according to the current tool and modifiers passed.	
----------------------------------------------------------------------------------------*/

void		ZPaintEngine::PickUpPaintColour( Point location, short modifiers )
{
	RGBColor		picked;
	Point			lastPt = { -1, -1 }, anchor;
	GrafPtr			savePort;
	
	ReportMouseLoc( anchor = location, location, TRUE );
	
	GetPort( &savePort );
	
	while( WaitMouseUp())
	{
		SetPort( savePort );
		GetMouse( &location );
		
		TranslatePointToOffscreen( &location );
		
		if ( DeltaPoint( location, lastPt ))
		{	
			GetClickedColour( location, &picked );
			PickUpPaintColour( &picked, modifiers );
			
			lastPt = location;
			
			ReportMouseLoc( anchor, location, TRUE );
		}
	}
}

/*-----------------------------*** PICKUPPAINTCOLOUR ***--------------------------------*/

	
void		ZPaintEngine::PickUpPaintColour( RGBColor* aColour, short modifiers )
{
	RGBColor		picked = *aColour;
	Boolean			isCmd;
	Boolean			isCmdCtrl;

	isCmd = ( modifiers & cmdKey ) == cmdKey;
	isCmdCtrl = isCmd && (( modifiers & controlKey) == controlKey );
	
	// set the erase, paint or line colour according to tool selected.
	
	switch ( currentTool )
	{
		case paintBrushTool:
		case airbrushTool:
		case pencilTool:
		case fillTool:
		case rectSolidTool:
		case roundRectSolidTool:
		case ovalSolidTool:
		case arcSolidTool:
		case regularPolySolidTool:
		case irregPolySolidTool:
		case colourPickTool:
		case textTool:
			if ( isCmd )
			{
				if ( isCmdCtrl )
				{
					eraseColour = picked;
					SendMessage( peMsgNewEraseColourPicked, &picked );
				}
				else
				{
					lineColour = picked;
					SendMessage( peMsgNewLineColourPicked, &picked );
				}
			}
			else
			{
				paintColour = picked;
				SendMessage( peMsgNewPaintColourPicked, &picked );
			}
			break;
		case eraserTool:
			eraseColour = picked;
			SendMessage( peMsgNewEraseColourPicked, &picked );
			break;
		default:
			if ( isCmd )
			{
				paintColour = picked;
				SendMessage( peMsgNewPaintColourPicked, &picked );
			}
			else
			{
				lineColour = picked;
				SendMessage( peMsgNewLineColourPicked, &picked );
			}
			break;	
	}		
}


/*------------------------------*** REPORTMOUSELOC  ***---------------------------------*/
/*	
send info about the mouse location as painting proceeds
----------------------------------------------------------------------------------------*/

void		ZPaintEngine::ReportMouseLoc( Point anchor, Point loc, Boolean inDrag )
{
	MouseLocInfo	mi;
	
	mi.mouseLoc = loc;
	mi.mouseAnchor = anchor;
	mi.inDrag = inDrag;
	mi.xDelta = loc.h - anchor.h;
	mi.yDelta = loc.v - anchor.v;
	mi.inWindow = TRUE;
	
	CGrafPtr	savePort;
	GDHandle	saveDev;
	
	GetGWorld( &savePort, &saveDev );
	SetGWorld((CGrafPtr) windowPort, GetMainDevice());
	
	SendMessage( peMsgMouseLocationInfo, &mi );
	SetGWorld( savePort, saveDev );
}


/*-------------------------------*** BLITTOWINDOW  ***----------------------------------*/
/*	
updates the window interactively as painting proceeds. Ha ha- just like that!
----------------------------------------------------------------------------------------*/

void		ZPaintEngine::BlitToWindow( Rect* srcRect, Rect* destRect, ZGWorld* aBuffer )
{
	if ( windowPort != NULL )
	{
		ZGWorld			*saveOS, *buf;
		ZWindow*		zw = GetZWindow( windowPort );
		Rect			r;
		
		if ( zw )
		{
			if ( aBuffer )
			{
				saveOS = ((ZGWorldWindow*) zw )->GetGWorld();
				((ZGWorldWindow*) zw )->SetGWorld( aBuffer );
			}
			
			zw->Focus();
			zw->GetContentRect( &r );
			
			ClipRect( &r );
			
			buf = ((ZGWorldWindow*) zw )->GetGWorld();
			
			ForeColor( blackColor );
			BackColor( whiteColor );
			
			buf->CopyOut( &windowPort->portBits, srcRect, destRect );
			
			if ( aBuffer )
				((ZGWorldWindow*) zw )->SetGWorld( saveOS );
		}	
	}
}


void		ZPaintEngine::UpdateWindow()
{
	if ( windowPort )
	{
		ZWindow*	zw = GetZWindow( windowPort );
	
		if ( zw )
			zw->PostRefresh();
	}
}
	

/*-------------------------------*** MAKEUNDOTASK  ***----------------------------------*/
/*	
constructs an undo task object that can be sent to the painting window to make the paint
operation undoable. This uses the current tool settings to set the name of the task,
and contains a GWorld copy of the current artBuffer. If there's not enough memory to
make the task, this will throw an exception.
----------------------------------------------------------------------------------------*/

ZUndoPaintTask*	ZPaintEngine::MakeUndoTask( short taskID )	
{
	ZWindow*			target = NULL;
	ZUndoTask*			pt = NULL;
	Str63				ptStr;
	
	// get the name of the painting task according to the current tool, etc
	
	if ( taskID )
		GetIndString( ptStr, kPaintingTaskStrListID, taskID );
	else
	{
		if ( currentTool == selectionTool )
			GetIndString( ptStr, kPaintingTaskStrListID, 21 );
		else
			GetIndString( ptStr, kPaintingTaskStrListID, currentTool );
	}
	// try to make the task. This indirectly makes a complete clone of the artBuffer
	// GWorld, so requires a lot of memory. We need to be ready for this to fail so we
	// can inform the user that no undo is possible for this op.
	
	if ( windowPort )
		target = GetZWindow( windowPort );
	
	if ( target )
	{
		try
		{
			switch( taskID )
			{
				case kTaskPlaceGraphic:
					FailNIL( pt = new ZUndoPaintSelectionTask( ptStr, target )); 
					((ZUndoPaintSelectionTask*) pt )->SetSelectionInfo( theSelection, fbHOffset, fbVOffset );
					break;
				default:	
					FailNIL( pt = new ZUndoPaintTask( ptStr, target )); 
					break;
			}
		}
		catch( OSErr err )
		{
			// report no undo available
			
			short	cont = Alert( kNoUndoAvailableAlertID, NULL );
			
			// if user cancels op, throw quiet exception to abort
			
			if ( cont == cancel )
			{
				target->PostRefresh();
				FailOSErr( userCanceledErr );
			}
			
			pt = NULL;
		}
	}
	
	return (ZUndoPaintTask*) pt;
}


/*-------------------------------*** SENDUNDOTASK  ***----------------------------------*/
/*	
send current undo task to the window
----------------------------------------------------------------------------------------*/

void		ZPaintEngine::SendUndoTask()	
{
	if ( windowPort )
	{
		ZWindow*	zw = GetZWindow( windowPort );
		
		if ( zw )
			zw->SetTask( curPTask );
	}
}


/*------------------------*** TRANSLATEPOINTTOOFFSCREEN  ***----------------------------*/
/*	
translates a point in screen coordinates to the offscreen buffer, allowing for the scaling.
This ensures that the user gets the pixel they actually see when clicking.
----------------------------------------------------------------------------------------*/

void		ZPaintEngine::TranslatePointToOffscreen( Point* aPt )
{
	// given a point that is relative to <artBounds>, this maps it so that it is
	// relative to <artFrame>. All calculation are done relative to <artFrame>, but
	// functions like GetMouse() will be relative to artBounds.
	
	Point	m;
	
	m = *aPt;
	
	m.h -= m.h % hAlign;
	m.v -= m.v % vAlign;
	
	MapPt( &m, &artBounds, &artFrame );
	
	*aPt = m;
}


/*-------------------------*** TRANSLATERECTTOOFFSCREEN  ***----------------------------*/
/*	
translates a rect in screen coordinates to the offscreen buffer, allowing for the scaling.
----------------------------------------------------------------------------------------*/

void		ZPaintEngine::TranslateRectToOffscreen( Rect* aRect )
{
	Rect	r;
	short	h, v;
	
	r = *aRect;
	
	h = r.left % hAlign;
	v = r.top % vAlign;
	
	OffsetRect( &r, -h, -v );
	
	MapRect( &r, &artBounds, &artFrame );
	
	*aRect = r;
}


/*-----------------------------*** SETPORTPENPATTERN  ***-------------------------------*/
/*	
sets the current port's pen pattern according to the passed constant.
----------------------------------------------------------------------------------------*/

void		ZPaintEngine::SetPortPenPattern( const short which )
{	
	// set the current port's pen pattern. <which> can be 0, 1 or 2, indicating
	// to set the erase, frame or fill pattern respectively. Colour patterns override
	// black and white ones.
	
	switch ( which )
	{
		case kErase:
			break;
		case kFrame:
			if ( cFramePattern )
				PenPixPat( cFramePattern );
			else
				PenPat( &framePattern );
			break;
		case kFill:
			if ( cFillPattern )
				PenPixPat( cFillPattern );
			else
				PenPat( &fillPattern );
			break;
	}
}


/*-----------------------------*** MAKEREGULARPOLY  ***---------------------------------*/
/*	
returns a region as a regular polygon with the current number of sides, aligned to the
points passed.
----------------------------------------------------------------------------------------*/

RgnHandle		ZPaintEngine::MakeRegularPoly( Point centre, Point aVertex )
{
	// constructs a regular polygon with the number of sides currently set, then draws it in
	// the current port. The point passed in represent the centre and current mouse poistion-
	// the poly is drawn with one vertex at <mouse> and the centre at <centre>. The bounding
	// rect of the poly is returned in <bounds>.
	
	RgnHandle		thePoly;
	float			radius, angle, alpha;
	float			dx, dy;
	short			x, y, side;
	
	FailNIL( thePoly = NewRgn());
	
	dx = aVertex.h - centre.h;
	dy = aVertex.v - centre.v;
	
	radius = sqrt(( dx * dx ) + ( dy * dy ));
	alpha = ( 2 * pi )/ regularPolySides;	
	// this is the angle delta in radians
	
	// find the starting angle in radians
	
	angle = atan( dy / dx );
	if (( dx < 0 ) && ( regularPolySides & 1 ))
		angle -= alpha / 2;
		
	aVertex.h = centre.h + ( radius * cos( angle ));
	aVertex.v = centre.v + ( radius * sin( angle ));

	OpenRgn();
	MoveTo( aVertex.h, aVertex.v );
	
	for ( side = 0; side < regularPolySides - 1; side++ )
	{
		// compute the next point in the poly
		
		angle -= alpha;
		
		x = centre.h + (radius * cos( angle ));
		y = centre.v + (radius * sin( angle ));
		
		LineTo( x, y );	
	}
	LineTo( aVertex.h, aVertex.v );
	CloseRgn( thePoly );
	
	// OK, we have the "polygon" as a region, so we can return it.

	return thePoly;
}


/*-------------------------------*** DRAWAIRINPORT  ***---------------------------------*/
/*	
draws the current airbrush effect in the current port, aligned with <location>. Some
airbrush effects are generated using another offscreen port member- airGW. This also
implements the clone tool copying operation which is virtually identical to spraying.
----------------------------------------------------------------------------------------*/

void		ZPaintEngine::DrawAirInPort( Point location, Rect* artSource )
{
	// draws the airbrush effect in the current port, applying all airbrush parameters.
	Rect			airFrame, spRect, src;
	short			i, n, mode;
	RgnHandle		saveClip, newClip;
	ZGWorld*		tp;
	CGrafPtr		savePort;
	GDHandle		saveDev;
	PixMapHandle	app, tpp, bpp;
	
	Pt2Rect( location, location, &airFrame );
	InsetRect( &airFrame, -airRadius, -airRadius );
	RGBForeColor( &paintColour );
	
	// we also clip to the airFrame oval
	GetClip( saveClip = NewRgn());
	OpenRgn();
	FrameOval( &airFrame );
	CloseRgn( newClip = NewRgn());
	
	SectRgn( saveClip, newClip, newClip );
	SetClip( newClip );
	
	switch ( currentAirMode )
	{
		case darkenArt:
			mode = adMin;
			break;
		case lightenArt:
			mode = adMax;
			break;
		default:
		case stdAirMode:
			mode = srcCopy + ditherCopy;
			break;
	}
	
	if ( currentTool == stampTool )
	{
		// copy from artBuffer at <artSource> to current destination through mask.
		
		airGW->GetSize( &spRect );
		
		app = airGW->GetPixMap( TRUE );
		tpp = artBuffer->GetPixMap( TRUE );
		bpp = tempBuffer->GetPixMap( TRUE );
		
		// we need to clip to the artFrame to avoid copying garbage from off the edge
		// of the image, so do that now.
		
		SectRect( artSource, &artFrame, &src );
		
		// all these rects need to agree in size, so do that...
		
		if( ! EqualRect( artSource, &src ))
		{
			spRect.top += src.top - artSource->top;
			spRect.left += src.left - artSource->left;
			spRect.right -= artSource->right - src.right;
			spRect.bottom -= artSource->bottom - src.bottom;
		
			airFrame.top += src.top - artSource->top;
			airFrame.left += src.left - artSource->left;
			airFrame.right -= artSource->right - src.right;
			airFrame.bottom -= artSource->bottom - src.bottom;
		}
		
		if ( ! EmptyRect( &src ))
		{
			ForeColor( blackColor );
			BackColor( whiteColor );
			
			CopyDeepMask((BitMap*) *tpp, (BitMap*) *app, (BitMap*) *bpp,
							&src, &spRect, &airFrame, mode, NULL );
		}				
		airGW->Unlock();
		artBuffer->Unlock();
		tempBuffer->Unlock();
	}
	else
	{
		switch ( currentAirMode )
		{
			case darkenArt:
			case lightenArt:
			case stdAirMode:
				// make tp- a port with just the solid paint colour in it
				
				airGW->GetSize( &spRect );
				
				FailNIL( tp = new ZGWorld( spRect, 32 ));
				tp->SetPortToGW( &savePort, &saveDev );
				tpp = tp->GetPixMap( TRUE );
				RGBForeColor( &paintColour );
				PaintRect( &spRect );
				SetGWorld( savePort, saveDev );
				
				// this is a bizarre fudge to make CopyDeepMask happy- it skews the image
				// if the dest bounds is not aligned on a 4-pixel boundary.
				
				OffsetRect( &airFrame, -( airFrame.left % 4 ), 0 );
				
				app = airGW->GetPixMap( TRUE );
				bpp = tempBuffer->GetPixMap( TRUE );
				
				ForeColor( blackColor );
				CopyDeepMask((BitMap*) *tpp, (BitMap*) *app, (BitMap*) *bpp,
								&spRect, &spRect, &airFrame, mode, NULL );

				
				ForgetObject( tp );
				
				airGW->Unlock();
				tempBuffer->Unlock();
				break;
			case multiSplatter:
				// multisplatter is similar to stdSplatter, but sets the colour at random
				
				RGBColor	ms;
				
				ms.red = Random();
				ms.green = Random();
				ms.blue = Random();
				RGBForeColor( &ms );
				
				// fall through to std splatter case	
			case stdSplatter:
				// splatter works by drawing numerous coloured circles at random sizes within
				// the airFrame rectangle. Pressure sets the max size of the "blobs"
					
					short	p = MAX( 1, airPressure / 5 );
					short	t = abs( Random()) % p;
					
					n = MAX( 1, MIN( t, 20 ));
					
					SetRect( &spRect, 0, 0, n, n );
					
					// place randomly within airFrame
					
					OffsetRect( &spRect,
								airFrame.left + abs( Random()) % (( airRadius * 2) - n ),
								airFrame.top + abs( Random()) % (( airRadius * 2 ) - n ));
								
					PaintOval( &spRect ); 
				break;
		}
	}
	SetClip( saveClip );
	DisposeRgn( saveClip );
	DisposeRgn( newClip );	
}


/*---------------------------------*** BEGINTEXT  ***-----------------------------------*/
/*
starts a text editing operation by dragging a text edition rectangle in the image, then
setting up the offscreen ports, textEdit record, etc. Once this has happened, all clicks
and drags affect the text operation until EndText is called.	
----------------------------------------------------------------------------------------*/

void		ZPaintEngine::BeginText( Point mouseStart )
{
	DragTextRect( mouseStart );	
	
	// make a new TextEdit record for doing the editing
	
	if ( ! EmptyRect( &textRect ))
	{
		inTextMode = TRUE;
		
		Rect		r = textRect;
		CGrafPtr	savePort;
		GDHandle	saveDev;
		
		InsetRect( &r, 2, 2 );
		
		Rect	tr = r;
		
		tr.right += textShadowWidth;
		tr.bottom += textShadowHeight;
		
		// the text will be drawn into the second buffer as we type (using a DRAWHook so
		// we can properly set the various colouring characteristics, etc. )
		
		FailNIL( tempBuffer = new ZGWorld( tr, artBuffer->GetDepth()));
		
		tempBuffer->Lock();
		tempBuffer->SetPortToGW( &savePort, &saveDev );
		
		TextFont( textFont );
		TextFace( textStyle );
		TextSize( textSize );
		
		FailNIL( currTE = TENew( &r, &tr ));
		
		// set up a DRAWHook so we can draw characters in Paint Engine style rather than
		// the default style. On 68k without Mixed Mode, this won't work since it requires
		// register-based parameters. Solution sought (TO DO).
		
		#if GENERATINGCFM
		DrawHookUPP oldHook = gPEHook;
		TECustomHook( intDrawHook, &oldHook, currTE );
		#endif
		
		// the lineHeight of the text needs to be increased if we have a vertical shadow:
		
		(*currTE)->lineHeight += textShadowHeight;
		
		TESetAlignment( textAlign, currTE );
		TEActivate( currTE );
		
		tempBuffer->Unlock();
		SetGWorld( savePort, saveDev );
		
		InsetRect( &tr, -2, -2 );
		textRect = tr;
		InvalRect( &textRect );
		
		SendMessage( peMsgTextEditingBegun, NULL );
	}
}


/*-------------------------------*** DRAGTEXTRECT  ***----------------------------------*/
/*
drag a new text editing rectangle	
----------------------------------------------------------------------------------------*/

void		ZPaintEngine::DragTextRect( Point startPt )
{
	Point	a, b;
	Rect	lastR;
	
	a = b = startPt;
	
	SetPort( windowPort );
	PenMode( patXor );
	PenPat( &qd.gray );
	
	SetRect( &lastR, 0, 0, 0, 0 );
	
	ReportMouseLoc( a, startPt, TRUE );
	
	while( WaitMouseUp())
	{
		GetMouse( &startPt );
		
		if ( DeltaPoint( startPt, b ))
		{
			FrameRect( &lastR );
			
			Pt2Rect( a, startPt, &lastR );
			FrameRect( &lastR );
		
			b = startPt;
			ReportMouseLoc( a, startPt, TRUE );
		}
	}
	
	FrameRect( &lastR );
	textRect = lastR;
	PenNormal();
}


/*---------------------------------*** TEXTCLICK  ***-----------------------------------*/
/*	
mouse clicks while text editing is in operation come here. This either passes the click to
TextEdit, or finishes the operation if the text rect was clicked out of.
----------------------------------------------------------------------------------------*/

void		ZPaintEngine::TextClick( Point mousePt, short modifiers )
{
	if ( inTextMode )
	{
		// if the click was outside the text rect, end the editing
		
		if ( PtInRect( mousePt, &textRect ))
		{
			Rect		r;
			CGrafPtr	savePort;
			GDHandle	saveDev;
			
			tempBuffer->Lock();
			tempBuffer->SetPortToGW( &savePort, &saveDev );
			
			if ( currTE )
				TEClick( mousePt, ( modifiers & shiftKey ) == shiftKey , currTE );
			
			tempBuffer->Unlock();
			SetGWorld( savePort, saveDev );
			CopyTextBufToWindow();
		}
		else
			EndText();
	}
	else
		BeginText( mousePt );
}


/*---------------------------------*** TEXTTYPE  ***------------------------------------*/
/*	
keyboard input during text edition comes in here.
----------------------------------------------------------------------------------------*/

void		ZPaintEngine::TextType( char theChar )
{
	if ( inTextMode && currTE )
	{
		CGrafPtr	savePort;
		GDHandle	saveDev;
		
		tempBuffer->Lock();
		tempBuffer->SetPortToGW( &savePort, &saveDev );

		TEKey( theChar, currTE );
		
		tempBuffer->Unlock();
		SetGWorld( savePort, saveDev );
		CopyTextBufToWindow();
	}
}

/*----------------------------------*** ENDTEXT  ***------------------------------------*/
/*	
finishes the text edition. The text is rendered into an offscreen port and converted to
a fully-floating selection. That can then be dragged into position like any graphic.
----------------------------------------------------------------------------------------*/


void		ZPaintEngine::EndText()
{
	if ( inTextMode )
	{
		Boolean		hasText;
		GrafPtr		savePort;
		GetPort( &savePort );
		
		inTextMode = FALSE;
		
		if ( currTE )
		{
			hasText = (*currTE)->teLength > 0;
			
			TEDeactivate( currTE );
			TEDispose( currTE );
		}	
		currTE = NULL;
		
		if ( hasText )
		{
			// make a float buffer from the text imaging buffer, then build a selection
			// region that exactly selects the text. This is done by building a mask
			// bitmap and converting to a region.
			
			// remove any existing selection & float buffer
			
			SetSelection( NULL );
			
			Rect	fr, r = textRect;
			InsetRect( &r, 2, 2 );
			fr = r;
			OffsetRect( &fr, -fr.left, -fr.top );
			
			try
			{
				// make a 1-bit mask of the text image
				
				FailNIL( airGW = new ZGWorld( fr, 1 ));
				
				airGW->Clear();
				
				// install mask search proc
				
				ColorSearchUPP		maskUPP = NewColorSearchProc( MakeMaskSearchProc );
				airGW->InstallSearchProc( maskUPP );
				
				airGW->CopyIn( tempBuffer, &r, &fr );
				airGW->RemoveSearchProc( maskUPP );
				DisposeRoutineDescriptor( maskUPP );
				
				// make region from bitmap
				
				PixMapHandle	pp = airGW->GetPixMap( TRUE );
				
				FailOSErr( BitMapToRegion( theSelection, (BitMap*) *pp ));
				airGW->Unlock();
				ForgetObject( airGW );	
				
				short	fH, fV;
				
				// offset region to zero and make floatbuffer
				
				fH = (*theSelection)->rgnBBox.left;
				fV = (*theSelection)->rgnBBox.top;
				
				OffsetRgn( theSelection, -fH, -fV );
				fr = (*theSelection)->rgnBBox;
				r = fr;
				OffsetRect( &r, textRect.left + 2 + fH, textRect.top + 2 + fV );
				
				FailNIL( floatBuffer = new ZGWorld( fr, artBuffer->GetDepth()));
				
				// copy text buf to float buf
				
				floatBuffer->CopyIn( tempBuffer, &r, &fr );
				RemoveSecondBuffer();

				// set fb offsets:
				
				fbHOffset = r.left;
				fbVOffset = r.top;
				
				// offset region to right place
				
				OffsetRgn( theSelection, fbHOffset, fbVOffset );
			}
			catch( OSErr err )
			{
				RemoveSecondBuffer();
				
				SetPort( windowPort );
				MapRect( &textRect, &artFrame, &artBounds );
				InvalRect( &textRect );
				SetRect( &textRect, 0, 0, 0, 0 );
				
				throw err;	
			}
		}	
		SetPort( windowPort );
		MapRect( &textRect, &artFrame, &artBounds );
		InvalRect( &textRect );
		SetRect( &textRect, 0, 0, 0, 0 );
		
		SetPort( savePort );

		SendMessage( peMsgTextEditingEnded, NULL );
	}
}


/*-----------------------------*** COPYTEXTBUFTOWINDOW  ***-----------------------------*/
/*
copies tempBuffer to the window, applying zoom/scale, etc during editing.	
----------------------------------------------------------------------------------------*/

void		ZPaintEngine::CopyTextBufToWindow()
{
	Rect		r, wr;
	CGrafPtr	savePort;
	GDHandle	saveDev;

	r = textRect;
			
	InsetRect( &r, 2, 2 );
	wr = r;
	
	MapRect( &wr, &artFrame, &artBounds );
	
	::GetGWorld( &savePort, &saveDev );
	::SetGWorld((CGrafPtr) windowPort, GetMainDevice());
	
	tempBuffer->CopyOut( &windowPort->portBits, &r, &wr );
	
	::SetGWorld( savePort, saveDev );
}


/*---------------------------------*** RENDERTEXT  ***----------------------------------*/
/*
render the range of characters according to the current text paint settings.	
----------------------------------------------------------------------------------------*/

#define		qdFontFudgeVal	10


void		ZPaintEngine::RenderText( Ptr textBuf, unsigned short charOffset, unsigned short chCount )
{
	ZGWorld*		charBuf;
	Rect			charRect, dest;
	FontInfo		fi;
	CGrafPtr		savePort;
	GDHandle		saveDev;
	short			tFont;
	short			tSize;
	short			tFace;
	RgnHandle		charRgn;
	PixMapHandle	pp;
	
	// to pull off this feat, we need to completely replace the way QuickDraw renders text. To
	// do this we need to render the char to yet another offscreen port, convert the shape to
	// a region, then paint/fill/frame this as needed to get the desired effect.
	
	// so how big's this character gonna be?
	
	GetFontInfo( &fi );
	SetRect( &charRect, -qdFontFudgeVal, 0, fi.widMax + 1, fi.ascent + fi.descent + fi.leading + 1 );
	
	dest = charRect;
	
	OffsetRect( &dest, qd.thePort->pnLoc.h, qd.thePort->pnLoc.v - fi.ascent );
	
	// get font settings from current port
	
	tFont = qd.thePort->txFont;
	tSize = qd.thePort->txSize;
	tFace = qd.thePort->txFace;
	
	// make 1-bit offscreen to draw this one in temporary memory
	
	FailNIL( charBuf = new ZGWorld( charRect, 1, NULL, TRUE ));
	
	charRgn = NewRgn();

	// for each char in the range, render it
	
	short	i;
	char	aChar;
	
	pp = charBuf->GetPixMap( TRUE );
	
	charBuf->SetPortToGW( &savePort, &saveDev );
	TextFont( tFont );
	TextSize( tSize );
	TextFace( tFace );
	TextMode( srcCopy );
	PenNormal();
	SetGWorld( savePort, saveDev );
	
	for( i = 0; i < chCount; i++ )
	{
		aChar = *( textBuf + charOffset + i );
		
		charBuf->SetPortToGW( &savePort, &saveDev );
		charBuf->Clear();
		MoveTo( 0, fi.ascent );
		DrawChar( aChar );
		
		// we now have the char as a bitmap. In order to render it with the
		// full range of effects we are capable of, we need to convert to
		// a region and paint it
		
		try
		{
			FailOSErr( BitMapToRegion( charRgn, (BitMap*) *pp ));
		
			// we have the region. Paint it into tempBuffer at the relevant location using the
			// paint settings we wish to apply. The region may be empty- e.g a space character,
			// so we can save time by ignoring it.
			
			OffsetRgn( charRgn, dest.left + qdFontFudgeVal, dest.top );
			OffsetRect( &dest, CharWidth( aChar ), 0 );

			SetGWorld( savePort, saveDev );
			
			if ( ! EmptyRgn( charRgn ))
			{
				if ( tempBuffer )
					tempBuffer->Lock();
				
				// render text shadow if any
				
				if ( textShadowWidth > 0 ||
					 textShadowHeight > 0 )
				{
					PenPat( &qd.black );
					RGBForeColor( &lineColour );
				
					OffsetRgn( charRgn, textShadowWidth, textShadowHeight );
					PaintRgn( charRgn );
					OffsetRgn( charRgn, -textShadowWidth, -textShadowHeight );
				}
				
				// render character in current colour and pattern, including gradient (funky!)
				
				if ( textFill == gradientFill && curGradient )
				{
					Point	centre;
					
					centre.h = ((*charRgn)->rgnBBox.left + (*charRgn)->rgnBBox.right ) / 2;
					centre.v = ((*charRgn)->rgnBBox.top + (*charRgn)->rgnBBox.bottom ) / 2;
					
					try
					{
						curGradient->GradientFillRegionInPort( charRgn, centre );
					}
					catch( OSErr err )
					{
						if ( tempBuffer )
							tempBuffer->Unlock();
						
						throw err;
					}	
				}
				else
				{
					RGBForeColor( &paintColour );
					RGBBackColor( &eraseColour );
					SetPortPenPattern( kFill );
					
					// if transparent text fill, set the appropriate mode
					
					if ( txFillTransparent )
						PenMode( patOr );
					else
						PenMode( patCopy );
					
					PaintRgn( charRgn );
				}
				// if there's a border set, draw that
				
				if ( textBorderWidth > 0 )
				{
					RGBForeColor( &paintColour );
					PenPat( &qd.black );
					PenSize( textBorderWidth, textBorderWidth );
					PenMode( patCopy );
					FrameRgn( charRgn );
				}
				
				ForeColor( blackColor );
				BackColor( whiteColor );
				
				if ( tempBuffer )
					tempBuffer->Unlock();
			}
		}
		catch( OSErr err )
		{
			SetGWorld( savePort, saveDev );
			charBuf->Unlock();
			ForgetObject( charBuf );
			DisposeRgn( charRgn );
			
			throw err;
		}
	}
	charBuf->Unlock();
	ForgetObject( charBuf );
	SetGWorld( savePort, saveDev );
	DisposeRgn( charRgn );
}


#pragma mark -


/*-------------------------------*** PTINSELECTION  ***---------------------------------*/
/*	
returns TRUE if mouse point passed is inside the current selection region. Pt is in
window coordinates.
----------------------------------------------------------------------------------------*/

Boolean		ZPaintEngine::PtInSelection( const Point mouse )
{
	Point mp = mouse;
	
	TranslatePointToOffscreen( &mp );
	
	return PtInRgn( mp, theSelection );
}


/*-------------------------------*** DRAGSELECTION  ***---------------------------------*/
/*	
makes a new selection region by dragging or clicking. This also allows the selection to be
moved, either with or without the image.
----------------------------------------------------------------------------------------*/

void		ZPaintEngine::DragSelection( Point mouseStart, short modifiers )
{
	// set the selection modifier according to the keyboard modifier- +shift is add,
	// +shift +option is subtract. If mouse is inside existing region, this is a move
	// operation unless we're adding or subtracting
	
	Point	anchor = mouseStart;
	
	currentSelMod = replaceSelection;
	
	if ( modifiers & shiftKey )
	{
		if ( modifiers & optionKey )
			currentSelMod = subtractFromSelection;
		else
			currentSelMod = addToSelection;
	}
	else
	{
		if ( PtInRgn( mouseStart, theSelection ))
		{
			if ( modifiers & cmdKey )
				currentSelMod = moveSelection;
			else
			{
				if ( modifiers & optionKey )
					currentSelMod = moveSelectionWithImageCopy;
				else
					currentSelMod = moveSelectionWithImage;
			}
		}	
	}
	
	if ( currentSelMod >= moveSelection )
	{
		// we're moving the selection, possibly with the image. If with the image, we need
		// to make an Undo task so that the move can be undone.
		
		long		mOffset;
		Point		lastMouse = mouseStart;
		Rect		src, dest, uRect, mURect, lastDest;
		
		if ( paintMaskActive )
			RefreshMask();
		
		SetUpSecondBuffer( &artFrame );
		
		if ( floatBuffer == NULL )
			MakeFloatingSelection( currentSelMod == moveSelectionWithImage );
		else
		{
			// the floater already exists so we're just moving it some more. If the option
			// is top copy, blit the contents down to the art buffer first
			
			if ( currentSelMod == moveSelectionWithImageCopy )
			{
				floatBuffer->GetSize( &src );
				dest = src;
				OffsetRect( &dest, fbHOffset, fbVOffset );
				artBuffer->CopyIn( floatBuffer, &src, &dest, ditherCopy, theSelection );
			}
		}
		
		// by setting lastDest to the selection bounds, outset a little, it will automatically
		// erase the selection ants as we start to drag, which is naturally what we want...

		lastDest = (*theSelection)->rgnBBox;
		InsetRect( &lastDest, -1, -1 );
		
		// track the mouse
		
		ReportMouseLoc( anchor, mouseStart, TRUE );
		
		while( WaitMouseUp())
		{
			GetMouse( &mouseStart );
			TranslatePointToOffscreen( &mouseStart );
			
			if ( mOffset = DeltaPoint( mouseStart, lastMouse ))
			{
				// mouse moved- where to?
				
				fbHOffset += LoWord( mOffset );
				fbVOffset += HiWord( mOffset );
				OffsetRgn( theSelection, LoWord( mOffset ), HiWord( mOffset ));
				
				// draw floating buffer at that position
				
				floatBuffer->GetSize( &src );
				dest = src;
				OffsetRect( &dest, fbHOffset, fbVOffset );
				
				// compute union of this move and the last
				
				UnionRect( &dest, &lastDest, &uRect );
				tempBuffer->CopyIn( artBuffer, &uRect, &uRect );
				tempBuffer->CopyIn( floatBuffer, &src, &dest, ditherCopy, theSelection );
				
				mURect = uRect;
				MapRect( &mURect, &artFrame, &artBounds );
				
				// blit temp buffer to screen
				
				tempBuffer->CopyOut( &windowPort->portBits, &uRect, &mURect, ditherCopy );
				
				lastMouse = mouseStart;
				lastDest = dest;
				
				ReportMouseLoc( anchor, mouseStart, TRUE );
			}
		}
		RemoveSecondBuffer();
	}
	else
	{
		// here we are forming a new or modified selection region, so drag the frame out and
		// set it as the region.
		
		Point		lastMouse, anchor;
		Rect		selRect, lastRect;
		long		ignored;
		PicHandle	freePic;
		RgnHandle	temp = NewRgn();
			
		FailNIL( temp );
		
		SetPort( windowPort );
		GetMouse( &lastMouse );
		anchor = lastMouse;
		
		ReportMouseLoc( anchor, mouseStart, TRUE );

		if ( currentSelType == freeSelection ||
			 currentSelType == lassoSelection )
		{
			freePic = OpenPicture( &artBounds );
			ShowPen();
		}
		
		if ( currentSelType == wandSelection )
		{
			// for "magic wand", we use a similar technique to the paint fill to form
			// a region, but in this case we set it as the selection instead of filling it.
			
			SeedFillToRegion( mouseStart, temp );
		}
		else
		{
			PenPat((Pattern*) &ants );
			PenMode( patXor );
			PenSize( 1, 1 );
			
			SetRect( &lastRect, 0, 0, 0, 0 );
			selRect = lastRect;
			
			while( WaitMouseUp())
			{
				GetMouse( &mouseStart );
				
				if ( DeltaPoint( mouseStart, lastMouse ))
				{
					Pt2Rect( mouseStart, anchor, &selRect );
					
					switch ( currentSelType )
					{
						case rectSelection:
							FrameRect( &lastRect );	// erase old
							FrameRect( &selRect );	// draw new
							break;
						case ovalSelection:
							FrameOval( &lastRect );
							FrameOval( &selRect );
							break;
						case lassoSelection:
						case freeSelection:
							MoveTo( lastMouse.h, lastMouse.v );
							LineTo( mouseStart.h, mouseStart.v );
							break;
					}
			
					lastRect = selRect;
					lastMouse = mouseStart;
					
					ReportMouseLoc( anchor, mouseStart, TRUE );
				}
			}
			if ( currentSelType == freeSelection ||
				 currentSelType == lassoSelection )
			{
				MoveTo( lastMouse.h, lastMouse.v );
				LineTo( anchor.h, anchor.v );
				HidePen();
				ClosePicture();
			}
			PenNormal();
			UpdateWindow();
			
			TranslateRectToOffscreen( &selRect );
			
			switch ( currentSelType )
			{
				case rectSelection:
					RectRgn( temp, &selRect );
					break;
				case ovalSelection:
					OpenRgn();
					FrameOval( &selRect );
					CloseRgn( temp );
					break;
				case freeSelection:
				case lassoSelection:
					OpenRgn();
					DrawPicture( freePic, &artFrame );
					CloseRgn( temp );
					KillPicture( freePic );
					break;
			}
			
			if ( currentSelType == lassoSelection )
			{
				TranslatePointToOffscreen( &anchor );
				ShrinkSelection( anchor, temp );
			}
		}
		SetSelection( temp );
		DisposeRgn( temp );
	}
	
	// reset this in case something external wants to manipulate the region
	
	currentSelMod = replaceSelection;	
}


/*--------------------------------*** SETSELECTION  ***---------------------------------*/
/*	
sets the new selection to the passed region, refreshing the old one as well. This honours
the current selection modifier member to see if the region should be replaced, added to
or subtracted from.
----------------------------------------------------------------------------------------*/

void		ZPaintEngine::SetSelection( RgnHandle aRgn, Boolean killFB )
{
	Boolean		rDispose = FALSE;
	Boolean		changed = FALSE;
	
	if ( aRgn == NULL )
	{
		aRgn = NewRgn();
		rDispose = TRUE;
	}
	
	changed = ! EqualRgn( aRgn, theSelection );
	
	// if a floater currently active, fix it in place
	
	if ( killFB )
		KillFloatingSelection( TRUE );
	
	InvalSelection();
	
	switch ( currentSelMod )
	{
		default:
		case replaceSelection:
			CopyRgn( aRgn, theSelection );
			break;
		case addToSelection:
			UnionRgn( theSelection, aRgn, theSelection );
			break;
		case subtractFromSelection:
			DiffRgn( theSelection, aRgn, theSelection );
			break;
	}
	
	if ( rDispose )
		DisposeRgn( aRgn );
		
	if ( changed )
		SendMessage( peMsgSelectionRgnChanged, (void*) theSelection );
}


/*----------------------------------*** SELECTALL  ***----------------------------------*/
/*
select the entire image
----------------------------------------------------------------------------------------*/

void		ZPaintEngine::SelectAll()
{
	currentSelMod = replaceSelection;
	
	RgnHandle	aRgn;
	
	FailNIL( aRgn = NewRgn());
	RectRgn( aRgn, &artFrame );
	SetSelection( aRgn, TRUE );
	DisposeRgn( aRgn );
}


/*-------------------------------*** INVALSELECTION  ***--------------------------------*/
/*
invalidate the selection region's border in the window	
----------------------------------------------------------------------------------------*/

void		ZPaintEngine::InvalSelection()
{
	if ( ! EmptyRgn( theSelection ))
	{
		RgnHandle	t = NewRgn();
		RgnHandle	h = NewRgn();
		
		CopyRgn( theSelection, t );
		MapRgn( t, &artFrame, &artBounds );
		
		CopyRgn( t, h );
		InsetRgn( t, -1, -1 );
		InsetRgn( h, 1, 1 );
		DiffRgn( t, h, t );
		
		SetPort( windowPort );
		InvalRgn( t );
		
		DisposeRgn( t );
		DisposeRgn( h );
	}
}


/*------------------------------*** ANIMATESELECTION  ***-------------------------------*/
/*
shows the current selection in <windowPort> using the ususal marching ants technique.	
----------------------------------------------------------------------------------------*/

void		ZPaintEngine::AnimateSelection()
{
	RgnHandle		temp;
	GrafPtr			sPort;
	static long		antTicks = 0;
	
	GetPort( &sPort );
	
	if ( ! EmptyRgn( theSelection ) && antsActive )
	{
		if ( TickCount() > antTicks )
		{
			antTicks = TickCount();
			
			temp = NewRgn();
			CopyRgn( theSelection, temp );
			
			MapRgn( temp, &artFrame, &artBounds );
			SetPort( windowPort );
			
			PenPat((Pattern*) &ants );
			FrameRgn( temp );
			
			DisposeRgn( temp );
			PenNormal();
			
			// cycle the ants pat to make them animate
			
			char firstRow = ants[0];
			
			for( short i = 0; i < 7; i++ )
				ants[i] = ants[i+1];
				
			ants[7] = firstRow;
		}
	}
	
	// also blink caret if text editing happening
	
	if ( inTextMode && currTE )
	{
		CGrafPtr	savePort;
		GDHandle	saveDev;
		short		carState;
		
		carState = (*currTE)->caretState;
		
		tempBuffer->Lock();
		tempBuffer->SetPortToGW( &savePort, &saveDev );
		
		TEIdle( currTE );
		
		tempBuffer->Unlock();
		SetGWorld( savePort, saveDev );
		
		// copy text buffer back to window if caret changed
		
		if ( carState != (*currTE)->caretState )
			CopyTextBufToWindow();
	}
	
	SetPort( sPort );
}


/*------------------------------*** ACTIVATESELECTION  ***------------------------------*/
/*
sets the activation state for the ants. Usually called when the window is activated and
deactivated to animate the ants only in the active window.	
----------------------------------------------------------------------------------------*/

void		ZPaintEngine::ActivateSelection( Boolean state )
{
	if ( state != antsActive )
	{
		if ( inTextMode && currTE && !state )
			EndText();
		
		antsActive = state;
		
		if ( ! antsActive )
			InvalSelection();
	}
}


/*---------------------------*** MAKEFLOATINGSELECTION  ***-----------------------------*/
/*
copy the part of artBuffer that is enclosed by <theSelection> into the offscreen world
<floatBuffer>. This is used for moving selections.
----------------------------------------------------------------------------------------*/

void		ZPaintEngine::MakeFloatingSelection( Boolean eraseBehind )
{
	Rect	sFrame, src;
	
	if ( ! EmptyRgn( theSelection ))
	{
		if ( floatBuffer )
			KillFloatingSelection( FALSE );
		
		sFrame = (*theSelection)->rgnBBox;
		src = sFrame;
		
		fbHOffset = sFrame.left;
		fbVOffset = sFrame.top;
		
		OffsetRect( &sFrame, -fbHOffset, -fbVOffset );
		
		CTabHandle	artColours;
		
		artColours = artBuffer->GetColours();
		
		FailNIL( floatBuffer = new ZGWorld( sFrame, artBuffer->GetDepth(), artColours ));
		floatBuffer->Clear();
		floatBuffer->CopyIn( artBuffer, &src, &sFrame );
		
		if ( eraseBehind )
		{
			CGrafPtr	savePort;
			GDHandle	saveDev;
			
			artBuffer->SetPortToGW( &savePort, &saveDev );
			RGBForeColor( &eraseColour );
			PaintRgn( theSelection );
			ForeColor( blackColor );
			
			SetGWorld( savePort, saveDev );
		}
	}
}


/*---------------------------*** KILLFLOATINGSELECTION  ***-----------------------------*/
/*
destroy the float buffer, optionally copying it's contents back to the artBuffer at the
current offset.
----------------------------------------------------------------------------------------*/

void		ZPaintEngine::KillFloatingSelection( Boolean copyBack )
{
	if ( floatBuffer )
	{
		if ( copyBack )
		{
			Rect	src, dest;
			
			// make undo task for this
			
			if ( ! EmptyRgn( theSelection ))
				curPTask = MakeUndoTask( kTaskPlaceGraphic );
			
			floatBuffer->GetSize( &src );
			dest = src;
			OffsetRect( &dest, fbHOffset, fbVOffset );
			
			// if paint mask active, subtract it from mask
			
			RgnHandle	mask = NewRgn();
			
			CopyRgn( theSelection, mask );
			
			if ( paintMaskActive )
				DiffRgn( mask, paintMask, mask );
			
			artBuffer->CopyIn( floatBuffer, &src, &dest, ditherCopy, mask );
			
			DisposeRgn( mask );
			
			// task will be sent to window when click method returns, unless
			// the task was NULL, in which case we need to send it here
			
			if ( curPTask == NULL )
				SendUndoTask();
		}	
		RgnHandle	h = NewRgn();
		
		CopyRgn( theSelection, h );
		MapRgn( h, &artFrame, &artBounds );
		
		SetPort( windowPort );
		InvalRgn( h );
		
		DisposeRgn( h );
		
		ForgetObject( floatBuffer );
	}
}


/*------------------------------*** SHRINKSELECTION  ***--------------------------------*/
/*
for lasso selection, and other "shrinking" selections, this will shrink the region around
bits in artBuffer that do not match the colour at cSeedPt. Normally this is the point where
the lasso drag started.
----------------------------------------------------------------------------------------*/

void		ZPaintEngine::ShrinkSelection( Point cSeedPt, RgnHandle aRgn )
{
	RGBColor		seed;
	CGrafPtr		savePort;
	GDHandle		saveDev, curDev;
	BitMap			maskBits;
	MatchRec		matchInfo;
	ColorSearchUPP	csUPP;
	PixMapHandle	srcPix;
	Rect			r;
	RgnHandle		tempRgn = NULL;
	
	if ( ! EmptyRgn( aRgn ))
	{
		// get the colour we're going to ignore
		
		artBuffer->SetPortToGW( &savePort, &saveDev );
		GetCPixel( cSeedPt.h, cSeedPt.v, &seed );
		
		// make a bitmap to hold the destination mask. This is the size of the bounds of the
		// selection region only
		r = (*aRgn)->rgnBBox;
		MakeBitMap( &r, &maskBits );
	
		// we're going to copy the artBuffer into the mask buffer using a custom search
		// proc that will only pass pixels that don't match the seed. We stuff the matchRec
		// struct into the current GDevice's refCon so that the search proc can find it.
		
		matchInfo.red = seed.red;
		matchInfo.green = seed.green;
		matchInfo.blue = seed.blue;
		matchInfo.matchData = matchTolerance;
		
		curDev = GetGDevice();
		(*curDev)->gdRefCon = (long) &matchInfo;
		
		// install a search proc
		
		csUPP = NewColorSearchProc( WandSearchProc );
		AddSearch( csUPP );
		
		// copy the bits to the mask, going through the search proc we just installed. This
		// will make 1 bits only where the colour doesn't match the seed, allowing for the
		// tolerance set.
		
		srcPix = artBuffer->GetPixMap( TRUE );	// lock it
		CopyBits((BitMap*) *srcPix, &maskBits, &r, &r, srcCopy, aRgn );
		artBuffer->Unlock();
		
		// Okey-doke, Now tear down a few things...
		
		DelSearch( csUPP );
		DisposeRoutineDescriptor( csUPP );
		(*curDev)->gdRefCon = 0;
		
		// make the bitmap into a region
		
		try
		{
			FailNIL( tempRgn = NewRgn());
			FailOSErr( BitMapToRegion( tempRgn, &maskBits ));
		}
		catch( OSErr err )
		{
			if ( tempRgn )
				DisposeRgn( tempRgn );			
			DisposeBitMap( &maskBits );
			
			SetGWorld( savePort, saveDev );
			throw err;
		}
		
		DisposeBitMap( &maskBits );
		
		// this is now the shrunk region, so replace the selection. We do this directly,
		// since this method is actually called by SetSelection if the tool is lasso.
		
		CopyRgn( tempRgn, aRgn );
		DisposeRgn( tempRgn );
		
		SetGWorld( savePort, saveDev );	
	}
}


/*-----------------------------*** UPDATEFLOATTOWINDOW  ***-----------------------------*/
/*
called by owning window to ensure the current floater, if any, is drawn in the window
----------------------------------------------------------------------------------------*/

void		ZPaintEngine::UpdateFloatToWindow()
{
	CGrafPtr	sPort;
	GDHandle	sDev;
		
	if ( floatBuffer )
	{
		Rect		src, dest;
		RgnHandle	h = NewRgn();
		
		floatBuffer->GetSize( &src );
		dest = src;
		OffsetRect( &dest, fbHOffset, fbVOffset );
		
		MapRect( &dest, &artFrame, &artBounds );
		
		CopyRgn( theSelection, h );
		MapRgn( h, &artFrame, &artBounds );
		floatBuffer->CopyOut( &windowPort->portBits, &src, &dest, ditherCopy, h );
		
		DisposeRgn( h );
	}
	
	// also draws the mask region, if it is visible & active
	
	RefreshMask();
		
	// this also draws the boundary of the text rect, if text editing is in progress
	
	if ( inTextMode )
	{
		Rect	wr = textRect;
		
		MapRect( &wr, &artFrame, &artBounds );
		
		GetGWorld( &sPort, &sDev );
		SetPort( windowPort );
		PenPat( &qd.gray );
		PenSize( 1, 1 );
		
		FrameRect( &wr );
		PenNormal();
		SetGWorld( sPort, sDev );
		
		tempBuffer->Lock();
		tempBuffer->SetPortToGW( &sPort, &sDev );
		
		if ( currTE )
			TEUpdate( &textRect, currTE );
		
		SetGWorld( sPort, sDev );
		tempBuffer->Unlock();
		
		CopyTextBufToWindow();		
	}
}


/*--------------------------------*** CLEARSELECTION  ***-------------------------------*/
/*
deletes the floating buffer or erases the selection region in the artbuffer, setting
the selection to empty in either case.
----------------------------------------------------------------------------------------*/

void		ZPaintEngine::ClearSelection()
{
	if ( floatBuffer )
	{
		KillFloatingSelection( FALSE );
		SetSelection( NULL );
	}
	else
	{
		if ( ! EmptyRgn( theSelection ))
		{
			curPTask = MakeUndoTask( kTaskClear );
			
			CGrafPtr	savePort;
			GDHandle	saveDev;
		
			artBuffer->SetPortToGW( &savePort, &saveDev );
			artBuffer->Lock();
			
			RGBForeColor( &eraseColour );
			PaintRgn( theSelection );
			ForeColor( blackColor );
			
			SetGWorld( savePort, saveDev );
			artBuffer->Unlock();
			
			SetPort( windowPort );
			MapRgn( theSelection, &artFrame, &artBounds );
			InvalRgn( theSelection );
			
			SetEmptyRgn( theSelection );
			SendUndoTask();	
		}
	}
}


/*-------------------------------------*** CLEAR  ***-----------------------------------*/
/*
erases the entire image to the background colour, with undo. Also kills selection.
----------------------------------------------------------------------------------------*/

void	ZPaintEngine::Clear()
{
	PaintTool	saveTool = currentTool;
	CGrafPtr	savePort;
	GDHandle	saveDev;
	
	SetSelection( NULL );
	
	if ( windowPort )
	{	
		currentTool = eraserTool;
		curPTask = MakeUndoTask();
		
		if ( artBuffer )
		{
			artBuffer->Lock();
			artBuffer->SetPortToGW( &savePort, &saveDev );
			
			PrepareMaskClipForPort();
		
			RGBForeColor( &eraseColour );
			PaintRect( &artFrame );
			
			ForeColor( blackColor );
			SetGWorld( savePort, saveDev );
			artBuffer->Unlock();
		}
	
		// send the undo task
		
		SendUndoTask();
				
		// put the tool back
		
		currentTool = saveTool;
		
		// update window
		
		BlitToWindow( &artFrame, &artBounds );
	}
}

/*---------------------------------*** COPYSELECTION  ***-------------------------------*/
/*
copies the float buffer to the clipboard, or the part of artBuffer under the selection.
----------------------------------------------------------------------------------------*/

void		ZPaintEngine::CopySelection()
{
	PicHandle	p;
		
	if ( ! EmptyRgn( theSelection ))
	{
		Boolean		noFloat = FALSE;
		
		if ( floatBuffer == NULL )
		{
			MakeFloatingSelection();
			noFloat = TRUE;
		}
		
		p = floatBuffer->MakePicture();
		gClipboard->PutData( p );
		KillPicture( p );
		
		if ( noFloat )
			KillFloatingSelection( FALSE );	
			
		gClipboard->AppendData('CRGN', (Handle) theSelection );
	}
}


/*--------------------------------*** PASTESELECTION  ***-------------------------------*/
/*
makes a floating buffer from the contents of the clipboard, and positions it initially in
the middle of the image.
----------------------------------------------------------------------------------------*/

void		ZPaintEngine::PasteSelection()
{
	PicHandle	p;
	
	if ( gClipboard->QueryType( 'PICT' ))
	{
		FailNIL( p = (PicHandle) gClipboard->GetData( 'PICT' ));
		
		if ( floatBuffer )
			KillFloatingSelection( TRUE );
			
		FailNIL( floatBuffer = new ZGWorld( p ));
		
		Rect	r;
		
		r = (*p)->picFrame;
		
		KillPicture( p );
		
		OffsetRect( &r, -r.left, -r.top );
		OffsetRect( &r, ( artFrame.right / 2 ) - ( r.right / 2 ),
						( artFrame.bottom / 2) - ( r.bottom / 2 ));
						
		fbHOffset = r.left;
		fbVOffset = r.top;
		
		RgnHandle	h;
		
		if ( gClipboard->QueryType( 'CRGN' ))
		{
			h = (RgnHandle) gClipboard->GetData('CRGN');
		
			OffsetRgn( h,fbHOffset - (*h)->rgnBBox.left,fbVOffset - (*h)->rgnBBox.top );
		}
		else
		{
			h = NewRgn();
			RectRgn( h, &r );
		}
		SetSelection( h, FALSE );
		DisposeRgn( h );
	
		SetPort( windowPort );
		MapRect( &r, &artFrame, &artBounds );
		InvalRect( &r );
	}
}


/*--------------------------------*** FILLSELECTION  ***--------------------------------*/
/*
fills the passed region as if the paint bucket were clicked in it. If the region is NULL,
the current selection is used instead. If the point passed is -1, -1, the centre of the
region is used.
----------------------------------------------------------------------------------------*/

void		ZPaintEngine::FillSelection( Point centre, RgnHandle aRgn )
{
	Boolean		madeRegion = FALSE;
	Boolean		fillFloat = FALSE;
	
	if (( aRgn == NULL ) && ! EmptyRgn( theSelection ))
	{
		FailNIL( aRgn = NewRgn());
		CopyRgn( theSelection, aRgn );
		madeRegion = TRUE;
	}
	
	if ( centre.h == -1 && centre.v == -1 )
	{
		centre.h = ((*aRgn)->rgnBBox.left + (*aRgn)->rgnBBox.right ) / 2;
		centre.v = ((*aRgn)->rgnBBox.top + (*aRgn)->rgnBBox.bottom ) / 2;
	}
	
	fillFloat = madeRegion && ( floatBuffer != NULL );
	
	// make a task so the fill is undoable
	
	if ( ! fillFloat )
	{
		try
		{
			curPTask = MakeUndoTask( fillTool );
		}
		catch( OSErr err )
		{
			if ( madeRegion )
				DisposeRgn( aRgn );
			throw err;
		}
	}
	// so far so good- now fill the region in the artbuffer, and update
	// the window.
	
	ZGWorld*	fBuf;
	CGrafPtr	savePort;
	GDHandle	saveDev;
	
	// if we're filling the selection and there's a floatbuffer, fill the floatbuffer
	// not the art buffer.
	
	if ( fillFloat )
	{
		fBuf = floatBuffer;
		OffsetRgn( aRgn, -fbHOffset, -fbVOffset );
		centre.h -= fbHOffset;
		centre.v -= fbVOffset;
	}
	else
		fBuf = artBuffer;
	
	// perform the fill
	
	fBuf->SetPortToGW( &savePort, &saveDev );
	
	// include mask if necessary
	
	if ( ! fillFloat )
		PrepareMaskClipForPort();
	
	try
	{
		FillRegionInPort( aRgn, centre );
	}
	catch( OSErr err )
	{
		SetGWorld( savePort, saveDev );
		
		if ( madeRegion )
			DisposeRgn( aRgn );
		throw err;
	}
	SetGWorld( savePort, saveDev );
	
	// update the window:
	
	UpdateWindow();
	
	if ( madeRegion )
	{
		DisposeRgn( aRgn );
		
		if ( ! fillFloat )
			SendUndoTask();
	}
}


/*-------------------------------------*** FLIP  ***------------------------------------*/
/*
flips the float buffer horizontally or vertically, using current selection as a mask
----------------------------------------------------------------------------------------*/

void		ZPaintEngine::Flip( const short flipDir )
{
	if ( floatBuffer == NULL )
		MakeFloatingSelection( FALSE );	
		
	RgnHandle	sel;
	
	FailNIL( sel = NewRgn());
	GetSelection( sel );
	
	// make sure selection is aligned with floatbuffer
	
	OffsetRgn( sel, -fbHOffset, -fbVOffset );
	
	if ( flipDir == kFlipHorizontal )
		floatBuffer->FlipHorizontal( sel );
	else
		floatBuffer->FlipVertical( sel );
	
	OffsetRgn( sel, fbHOffset, fbVOffset );	
	SetSelection( sel, FALSE );
	DisposeRgn( sel );
	
	// refresh window to show change
	
	curPTask = NULL;
	SendUndoTask();
	UpdateFloatToWindow();
}


/*------------------------------------*** INVERT  ***-----------------------------------*/
/*
inverts the selection, the floatbuffer or the entire image depending on the selection state
----------------------------------------------------------------------------------------*/

void	ZPaintEngine::Invert()
{
	if ( EmptyRgn( theSelection ))
	{
		// invert the whole image. This requires an undo task
		
		try
		{
			curPTask = MakeUndoTask( invertTool );
		}
		catch( OSErr err )
		{
			curPTask = NULL;
		}
		
		artBuffer->Invert();
		BlitToWindow( &artFrame, &artBounds );
		
		SendUndoTask();	
	}
	else
	{
		// invert the floatbuffer, making it if need be.
		
		if ( floatBuffer == NULL )
			MakeFloatingSelection();
			
		floatBuffer->Invert();	
		
		curPTask = NULL;
		SendUndoTask();
		UpdateFloatToWindow();
	}
}


#pragma mark -


/*------------------------------*** SETMASKTOSELECTION  ***-----------------------------*/
/*
sets the mask to be equal to the current selection
----------------------------------------------------------------------------------------*/

void	ZPaintEngine::SetMaskToSelection()
{
	CopyRgn( theSelection, paintMask );
	RefreshMask();
}

/*-----------------------------------*** ADDTOMASK  ***---------------------------------*/
/*
adds the selection region to the mask region
----------------------------------------------------------------------------------------*/

void	ZPaintEngine::AddToMask()
{
	UnionRgn( theSelection, paintMask, paintMask );
	RefreshMask();
}


/*-------------------------------*** SUBTRACTFROMMASK  ***------------------------------*/
/*
subtracts the selection region from the mask region
----------------------------------------------------------------------------------------*/

void	ZPaintEngine::SubtractFromMask()
{
	DiffRgn( paintMask, theSelection, paintMask );
	RefreshMask();
}


/*-----------------------------------*** CLEARMASK  ***---------------------------------*/
/*
makes the mask region empty
----------------------------------------------------------------------------------------*/

void	ZPaintEngine::ClearMask()
{
	SetEmptyRgn( paintMask );
}


/*-----------------------------------*** INVERTMASK  ***--------------------------------*/
/*
inverts the current paint mask
----------------------------------------------------------------------------------------*/

void	ZPaintEngine::InvertMask()
{
	RgnHandle	temp;
	
	FailNIL( temp = NewRgn());
	
	RectRgn( temp, &artFrame );
	DiffRgn( temp, paintMask, paintMask );
	
	DisposeRgn( temp );
	
	RefreshMask();
}


/*----------------------------------*** SHOWHIDEMASK  ***-------------------------------*/
/*
sets the visible state of the mask
----------------------------------------------------------------------------------------*/

void	ZPaintEngine::ShowHideMask( Boolean showIt )
{
	if ( showIt != paintMaskVisible )
	{
		paintMaskVisible = showIt;
	
		RefreshMask();
	}
}


/*---------------------------------*** ACTIVATEMASK  ***--------------------------------*/
/*
sets the activae state of the mask- i.e. whether it affects painting or not.
----------------------------------------------------------------------------------------*/

void	ZPaintEngine::ActivateMask( Boolean isActive )
{
	if ( isActive != paintMaskActive )
		paintMaskActive = isActive;
}


/*----------------------------------*** REFRESHMASK  ***--------------------------------*/
/*
update mask on screen
----------------------------------------------------------------------------------------*/

#define		kMaskPPatID		128


void	ZPaintEngine::RefreshMask()
{
	if ( paintMaskVisible && paintMaskActive )
	{
		RgnHandle		temp;
		CGrafPtr		sPort;
		GDHandle		sDev;
		PixPatHandle	pp;
	
		FailNIL( temp = NewRgn());
		
		CopyRgn( paintMask, temp );
		MapRgn( temp, &artFrame, &artBounds );
		
		GetGWorld( &sPort, &sDev );
		SetPort( windowPort );
		
		pp = GetPixPat( kMaskPPatID );
		
		if ( pp )
		{
			PenPixPat( pp );
			PenMode( patCopy );
		}
		else
		{
			PenPat( &qd.gray );
			ForeColor( redColor );
			PenMode( patOr );	
		}
		
		PaintRgn( temp );
		PenNormal();
		ForeColor( blackColor );
		
		if ( pp )
			DisposePixPat( pp );
		
		DisposeRgn( temp );
		SetGWorld( sPort, sDev );
	}
}


/*---------------------------*** PREPAREMASKCLIPFORPORT  ***----------------------------*/
/*
removes the mask region fromn the current port's clip region if active
----------------------------------------------------------------------------------------*/

void	ZPaintEngine::PrepareMaskClipForPort()
{
	GrafPtr	savePort;
	
	GetPort( &savePort );
	
	if ( paintMaskActive )
	{
		RgnHandle	clip;
		
		FailNIL( clip = NewRgn());
		
		GetClip( clip );
		DiffRgn( clip, paintMask, clip );
	
		SetClip( clip );
		
		// also remove mapped clip from windowPort
	
		SetPort( windowPort );
		
		CopyRgn( paintMask, clip );
		MapRgn( clip, &artFrame, &artBounds );
		
		DiffRgn( qd.thePort->clipRgn, clip, qd.thePort->clipRgn );
		DisposeRgn( clip );
	}
	else
	{
		ClipRect( &qd.thePort->portRect );
		SetPort( windowPort );
		
		ZWindow*	zw = GetZWindow( windowPort );
		
		Rect	r;
		
		zw->GetContentRect( &r );
		ClipRect( &r );
	}	
	SetPort( savePort );
}


#pragma mark -
/*--------------------------------------------------------------------------------------*/


// macro ABS() is unreliable- unsure why at present but this works instead.

static short abs( short a )
{	
	if ( a  < 0 )
		return -a;
	else
		return a;
}


/*--------------------------------------------------------------------------------------*/


Boolean	EqualColour( RGBColor* a, RGBColor* b )
{
	// return TRUE if colours are equal, else FALSE.

	return  ( a->red   & 0xF000 ) == ( b->red   & 0xF000 )	&&
			( a->green & 0xF000 ) == ( b->green & 0xF000 )	&&
			( a->blue  & 0xF000 ) == ( b->blue  & 0xF000 );
}


/*--------------------------------------------------------------------------------------*/

static pascal Boolean WandSearchProc( RGBColor* rgb, long* position )
{
	// search procedure that assists in implementing the wand and lasso tools
	
	register MatchRec*	mrp;
	register GDHandle	curDevice;
	register short		delta;
	
	curDevice = GetGDevice();
	mrp = (MatchRec*)(*curDevice)->gdRefCon;
	delta = LoWord( mrp->matchData );
	
	if ((rgb->red 	>= mrp->red    - delta) &&
		(rgb->red 	<= mrp->red    + delta) &&
		(rgb->green >= mrp->green  - delta) &&
		(rgb->green <= mrp->green  + delta) &&
		(rgb->blue 	>= mrp->blue   - delta) &&
		(rgb->blue 	<= mrp->blue   + delta))
		*position = 0;
	else
		*position = 1;

	return TRUE;
}


/*--------------------------------------------------------------------------------------*/


static pascal Boolean	MakeMaskSearchProc( RGBColor* rgb, long* position )
{
	// search procedure that maps everything to black except white, This is used to build
	// masks from coloured images.

	if ( rgb->red 	>= 0xFF00 	&&
		 rgb->green >= 0xFF00 	&&
		 rgb->blue 	>= 0xFF00 )
		*position = 0;
	else
		*position = 1;
		
	return TRUE;
}

/*--------------------------------------------------------------------------------------*/


static pascal void 		PEDrawHookProc( unsigned short textOffset, unsigned short drawLen, void *textBufferPtr, TEPtr pTE, TEHandle hTE )
{
	// call gPaintEngine to render the character in the port
	try
	{
		gPaintEngine->RenderText((Ptr) textBufferPtr, textOffset, drawLen );
	}
	catch( OSErr err )
	{
		SysBeep( 1 );
	}
}
