/*************************************************************************************************
*
*
*			ObjectMacZapp		-- a standard Mac OOP application template
*
*
*
*			ZPaintWindow.cpp 	-- a window you can paint in with the paint engine
*
*
*
*
*			 1996, Graham Cox
*
*
*
*
*************************************************************************************************/

#include	"ZPaintWindow.h"
#include	"ZPaintEngine.h"
#include	"ZPaintApplication.h"
#include	"MacZoop.h"
#include	"ZGradientDialog.h"
#include	"ZUndoPaintTask.h"
#include	"ZMouserWindow.h"
#include	"ZImageFilter.h"
#include	"ZFilterDialog.h"
#include	"ZJPEGFile.h"
#include	"ZGIFFile.h"

ZPaintEngine*	gPaintEngine = NULL;


extern PaintPrefsHdl	gPrefsH;


/*---------------------------------*** CONSTRUCTOR ***----------------------------------*/

ZPaintWindow::ZPaintWindow( ZCommander* aBoss, short windID )
	: ZGWorldWindow( aBoss, windID )
{
	FailNIL( gPaintEngine );
	itsSelection = NewRgn();
	rulersVisible = FALSE;
}


/*---------------------------------*** DESTRUCTOR ***-----------------------------------*/

ZPaintWindow::~ZPaintWindow()
{
	DisposeRgn( itsSelection );
}


/*------------------------------------*** CLOSE ***-------------------------------------*/
/*	
kill text editing before closing as usual
----------------------------------------------------------------------------------------*/

Boolean	ZPaintWindow::Close( const short phase )
{
	// kill any text editing going on in the paint engine
	
	gPaintEngine->EndText();
	gPaintEngine->SetArtBuffer( NULL );

	return inherited::Close( phase );
}


/*--------------------------------*** CLICKCONTENT ***----------------------------------*/
/*	
initialise the window, and make a new, blank GWorld for it.
----------------------------------------------------------------------------------------*/

void	ZPaintWindow::InitZWindow()
{
	inherited::InitZWindow();
	
	// use the size and depth settings from the global prefs
	// if the size is set to "same as monitor", find out that size now
	
	if ((*gPrefsH)->sizeSameAsMonitor )
	{
		GDHandle	mainGD;
		
		mainGD = GetMainDevice();
		
		(*gPrefsH)->imgWidth = (*mainGD)->gdRect.right - (*mainGD)->gdRect.left;
		(*gPrefsH)->imgHeight = (*mainGD)->gdRect.bottom - (*mainGD)->gdRect.top;
	}
	
	MakeNewGWorld(	(*gPrefsH)->imgWidth,
					(*gPrefsH)->imgHeight,
					(*gPrefsH)->depthSameAsMonitor? 0 : (*gPrefsH)->imgColourDepth );
					
	SetupWindowFromGWorld();
					
	// set size constraint of window to be a bit bigger than default so hBar can't disappear
	
	Rect	r = sizeRect;
	
	r.left += 100;
	SetSizeRect( r );
}


/*--------------------------------*** CLICKCONTENT ***----------------------------------*/
/*	
pass mouse clicks in the window to the paint engine, first ensuring that the paint engine
is set up correctly for the current display state
----------------------------------------------------------------------------------------*/

void	ZPaintWindow::ClickContent( const Point mouse, const short modifiers )
{
	if ( gPaintEngine )
	{
		Rect		r;
		short		h, v;
		Point		cmpMouse;
		PaintTool	pt;
		
		CalcBoundsRect( &r );
		GetPosition( &h, &v );
		OffsetRect( &r, -h, -v );
		
		gPaintEngine->SetArtBounds( r, scale );
		gPaintEngine->SetWindowPort( macWindow );
		gPaintEngine->SetArtBuffer( itsOffscreen );
		
		// if the zoom tool is the one we're currently using, handle the click directly
		// and don't pass it to the paint engine.
		
		pt = gPaintEngine->GetPaintTool();
		
		if ( pt == zoomTool )
			ZoomToPoint( mouse, ( modifiers & optionKey ) == optionKey );
		else
		{
			GetContentRect( &r );
			ClipRect( &r );
			
			cmpMouse = mouse;
			cmpMouse.h -= h;
			cmpMouse.v -= v;
			
			gPaintEngine->MouseClick( cmpMouse, modifiers );
		}
	}
}


/*--------------------------------*** HANDLECOMMAND ***---------------------------------*/
/*
set the paint engine tool from the passed command- this is a quick-and-dirty user interface
for testing the paint engine (menu command used to select tool)	
----------------------------------------------------------------------------------------*/

void	ZPaintWindow::HandleCommand( const long aCmd )
{
	if (( aCmd > kFirstToolCmd ) && ( aCmd < 1100 ) && gPaintEngine )
	{
		PaintTool cmd = ( PaintTool )( aCmd - kFirstToolCmd );
		
		gPaintEngine->SetCurrentTool( cmd );
		
		// if the command was for the text tool, enable the font/size/style menus
		
		if ( cmd == textTool )
		{
			gMenuBar->SetMenuDimming( kFontMenuID, neverDim );
			gMenuBar->SetMenuDimming( kStyleMenuID, neverDim );
			gMenuBar->SetMenuDimming( kFontSizeMenuID, neverDim );
		}
		else
		{
			gMenuBar->SetMenuDimming( kFontMenuID, neverDim + dimTitle );
			gMenuBar->SetMenuDimming( kStyleMenuID, neverDim + dimTitle );
			gMenuBar->SetMenuDimming( kFontSizeMenuID, neverDim + dimTitle );
		}	
	}
	else
	{
		if (( aCmd > 1100 ) && gPaintEngine)
		{
			// it's a selection command, so set the selection tool and sub-type
			
			gPaintEngine->SetCurrentTool( selectionTool );	
			gPaintEngine->SetSelectionType(( SelectionType )( aCmd - 1100 ));
		}
		else
		{
			switch ( aCmd )
			{
				case kCmdEditGradients:
					DoDialog( kGradientDialogID );
					break;
				case kCmdEditPatterns:
					DoDialog( kPatternsDialogID );
					break;
				case kCmdEditFilter:
					DoDialog( kFilterDialogID );
					break;
				case kCmdEditStamps:
					DoDialog( kStampsDialogID );
					break;
				case kCmdSelectNone:
					gPaintEngine->SetSelection( NULL );
					break;	
				case kCmdFatbitsGrid:
					fatBitsGrid = !fatBitsGrid;
					if ( scale > 400 )
						PostRefresh();
					break;
				case kCmdFlipHorizontal:
					gPaintEngine->Flip( kFlipHorizontal );
					break;
				case kCmdFlipVertical:
					gPaintEngine->Flip( kFlipVertical );
					break;
				case kCmdFillSelection:
					Point centre = { -1, -1};
					gPaintEngine->FillSelection( centre, NULL );
					break;
				case kCmdInvert:
					gPaintEngine->Invert();
					break;
				case kCmdAddToMask:
					gPaintEngine->AddToMask();
					break;
				case kCmdSubFromMask:
					gPaintEngine->SubtractFromMask();
					if ( gPaintEngine->MaskVisible())
						PostRefresh();
					break;
				case kCmdClearMask:
					gPaintEngine->ClearMask();
					if ( gPaintEngine->MaskVisible())
						PostRefresh();
					break;
				case kCmdInvertMask:
					gPaintEngine->InvertMask();
					if ( gPaintEngine->MaskVisible())
						PostRefresh();
					break;
				case kCmdShowHideMask:
					gPaintEngine->ShowHideMask( ! gPaintEngine->MaskVisible());
					PostRefresh();
					break;
				case kCmdActivateMask:
					gPaintEngine->ActivateMask( ! gPaintEngine->MaskActive());
					if ( gPaintEngine->MaskVisible())
						PostRefresh();
					break;
				case kCmdSetMask:
					gPaintEngine->SetMaskToSelection();
					if ( gPaintEngine->MaskVisible())
						PostRefresh();
					break;
				case kCmdSwitchToBlackAndWhite:
					SwitchDepth( 1 );
					break;
				case kCmdSwitchTo16Greys:
					SwitchDepth( 4 | 0x8000 );
					break;
				case kCmdSwitchTo16Colours:
					SwitchDepth( 4 );
					break;
				case kCmdSwitchTo256Greys:
					SwitchDepth( 8 | 0x8000 );
					break;
				case kCmdSwitchTo256Colours:
					SwitchDepth( 8 );
					break;
				case kCmdSwitchToThousands:
					SwitchDepth( 16 );
					break;
				case kCmdSwitchToMillions:
					SwitchDepth( 32 );
					break;
				case kCmdSet100PercentZoom:
					SetScale( 100 );
					PostRefresh();
					break;
				default:
					inherited::HandleCommand( aCmd );
					break;
			}
		}
	}
}


/*--------------------------------*** GETCONTENTRECT ***--------------------------------*/
/*
get content area, allowing for rulers if visible.	
----------------------------------------------------------------------------------------*/

void	ZPaintWindow::GetContentRect( Rect* aRect )
{
	inherited::GetContentRect( aRect );
	
	if ( rulersVisible )
	{
		aRect->top += kRulerWidth;
		aRect->left += kRulerWidth;
	}
}


/*-------------------------------------*** TYPE ***-------------------------------------*/
/*
if backspace key, do clear operation	
----------------------------------------------------------------------------------------*/

void	ZPaintWindow::Type( const char theKey )
{
	if (( theKey == BACKSPACE_KEY ) && ! gPaintEngine->InTextMode())
		DoClear();
	else
	{
		gPaintEngine->TextType( theKey );
		inherited::Type( theKey );
	}
}


/*------------------------------------*** DOCUT ***-------------------------------------*/
/*
cut paint selection to clipboard	
----------------------------------------------------------------------------------------*/

void	ZPaintWindow::DoCut()
{
	if ( gPaintEngine )
	{
		gPaintEngine->CopySelection();
		gPaintEngine->ClearSelection();
	}
}


/*-----------------------------------*** DOCOPY ***-------------------------------------*/
/*	
copy paint selection to the clipboard
----------------------------------------------------------------------------------------*/

void	ZPaintWindow::DoCopy()
{
	if ( gPaintEngine )
		gPaintEngine->CopySelection();
}


/*-----------------------------------*** DOPASTE ***------------------------------------*/
/*	
paste clipboard image as floating selection in image
----------------------------------------------------------------------------------------*/

void	ZPaintWindow::DoPaste()
{
	if ( gPaintEngine )
		gPaintEngine->PasteSelection();
}


/*-----------------------------------*** DOCLEAR ***------------------------------------*/
/*	
remove or erase selection in image
----------------------------------------------------------------------------------------*/

void	ZPaintWindow::DoClear()
{
	if ( gPaintEngine )
		gPaintEngine->ClearSelection();
}


/*--------------------------------*** DOSELECTALL ***-----------------------------------*/
/*	
select the whole image
----------------------------------------------------------------------------------------*/

void	ZPaintWindow::DoSelectAll()
{
	if ( gPaintEngine )
		gPaintEngine->SelectAll();
}


/*---------------------------------*** UPDATEMENUS ***----------------------------------*/
/*
enable relevant menu commands	
----------------------------------------------------------------------------------------*/

void	ZPaintWindow::UpdateMenus()
{
	if ( gPaintEngine )
	{
		if ( gPaintEngine->HasSelection())
		{
			// enable commands relating to selection region
			
			gMenuBar->EnableCommand( kCmdCut );
			gMenuBar->EnableCommand( kCmdCopy );
			gMenuBar->EnableCommand( kCmdClear );
			gMenuBar->EnableCommand( kCmdSelectNone );
			
			gMenuBar->EnableCommand( kCmdFlipHorizontal );
			gMenuBar->EnableCommand( kCmdFlipVertical );
			
			gMenuBar->EnableCommand( kCmdFillSelection );
			
			// activate mask manipulation commands
			
			gMenuBar->EnableCommand( kCmdAddToMask );
			gMenuBar->EnableCommand( kCmdSubFromMask );
			gMenuBar->EnableCommand( kCmdSetMask );
		}
		
		// enable general image commands
		
		gMenuBar->EnableCommand( kCmdSelectAll );
		gMenuBar->EnableCommand( kCmdEditGradients );
		gMenuBar->EnableCommand( kCmdEditPatterns );
		gMenuBar->EnableCommand( kCmdEditStamps );
		
		// enable mask commands and set wording
		
		gMenuBar->EnableCommand( kCmdShowHideMask );
		
		if ( gPaintEngine->MaskVisible())
			gMenuBar->SetCommandText( kCmdShowHideMask, kAltMenuStringsID, 4 );
		else
			gMenuBar->SetCommandText( kCmdShowHideMask, kAltMenuStringsID, 3 );
		
		gMenuBar->EnableCommand( kCmdClearMask );
		gMenuBar->EnableCommand( kCmdInvertMask );
		gMenuBar->EnableCommand( kCmdActivateMask );
		
		if ( gPaintEngine->MaskActive())
			gMenuBar->SetCommandText( kCmdActivateMask, kAltMenuStringsID, 2 );
		else
			gMenuBar->SetCommandText( kCmdActivateMask, kAltMenuStringsID, 1 );
			
		// set a check mark against the current depth
		
		if ( itsOffscreen )
		{
			gMenuBar->EnableCommand( kCmdEditFilter );
			
			// enable depth switch commands
			
			gMenuBar->EnableCommand( kCmdSwitchToBlackAndWhite );
			gMenuBar->EnableCommand( kCmdSwitchTo16Greys );
			gMenuBar->EnableCommand( kCmdSwitchTo16Colours );
			gMenuBar->EnableCommand( kCmdSwitchTo256Greys );
			gMenuBar->EnableCommand( kCmdSwitchTo256Colours );
			gMenuBar->EnableCommand( kCmdSwitchToThousands );
			gMenuBar->EnableCommand( kCmdSwitchToMillions );
			
			short	aDepth = itsOffscreen->GetDepth();
			short	cInfo  = itsOffscreen->GetColourInfo();
			short	dCmd;
			
			// determine if grey or not
			
			Boolean	isGrey = ( cInfo == 40 || cInfo == 36 );
			
			// figure out associated command
			
			switch ( aDepth )
			{
				case 1:
					dCmd = kCmdSwitchToBlackAndWhite;
					break;
				case 4:
					if ( isGrey )
						dCmd = kCmdSwitchTo16Greys;
					else
						dCmd = kCmdSwitchTo16Colours;
					break;
				case 8:
					if ( isGrey )
						dCmd = kCmdSwitchTo256Greys;
					else
						dCmd = kCmdSwitchTo256Colours;
					break;
				case 16:
					dCmd = kCmdSwitchToThousands;
					break;
				case 32:
					dCmd = kCmdSwitchToMillions;
					break;
			}
			gMenuBar->CheckCommand( dCmd, TRUE );
		}
	}	

	inherited::UpdateMenus();
}


/*--------------------------------*** ADJUSTCURSOR ***----------------------------------*/
/*
set cursor shape according to paint tool selected	
----------------------------------------------------------------------------------------*/

void	ZPaintWindow::AdjustCursor( const Point mouse, const short modifiers)
{
	PaintTool 	curTool;
	short		cursID;
	CursHandle	cH;
	Rect		cr;
	
	GetContentRect( &cr );
	
	if ( PtInRect( mouse, &cr ) && gPaintEngine )
	{
		curTool = gPaintEngine->GetPaintTool();
		
		if ( curTool == zoomTool )
		{
			if ( modifiers & optionKey )
				cursID = kZoomMinusCursorID;
			else
				cursID = kZoomPlusCursorID;
		}
		else
		{
			if ( curTool == stampTool )
			{
				if ( modifiers & optionKey )
					cursID = kAltStampCursorID;
				else
					cursID = ((short) curTool ) + kFirstToolCmd;
			}
			else
			{
				if ( curTool == selectionTool )
				{
					// display different cursor according to modifiers/ selection type. If mouse within selection,
					// change to reverse arrow unless shift or option is down.
					
					if ( gPaintEngine->PtInSelection( mouse ) && ( modifiers & ( shiftKey | optionKey )) == 0 )
						cursID = kAltArrowCURSID;
					else
					{
						SelectionType	st = gPaintEngine->GetSelectionType();
						
						switch ( st )
						{
							case rectSelection:
							case ovalSelection:
							case freeSelection:
								cursID = kRectSelectionCursorID;
								break;
							case lassoSelection:
								cursID = kLassoSelectCursorID;
								break;
							case wandSelection:
								cursID = kMagicWandCursorID;
								break;
						}
						
						// cursor IDs are arranged so that adding 4 or 8 to the ID gets the
						// associated plus or minus cursor
						
						if ( modifiers & shiftKey )
						{
							if ( modifiers & optionKey )
								cursID += 8;
							else
								cursID += 4;
						}	
					}
				}
				else
				{
					// if option key is down, display picker
					
					if ( modifiers & optionKey )
						cursID = kColourPickerCursorID;
					else
					{
						cursID = ((short) curTool ) + kFirstToolCmd;
						
						if ( curTool >= rectFrameTool &&
							 curTool < stampTool )
							cursID = ((short) rectFrameTool ) + kFirstToolCmd;
					}
				}
			}
		}
		cH = GetCursor( cursID );
		
		if ( cH )
			SetCursor( *cH );
	}
	else
		SetCursor( &qd.arrow );
		
	ReportMouseLoc( mouse, mouse, FALSE );
}


/*------------------------------------*** IDLE ***--------------------------------------*/
/*
animate selection if necessary	
----------------------------------------------------------------------------------------*/

void	ZPaintWindow::Idle()
{
	if ( gPaintEngine )
	{
		Rect	r;
		short	h, v;
		
		Focus();
		CalcBoundsRect( &r );
		GetPosition( &h, &v );
		OffsetRect( &r, -h, -v );
		
		gPaintEngine->SetArtBounds( r, scale );
		gPaintEngine->SetWindowPort( macWindow );
		gPaintEngine->SetArtBuffer( itsOffscreen );
		
		GetContentRect( &r );
		ClipRect( &r );
		
		gPaintEngine->AnimateSelection();
	}
}


/*-------------------------------------*** DRAW ***-------------------------------------*/
/*
draw window, including rulers
----------------------------------------------------------------------------------------*/

void	ZPaintWindow::Draw()
{
	inherited::Draw();
	
	if ( rulersVisible )
	{
		MoveTo( kRulerWidth - 1, 0 );
		LineTo( kRulerWidth - 1, macWindow->portRect.bottom - kStdScrollbarWidth );
		MoveTo( 0, kRulerWidth - 1 );
		LineTo( macWindow->portRect.right - kStdScrollbarWidth, kRulerWidth - 1 );
	}
}


/*---------------------------------*** DRAWCONTENT ***----------------------------------*/
/*
update window, also overlaying any floating buffer and selection. This ensures all elements
of the image are shown, even when scrolling.	
----------------------------------------------------------------------------------------*/

void	ZPaintWindow::DrawContent()
{
	inherited::DrawContent();
	
	if ( gPaintEngine )
	{
		Rect	r;

		CalcBoundsRect( &r );
		
		gPaintEngine->SetArtBounds( r, scale );
		gPaintEngine->SetWindowPort( macWindow );
		
		GetContentRect( &r );
		ClipRect( &r );

		gPaintEngine->UpdateFloatToWindow();
	}	
	Idle();
	
	DrawImageInfo();
}


/*---------------------------------*** MAKEGWORLD ***-----------------------------------*/
/*
make offscreen image, and also paint engine object for editing it.	
----------------------------------------------------------------------------------------*/

void	ZPaintWindow::MakeGWorld()
{
	inherited::MakeGWorld();
	
	gPaintEngine->ActivateSelection( FALSE );
	gPaintEngine->SetArtBuffer( itsOffscreen, FALSE );
}



/*--------------------------------*** MAKENEWGWORLD ***---------------------------------*/
/*
make a new blank GWorld when the window is first ogPaintEnginened	
----------------------------------------------------------------------------------------*/

void	ZPaintWindow::MakeNewGWorld( const short aWidth, const short aHeight, const short aDepth )
{
	Rect		aSize;
	short		depth;
	CTabHandle	ct = NULL;
	
	SetRect( &aSize, 0, 0, aWidth, aHeight );
	
	if ( aDepth == 0 )
	{
		// use depth of main monitor if depth passed is 0.
		
		GDHandle	mainGD = GetMainDevice();
		
		depth = (*(*mainGD)->gdPMap)->pixelSize;
	}
	else
	{
		switch ( aDepth )
		{
			case 40:
				depth = 8;
				ct = GetCTable( 40 );
				break;
			case 36:
				depth = 4;
				ct = GetCTable( 36 );
				break;
			default:
				depth = aDepth;
				break;
		}
	}
	
	if ( itsOffscreen )
		ForgetObject( itsOffscreen );
			
	FailNIL( itsOffscreen = new ZGWorld( aSize, depth, ct ));
	itsOffscreen->Clear();
	
	if ( ct )
		DisposeCTable( ct );
	
	gPaintEngine->ActivateSelection( FALSE );
}


/*----------------------------*** SETUPWINDOWFROMGWORLD ***-----------------------------*/
/*
set scroll bounds and window size from GWorld dimensions	
----------------------------------------------------------------------------------------*/

void	ZPaintWindow::SetupWindowFromGWorld()
{
	Rect	aSize;
	
	SetScrollAmount(8,8);
	CalcBoundsRect( &aSize );
	SetBounds( aSize );
	
	SetSize( aSize.right + kStdScrollbarWidth, aSize.bottom + kStdScrollbarWidth );
	
	MakePaletteForWindow();
	
	gPaintEngine->SetArtBuffer( itsOffscreen, FALSE );
}


/*----------------------------------*** ACTIVATE ***------------------------------------*/
/*
activate selection	
----------------------------------------------------------------------------------------*/

void	ZPaintWindow::Activate()
{
	inherited::Activate();
	
	if ( gPaintEngine )
	{
		gPaintEngine->SetSelection( itsSelection );
		gPaintEngine->ActivateSelection( TRUE );
		gPaintEngine->SetWindowPort( macWindow );
		gPaintEngine->SetArtBuffer( itsOffscreen );
	}	
	SendMessage( kPaintWindowBecomeActive, (void*)((long)( gPaintEngine->GetPaintTool()) + kFirstToolCmd ));
}


/*---------------------------------*** DEACTIVATE ***-----------------------------------*/
/*
deactivate selection	
----------------------------------------------------------------------------------------*/

void	ZPaintWindow::Deactivate()
{
	if ( gPaintEngine )
	{
		gPaintEngine->ActivateSelection( FALSE );
		gPaintEngine->GetSelection( itsSelection );
	}
	inherited::Deactivate();
}


/*-----------------------------------*** REVERT ***-------------------------------------*/
/*
revert the image, then activate to keep paint engine sweet.	
----------------------------------------------------------------------------------------*/

void	ZPaintWindow::Revert()
{
	inherited::Revert();
	Activate();
}


/*----------------------------------*** OPENFILE ***------------------------------------*/
/*
open the file. If JPEG, handle usinga JPEGFile object
----------------------------------------------------------------------------------------*/

void	ZPaintWindow::OpenFile( const OSType fType )
{
	SetScale( 100 );
	
	if ( fType == 'JPEG' ||
		 fType == 'JFIF' ||
		 fType == 'GIFf')
	{
		ZJPEGFile*	zf;
		
		if ( fType == 'GIFf' )
			FailNIL( zf = ( ZJPEGFile*) new ZGIFFile( macFile ));
		else
			FailNIL( zf = new ZJPEGFile( macFile ));
		
		try
		{
			zf->Open();
			
			// make a GWorld object- this will be replaced by opening the file, so
			// we just make a very small one to conserve memory
			
			MakeNewGWorld( 4, 4, 1 );
			
			zf->Read( itsOffscreen );
			zf->Close();
			
			// set up window parameters
			
			SetupWindowFromGWorld();
			
			ZWindow::OpenFile( fType );		// set window title, etc.
		}
		catch( OSErr err )
		{
			ForgetObject( zf );
			
			throw err;
		}
		
		ForgetObject( zf );
	}
	else
		inherited::OpenFile( fType );	// inherited method can handle PICT files.
	
	SendMessage( kPaintWindowImageLoaded, NULL );
}


/*-------------------------------*** MOVESCROLLBARS ***---------------------------------*/
/*
move scrollbars- this is almost the same as the default behaviour, but shortens the
horizointal scrollbar on the left to make room for some image info.	
----------------------------------------------------------------------------------------*/

void	ZPaintWindow::MoveScrollbars()
{
	inherited::MoveScrollbars();
	
	if ( hasHBar )
	{
		short	w, h;
		
		w = macWindow->portRect.right - macWindow->portRect.left - kStdScrollbarWidth - kInfoWidth + 1;
		h = kStdScrollbarWidth + 1;
		
		SizeControl( theHBar, w, h );
		MoveControl( theHBar, macWindow->portRect.left + kInfoWidth, macWindow->portRect.bottom - kStdScrollbarWidth );
	}
}


/*-------------------------------*** RECEIVEMESSAGE ***---------------------------------*/
/*
receive message. Here we just wait for a filter dialog to close.	
----------------------------------------------------------------------------------------*/

 void	ZPaintWindow::ReceiveMessage( ZComrade* aSender, long aMsg, void* msgData )
{
	if ( aMsg == kFilterMsgDialogClosed )
	{
		// about to filter- make undo task.
		
		ZUndoPaintTask*	undoTask;
		Str255			taskStr;
		
		SetWatchCursor();
		GetIndString( taskStr, 201, 26 );
		FailNIL( undoTask = new ZUndoPaintTask( taskStr, this ));
		
		try
		{
			ZImageFilter	zf( itsOffscreen, NULL );
		
			zf.SetKernel((double*) msgData );
			zf.SetOffset(((ZFilterDialog*) aSender)->GetAddOffset());
			zf.ApplyFilter();
			
			SetTask( undoTask );
			PostRefresh();
		}
		catch( OSErr err )
		{
			ForgetObject( undoTask );
			
			throw err;
		}
	}
}

/*----------------------------------*** DODIALOG ***------------------------------------*/
/*
handle the various auxiliary dialog	boxes
----------------------------------------------------------------------------------------*/

void	ZPaintWindow::DoDialog( const short dialogID )
{
	ZDialog*	zd;
	
	switch ( dialogID )
	{
		case kGradientDialogID:
			FailNIL( zd = new ZGradientDialog( gApplication, kGradientDialogID ));
			break;
		case kPatEditorDialogID:
			FailNIL( zd = new ZAdvancedDialog( gApplication, kPatEditorDialogID ));
			break;
		case kFilterDialogID:
			FailNIL( zd = new ZFilterDialog( this, kFilterDialogID ));
			break;
		case kStampsDialogID:
			FailNIL( zd = new ZAdvancedDialog( this, kStampsDialogID ));
			break;
	}
	
	try
	{
		zd->InitZWindow();
	}
	catch( OSErr err )
	{
		ForgetObject( zd );
		
		throw err;
	}
	zd->Select();
}


/*--------------------------------*** DRAWIMAGEINFO ***---------------------------------*/
/*
draw the size and depth info in the bottom left corner
----------------------------------------------------------------------------------------*/

void	ZPaintWindow::DrawImageInfo()
{
	Rect	r, s;
	Str15	temp;
	Str31	infStr;
	
	if ( itsOffscreen )
	{
	
		r = macWindow->portRect;
		r.top = r.bottom - kStdScrollbarWidth + 2;
		r.right = r.left + kInfoWidth;
		
		ClipRect( &r );
	
		itsOffscreen->GetSize( &s );
		
		NumToString( s.right - s.left, infStr );
		ConcatPStrings( infStr, "\p x " );
		NumToString( s.bottom - s.top, temp );
		ConcatPStrings( infStr, temp );
		ConcatPStrings( infStr, "\p x " );
		NumToString( itsOffscreen->GetDepth(), temp );
		ConcatPStrings( infStr, temp );
		ConcatPStrings( infStr, "\p bits." );
		
		TextFont( geneva );
		TextSize( 9 );
		TextFace( 0 );
		TETextBox( &infStr[1], infStr[0], &r, teJustCenter );
	}
}


/*---------------------------------*** SWITCHDEPTH ***----------------------------------*/
/*
change the colour depth of the image to that passed. A grey image is indicated by the high
bit being set.
----------------------------------------------------------------------------------------*/

void	ZPaintWindow::SwitchDepth( const short aDepth )
{
	if ( itsOffscreen )
	{
		short		depth, curDepth = itsOffscreen->GetDepth();
		short		cInfo = itsOffscreen->GetColourInfo();
		Boolean		isGrey = ( cInfo == 36 || cInfo == 40 );
		
		depth = aDepth & 0x7FFF;
		
		if (( curDepth != depth ) ||
			( isGrey != (( aDepth & 0x8000 ) == 0x8000 )))
		{
			// we could do with an undo task here so this can be undone.
			
			Str63			ptStr;
			ZUndoPaintTask*	pTask;
			
			GetIndString( ptStr, 201, 25 );
			FailNIL( pTask = new ZUndoPaintTask( ptStr, this ));
			
			// want a grey image?
			CTabHandle	ct = NULL;
			
			if ( depth < 16 )
			{ 
				if ( aDepth & 0x8000 )
					ct = GetCTable( depth + 32 );
				else
					ct = GetCTable( depth );
			}
			
			// set image to new colours and depth
			
			try
			{
				if ( curDepth != depth )
					itsOffscreen->SetDepth( depth, ct, TRUE );
				else
					itsOffscreen->SetColours( ct, TRUE );
			}
			catch( OSErr err )
			{
				if ( ct )
					DisposeCTable( ct );
					
				ForgetObject( pTask );
				
				throw err;
			}
			
			if ( ct )
				DisposeCTable( ct );
			
			SetTask( pTask );
			PostRefresh();
			
			SendMessage( kPaintWindowDepthChanged, &aDepth );
		}
	}
}


/*------------------------------*** REPORTMOUSELOC ***----------------------------------*/
/*
send a message about the mouse location. This is called from many click and drag ops in
this window, as well as when the mouse is over the window.
----------------------------------------------------------------------------------------*/


void	ZPaintWindow::ReportMouseLoc( Point anchor, Point loc, Boolean inDrag )
{
	static Point	lastMouse = { -1, -1 };
	
	MouseLocInfo	mi;
	Rect			r;
	
	if ( DeltaPoint( loc, lastMouse ))
	{
		GetContentRect( &r );
		
		// mouse location is transmitted in terms of the point in the art buffer,
		// not the window, so that we report correctly regardless of scroll and
		// zoom of the image.
		
		gPaintEngine->TranslatePointToOffscreen( &anchor );
		gPaintEngine->TranslatePointToOffscreen( &loc );
		
		mi.mouseLoc = loc;
		mi.mouseAnchor = anchor;
		mi.inDrag = inDrag;
		mi.xDelta = loc.h - anchor.h;
		mi.yDelta = loc.v - anchor.v;
		mi.inWindow = PtInRect( loc, &r );
		
		SendMessage( kPaintWindowMouseInfo, &mi );
		
		lastMouse = loc;
	}
}
